Package: vitals
Title: Large Language Model Evaluation
Version: 0.3.0
Authors@R: c(
    person("Simon", "Couch", , "simon.couch@posit.co", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0001-5676-5107")),
    person("Max", "Kuhn", , "max@posit.co", role = "ctb"),
    person("Hadley", "Wickham", , "hadley@posit.co", role = "ctb",
           comment = c(ORCID = "0000-0003-4757-117X")),
    person("Mine", "Cetinkaya-Rundel", , "mine@posit.co", role = "ctb",
           comment = c(ORCID = "0000-0001-6452-2420")),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: A port of 'Inspect', a widely adopted 'Python' framework for
    large language model evaluation. Specifically aimed at 'ellmer' users
    who want to measure the effectiveness of their large language model-based
    products, the package supports prompt engineering, tool usage,
    multi-turn dialog, and model graded evaluations.
License: MIT + file LICENSE
URL: https://github.com/tidyverse/vitals, https://vitals.tidyverse.org
BugReports: https://github.com/tidyverse/vitals/issues
Depends: R (>= 4.1)
Imports: cli, dplyr, ellmer (>= 0.4.0), glue, httpuv, httr2, jsonlite,
        purrr, R6, rlang, S7, tibble, tidyr, withr
Suggests: ggplot2, here, htmltools, knitr, magick, ordinal, rmarkdown,
        testthat (>= 3.0.0), vcr (>= 2.0.0)
Config/Needs/website: tidyverse/tidytemplate, rmarkdown, btw,
        tidyverse, gt, brms, RcppEigen, broom, gt
Config/testthat/edition: 3
Config/testthat/parallel: true
Config/usethis/last-upkeep: 2025-04-25
Config/testthat/start-first: task, translate
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-05-15 15:30:21 UTC; simoncouch
Author: Simon Couch [aut, cre] (ORCID: <https://orcid.org/0000-0001-5676-5107>),
  Max Kuhn [ctb],
  Hadley Wickham [ctb] (ORCID: <https://orcid.org/0000-0003-4757-117X>),
  Mine Cetinkaya-Rundel [ctb] (ORCID:
    <https://orcid.org/0000-0001-6452-2420>),
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Simon Couch <simon.couch@posit.co>
Repository: CRAN
Date/Publication: 2026-05-15 20:20:02 UTC
Built: R 4.6.0; ; 2026-05-15 22:43:35 UTC; unix

Package: vitals
Title: Large Language Model Evaluation
Version: 0.3.0
Authors@R: c(
    person("Simon", "Couch", , "simon.couch@posit.co", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0001-5676-5107")),
    person("Max", "Kuhn", , "max@posit.co", role = "ctb"),
    person("Hadley", "Wickham", , "hadley@posit.co", role = "ctb",
           comment = c(ORCID = "0000-0003-4757-117X")),
    person("Mine", "Cetinkaya-Rundel", , "mine@posit.co", role = "ctb",
           comment = c(ORCID = "0000-0001-6452-2420")),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: A port of 'Inspect', a widely adopted 'Python' framework for
    large language model evaluation. Specifically aimed at 'ellmer' users
    who want to measure the effectiveness of their large language model-based
    products, the package supports prompt engineering, tool usage,
    multi-turn dialog, and model graded evaluations.
License: MIT + file LICENSE
URL: https://github.com/tidyverse/vitals, https://vitals.tidyverse.org
BugReports: https://github.com/tidyverse/vitals/issues
Depends: R (>= 4.1)
Imports: cli, dplyr, ellmer (>= 0.4.0), glue, httpuv, httr2, jsonlite,
        purrr, R6, rlang, S7, tibble, tidyr, withr
Suggests: ggplot2, here, htmltools, knitr, magick, ordinal, rmarkdown,
        testthat (>= 3.0.0), vcr (>= 2.0.0)
Config/Needs/website: tidyverse/tidytemplate, rmarkdown, btw,
        tidyverse, gt, brms, RcppEigen, broom, gt
Config/testthat/edition: 3
Config/testthat/parallel: true
Config/usethis/last-upkeep: 2025-04-25
Config/testthat/start-first: task, translate
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-05-15 15:30:21 UTC; simoncouch
Author: Simon Couch [aut, cre] (ORCID: <https://orcid.org/0000-0001-5676-5107>),
  Max Kuhn [ctb],
  Hadley Wickham [ctb] (ORCID: <https://orcid.org/0000-0003-4757-117X>),
  Mine Cetinkaya-Rundel [ctb] (ORCID:
    <https://orcid.org/0000-0001-6452-2420>),
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Simon Couch <simon.couch@posit.co>
Repository: CRAN
Date/Publication: 2026-05-15 20:20:02 UTC
Built: R 4.5.2; ; 2026-05-15 22:08:32 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,Task)
export(Task)
export(detect_answer)
export(detect_exact)
export(detect_includes)
export(detect_match)
export(detect_pattern)
export(generate)
export(generate_structured)
export(model_graded_fact)
export(model_graded_qa)
export(vitals_bind)
export(vitals_bundle)
export(vitals_log_dir)
export(vitals_log_dir_set)
export(vitals_view)
if (getRversion() < "4.3.0") importFrom("S7", "@")
import(rlang)
importFrom(R6,R6Class)
importFrom(cli,cat_line)
importFrom(cli,col_grey)
importFrom(cli,format_inline)
importFrom(ellmer,content_image_url)
importFrom(tibble,tibble)
importFrom(withr,defer)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# vitals <a href="https://vitals.tidyverse.org"><img src="man/figures/logo.png" align="right" height="240" alt="vitals website" /></a>

<!-- badges: start -->

[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![CRAN
status](https://www.r-pkg.org/badges/version/vitals)](https://CRAN.R-project.org/package=vitals)
[![R-CMD-check](https://github.com/tidyverse/vitals/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/tidyverse/vitals/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

vitals is a framework for large language model evaluation in R. It’s
specifically aimed at [ellmer](https://ellmer.tidyverse.org/) users who
want to measure the effectiveness of their LLM products like [custom
chat apps](https://posit.co/blog/custom-chat-app/) and
[querychat](https://github.com/posit-dev/querychat) apps. You can use it
to:

- Measure whether changes in your prompts or additions of new tools
  improve performance in your LLM product
- Compare how different models affect performance, cost, and/or latency
  of your LLM product
- Surface problematic behaviors in your LLM product

The package is an R port of the widely adopted Python framework
[Inspect](https://inspect.ai-safety-institute.org.uk/). While the
package doesn’t integrate with Inspect directly, it allows users to
interface with the [Inspect log
viewer](https://inspect.ai-safety-institute.org.uk/log-viewer.html) and
provides an on-ramp to transition to Inspect if need be by writing
evaluation logs to the same file format.

## Installation

Install the vitals package from CRAN with:

``` r
install.packages("vitals")
```

You can install the developmental version of vitals using:

``` r
pak::pak("tidyverse/vitals")
```

## Example

LLM evaluation with vitals is composed of two main steps.

``` r
library(vitals)
library(ellmer)
library(tibble)
```

1)  First, create an evaluation **task** with the `Task$new()` method.

``` r
simple_addition <- tibble(
  input = c("What's 2+2?", "What's 2+3?", "What's 2+4?"),
  target = c("4", "5", "6")
)

tsk <- Task$new(
  dataset = simple_addition, 
  solver = generate(chat_claude(model = "claude-sonnet-4-5")), 
  scorer = model_graded_qa()
)
```

Tasks are composed of three main components:

- **Datasets** are a data frame with, minimally, columns `input` and
  `target`. `input` represents some question or problem, and `target`
  gives the target response.
- **Solvers** are functions that take `input` and return some value
  approximating `target`, likely wrapping ellmer chats. `generate()` is
  the simplest scorer in vitals, and just passes the `input` to the
  chat’s `$chat()` method, returning its result as-is.
- **Scorers** juxtapose the solvers’ output with `target`, evaluating
  how well the solver solved the `input`.

2)  Evaluate the task.

``` r
tsk$eval()
```

`$eval()` will run the solver, run the scorer, and then situate the
results in a persistent log file that can be explored interactively with
the Inspect log viewer.

<img src="man/figures/log_viewer.png" alt="A screenshot of the Inspect log viewer, an interactive app displaying information on the 3 samples evaluated in this eval." width="100%" />

Any arguments to the solver or scorer can be passed to `$eval()`,
allowing for straightforward parameterization of tasks. For example, if
I wanted to evaluate `chat_openai()` on this task rather than
`chat_claude()`, I could write:

``` r
tsk_openai <- tsk$clone()
tsk_openai$eval(solver_chat = chat_openai(model = "gpt-4.1"))
```

For an applied example, see the “Getting started with vitals” vignette
at `vignette("vitals", package = "vitals")`.

## Package system files

**test/**

The package uses a number of cached objects during testing in `test/`.

The `.json` files in `inspect-example/logs` are the resulting log file from running evals in `inst/test/inspect`, e.g. with: 

```bash
inspect eval inst/test/inspect/basics.py  --model anthropic/claude-sonnet-4-5 --log-format=json
```

...or:

```bash
inspect eval test/inspect/tools.py  --model anthropic/claude-sonnet-4-5 --log-format=json
```

**dist/**

`/dist` is a bundled version of the Inspect viewer. (See [here](https://github.com/UKGovernmentBEIS/inspect_ai/blob/88d1cd98041a245c1d0cca4536d60e3244630b78/src/inspect_ai/_view/www/README.md) for more information.)

**regenerate-example-objects.R**

The package defines a function `regenerate_example_objects()` in the source that sources the script `inst/regenerate-example-objects.R`.

create_vitals_venv <- function() {
  env_name <- "vitals-venv"
  reticulate::virtualenv_create(env_name)

  reticulate::virtualenv_install(
    env_name,
    packages = c("inspect-ai", "pydantic")
  )

  reticulate::use_virtualenv(env_name)
}

reprex::reprex({
  library(vitals)
  library(ellmer)
  library(tidyverse)

  type_answer <- type_object(
    answer = type_string(
      "The author's first name, with no other formatting."
    )
  )

  names <- tribble(
    ~input                                 , ~target  ,
    "Name's Josiah, how's it going?"       , "Josiah" ,
    "I'm Lin, what's your name?"           , "Lin"    ,
    "My name is Em Fields, how about you?" , "Em"
  )

  tsk <-
    Task$new(
      dataset = names,
      solver = generate_structured(
        solver_chat = chat_anthropic(model = "claude-sonnet-4-20250514"),
        type = type_answer
      ),
      scorer = detect_match("any")
    )

  tsk$eval()

  tsk$get_samples()$solver_chat[[1]]
})

library(tidyverse)
library(ellmer)
library(vitals)
devtools::load_all()
plt <- tibble(
  input = "Please make a ggplot of mpg vs hp in mtcars and tell me what you see.",
  target = "Something like `ggplot(mtcars) + aes(x = hp, y = mpg)"
)

tsk <- Task$new(
  dataset = plt,
  solver = generate(
    solver_chat = chat_claude()$register_tool(predictive:::tool_run_r_code)
  ),
  scorer = model_graded_qa(scorer_chat = chat_claude())
)

tsk$eval()

# ------------------------------------------------------------------

# ch <- chat_claude()$register_tool(predictive:::tool_run_r_code)
# ch$chat(
#   "Please make a ggplot of mpg vs hp in mtcars and tell me what you see.",
#   echo = FALSE
# )

# Runs the code to regenerate example eval objects used in the package.
regenerate_example_objects <- function() {
  devtools::load_all()

  cli::cli_progress_step("Regenerating example task")
  regenerate_example_task()

  cli::cli_progress_step("Regenerating `are` on 1 epoch")
  regenerate_are_1e()

  cli::cli_progress_step("Regenerating `are` with custom solvers")
  regenerate_are_custom_solvers()

  cli::cli_progress_step(
    "Regenerating `are` with 3 epochs for analysis article"
  )
  regenerate_are_3e_analysis()

  cli::cli_progress_done()
}

# example_task -- a minimal, 2-row toy eval ---------------------------
regenerate_example_task <- function() {
  library(ellmer)
  library(tibble)

  simple_addition <- tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  ch <- chat_claude(model = "claude-sonnet-4-5-20250929")
  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(ch$set_system_prompt("Respond in the format 'a+b=c'.")),
    scorer = model_graded_qa(scorer_chat = ch)
  )

  tsk$eval()
  save(tsk, file = "inst/test/example-task.rda")
}

# An R Eval on 1 epoch, via the intro vignette ------------------------
regenerate_are_1e <- function() {
  json_files <- list.files(
    "vignettes/articles/data/vitals/logs",
    pattern = "\\.json$",
    full.names = TRUE
  )
  if (length(json_files) > 0) {
    file.remove(json_files)
  }

  withr::local_envvar(VITALS_SHOULD_EVAL = "true")
  rmarkdown::render('vignettes/vitals.Rmd')
}

# An R Eval on 1 epoch, via the intro vignette ------------------------
regenerate_are_custom_solvers <- function() {
  json_files <- list.files(
    "vignettes/articles/data/solvers/logs",
    pattern = "\\.json$",
    full.names = TRUE
  )
  if (length(json_files) > 0) {
    file.remove(json_files)
  }

  withr::local_envvar(VITALS_SHOULD_EVAL = "true")
  rmarkdown::render('vignettes/articles/solvers.Rmd')
}

# An R Eval on 3 epoch x 3 models, via the analysis article -----------------
regenerate_are_3e_analysis <- function() {
  json_files <- list.files(
    "vignettes/articles/data/analysis/logs",
    pattern = "\\.json$",
    full.names = TRUE
  )
  if (length(json_files) > 0) {
    file.remove(json_files)
  }

  withr::local_envvar(VITALS_SHOULD_EVAL = "true")
  rmarkdown::render("vignettes/articles/analysis.Rmd")
}

regenerate_example_objects()

eval_log <- function(
  eval = translate_to_eval(),
  plan = translate_to_plan(),
  results = translate_to_results(),
  stats = translate_to_stats(),
  samples
) {
  res <-
    list(
      version = 2L,
      status = "success",
      eval = eval,
      plan = plan,
      results = results,
      stats = stats,
      samples = samples
    )

  structure(res, class = c("eval_log", class(res)))
}

# top-level entries ------------------------------------------------------------
translate_to_eval <- function(
  run_id,
  created = eval_log_timestamp(),
  task,
  task_id,
  task_version = 0,
  task_file = "",
  task_attribs = c(),
  task_args = c(),
  dataset,
  model,
  model_args = c(),
  config = c(),
  # TODO: look into what this actually does
  revision = list(
    type = "git",
    origin = "https://github.com/UKGovernmentBEIS/inspect_ai.git",
    commit = "9140d8a2"
  ),
  packages = list(inspect_ai = "0.3.63"),
  scorers
) {
  list(
    run_id = run_id,
    created = created,
    task = task,
    task_id = task_id,
    task_version = task_version,
    task_file = task_file,
    task_attribs = task_attribs,
    task_args = task_args,
    dataset = dataset,
    model = model,
    model_args = model_args,
    config = config,
    revision = revision,
    packages = packages,
    scorers = scorers
  )
}

translate_to_plan <- function(
  name = "plan",
  steps = translate_to_plan_steps(),
  config = c()
) {
  list(
    name = name,
    steps = steps,
    config = config
  )
}

translate_to_results <- function(
  total_samples,
  completed_samples,
  scores
) {
  list(
    total_samples = total_samples,
    completed_samples = completed_samples,
    scores = scores
  )
}

translate_to_plan_steps <- function(name, arguments, system_prompt = NULL) {
  steps <- list()

  if (!is.null(system_prompt)) {
    steps <- append(
      steps,
      list(list(
        solver = "system_message",
        params = list(template = system_prompt)
      ))
    )
  }

  steps <- append(
    steps,
    list(list(
      solver = name,
      params = arguments
    ))
  )

  steps
}

translate_to_stats <- function(
  started_at,
  completed_at,
  model_usage
) {
  list(
    started_at = started_at,
    completed_at = completed_at,
    model_usage = model_usage
  )
}

translate_to_samples <- function(dataset, scores, timestamps) {
  res <- list()
  for (i in seq_len(nrow(dataset))) {
    sample <- dataset[i, , drop = FALSE]
    res[[i]] <- translate_to_sample(
      sample,
      scores = scores,
      timestamps = timestamps
    )
  }
  res
}

translate_to_sample <- function(sample, scores, timestamps) {
  chat <- sample$solver_chat[[1]]
  scorer_name <- scores$name
  if ("scorer_chat" %in% names(sample)) {
    scorer_name <- paste0(
      c(
        scorer_name,
        " (",
        sample$scorer_chat[[1]]$get_model(),
        ")"
      ),
      collapse = ""
    )
  }
  turns <- chat$get_turns()
  first_turn <- turns[[1]]
  input_value <- message_content_from_turn(first_turn)
  sample_input <- if (is_content_list(input_value)) {
    list(list(
      content = input_value,
      source = "input",
      role = "user"
    ))
  } else {
    input_value
  }
  list(
    id = sample$id,
    epoch = if ("epoch" %in% colnames(sample)) {
      sample$epoch
    } else {
      1
    },
    input = sample_input,
    target = as.character(sample$target),
    messages = translate_to_messages(chat),
    output = translate_to_output(chat),
    scores = dots_list(
      !!scorer_name := translate_to_score(
        output = sample$result[[1]],
        score = sample$score[[1]],
        scorer = scorer_name,
        scorer_chat = if ("scorer_chat" %in% names(sample)) {
          sample$scorer_chat[[1]]
        } else {
          NULL
        },
        scorer_explanation = if ("scorer_explanation" %in% names(sample)) {
          sample$scorer_explanation[[1]]
        } else {
          NULL
        },
        metadata = as_metadata(sample[names(sample) == "scorer_metadata"])
      )
    ),
    metadata = as_metadata(sample[names(sample) == "solver_metadata"]),
    store = c(),
    events = translate_to_events(sample = sample, timestamps = timestamps),
    model_usage = sum_model_usage(list(chat)),
    attachments = c()
  )
}

# sub-level entries ------------------------------------------------------------
translate_to_metrics <- function(
  name = character(),
  value = numeric(),
  options = list()
) {
  list(
    name = name,
    value = value,
    options = options
  )
}

translate_to_score <- function(
  output,
  score,
  scorer,
  scorer_chat = NULL,
  scorer_explanation = NULL,
  metadata
) {
  if (is.null(scorer_chat)) {
    explanation <- if (!is.null(scorer_explanation)) {
      scorer_explanation
    } else {
      scorer
    }
    return(list(
      value = score,
      answer = as.character(output),
      explanation = explanation,
      metadata = metadata
    ))
  }

  turns <- scorer_chat$get_turns()
  explanation <- scorer_chat$last_turn()@text

  list(
    value = score,
    answer = as.character(output),
    explanation = explanation,
    metadata = c(
      list(
        grading = lapply(turns, translate_to_metadata_grading)
      ),
      metadata
    )
  )
}

# Inspect formats the content a bit differently depending
# on whether the turn role is user vs. assistant.
translate_to_metadata_grading <- function(turn) {
  if (turn@role == "user") {
    return(
      list(
        id = generate_id(),
        content = turn@text,
        role = "user"
      )
    )
  }

  list(
    id = generate_id(),
    content = list(list(type = "text", text = turn@text)),
    source = "generate",
    role = turn@role
  )
}

translate_to_eval_scorers <- function(name) {
  list(list(
    name = name,
    options = c(),
    # TODO: make this dynamic once implemented
    metrics = list(
      list(name = "mean", options = c())
    ),
    metadata = c()
  ))
}

# format single-row tibble as a string with column names as prefixes
# followed by a colon, with entries separated by newlines.
input_string <- function(x) {
  if (is.character(x)) {
    return(x)
  }
  paste0(
    mapply(
      function(name, value) paste0(name, ": ", value),
      names(x),
      as.list(x[1, ])
    ),
    collapse = "\n\n"
  )
}

# metadata entries must ultimately be a "flattened", named list.
# note `c()` is interpreted as a named list by jsonlite.
# we're not interested in recursing here; provide a lossy summary of the object
as_metadata <- function(x) {
  result <- if (is.list(x) && !has_class(x)) {
    x
  } else if (is.data.frame(x) || (is.atomic(x) && !is.null(names(x)))) {
    as.list(x)
  } else {
    list(capture.output(print(x)))
  }

  result <- set_reasonable_names(result)

  lapply(result, function(el) {
    if (!is.list(el)) {
      el
    } else if (length(el) == 1 && has_class(el[[1]])) {
      capture.output(print(el[[1]]))
    } else {
      capture.output(if (has_class(el)) print(el) else str(el))
    }
  })
}

has_class <- function(x) {
  !is.null(class(x)) && !identical(class(x), "list")
}

set_reasonable_names <- function(x) {
  if (is.null(names(x))) {
    names(x) <- as.character(seq_along(x))
    return(x)
  }

  unnamed_idx <- which(names(x) == "")
  names(x)[unnamed_idx] <- as.character(unnamed_idx)
  x
}

#' An R Eval
#'
#' @description
#' An R Eval is a dataset of challenging R coding problems. Each `input` is a
#' question about R code which could be solved on first-read only by experts
#' and, with a chance to read documentation and run some code, by
#' fluent data scientists. Solutions are in `target()` and enable a fluent
#' data scientist to evaluate whether the solution deserves full, partial, or
#' no credit.
#'
#' Pass this dataset to `Task$new()` to situate it inside of an evaluation
#' task.
#'
#' @format A tibble with `r nrow(are)` rows and `r ncol(are)` columns:
#' \describe{
#'   \item{id}{Character. Unique identifier/title for the code problem.}
#'   \item{input}{Character. The question to be answered.}
#'   \item{target}{Character. The solution, often with a description of notable
#'   features of a correct solution.}
#'   \item{domain}{Character. The technical domain
#'   (e.g., Data Analysis, Programming, or Authoring).}
#'   \item{task}{Character. Type of task
#'   (e.g., Debugging, New feature, or Translation.)}
#'   \item{source}{Character. URL or source of the problem. `NA`s indicate that
#'   the problem was written originally for this eval.}
#'   \item{knowledge}{List. Required knowledge/concepts for solving the problem.}
#' }
#'
#' @source Posit Community, GitHub issues, R4DS solutions, etc. For row-level
#' references, see `source`.
#'
#' @examples
#' are
#'
#' dplyr::glimpse(are)
"are"

#' Interactively view local evaluation logs
#'
#' @description
#' vitals bundles the Inspect log viewer, an interactive app for exploring
#' evaluation logs. Supply a path to a directory of tasks written to json.
#' For individual [Task] objects, use the `$view()` method instead.
#'
#' @param dir Path to a directory containing task eval logs.
#' @param host Host to serve on. Defaults to "127.0.0.1".
#' @param port Port to serve on. If NULL, will find a random available port.
#'
#' @inherit Task examples
#' @return The server object (invisibly)
#' @export
vitals_view <- function(
  dir = vitals_log_dir(),
  host = "127.0.0.1",
  port = NULL
) {
  vitals_view_impl(dir = dir, host = host, port = port)
}

vitals_view_impl <- function(
  dir = vitals_log_dir(),
  host = "127.0.0.1",
  port = NULL,
  call = caller_env()
) {
  dist_dir <- system.file("dist", package = "vitals")

  if (is.null(port)) {
    port <- httpuv::randomPort()
  }

  tryCatch(
    {
      existing_server <- httpuv::listServers()
      if (length(existing_server) > 0) {
        httpuv::stopServer(existing_server[[1]])
      }
    },
    error = function(cnd) {
      cli::cli_abort(
        "Unable to terminate the existing server.",
        parent = cnd,
        call = call
      )
    }
  )

  if (!dir.exists(dir)) {
    cli::cli_abort(
      "Log directory {.file {dir}} not found.",
      call = call
    )
  }

  log_metadata <- get_log_files_metadata(dir)
  task_ids <- vapply(log_metadata, function(x) x$task_id %||% NA_character_, character(1))
  if (any(vapply(task_ids, is_old_task_id_format, logical(1)), na.rm = TRUE)) {
    warn_old_task_id_format()
  }

  server <- httpuv::startServer(
    host = host,
    port = port,

    app = list(
      call = function(req) {
        tryCatch(
          {
            # parse query parameters
            query <- parse_query_string(req$QUERY_STRING)

            # handle API routes first
            if (startsWith(req$PATH_INFO, "/api/")) {
              # handle dynamic route first
              if (startsWith(req$PATH_INFO, "/api/logs/")) {
                file <- substr(req$PATH_INFO, 11, nchar(req$PATH_INFO))
                file <- utils::URLdecode(file)
                return(get_api_log_file(dir, file, query))
              }

              # handle static routes
              return(switch(
                req$PATH_INFO,
                "/api/logs" = get_api_logs(dir),
                "/api/log-dir" = get_api_log_dir(dir),
                "/api/log-files" = get_api_log_files(dir),
                "/api/log-headers" = get_api_log_headers(dir, query),
                "/api/events" = get_api_events(),
                "/api/eval-set" = get_api_eval_set(),
                "/api/flow" = get_api_flow(),
                list(status = 404, body = "API endpoint not found")
              ))
            }

            # then handle static files
            if (req$PATH_INFO == "/" || !grepl("\\.", req$PATH_INFO)) {
              # serve index.html for root or routes without file extensions
              index_path <- file.path(dist_dir, "index.html")
              return(list(
                status = 200,
                headers = list(
                  'Content-Type' = 'text/html',
                  'Cache-Control' = 'no-cache'
                ),
                body = readBin(index_path, "raw", file.info(index_path)$size)
              ))
            } else {
              # serve static files from dist
              file_path <- file.path(dist_dir, substring(req$PATH_INFO, 2))
              if (file.exists(file_path)) {
                content_type <- switch(
                  tools::file_ext(file_path),
                  "html" = "text/html",
                  "js" = "application/javascript",
                  "css" = "text/css",
                  "svg" = "image/svg+xml",
                  "application/octet-stream"
                )
                return(list(
                  status = 200,
                  headers = list(
                    'Content-Type' = content_type,
                    'Cache-Control' = 'no-cache'
                  ),
                  body = readBin(file_path, "raw", file.info(file_path)$size)
                ))
              }
            }

            list(status = 404, body = "Not found")
          },
          error = function(e) {
            # log the error for debugging
            message("Error processing request: ", e$message)
            list(
              status = 500,
              headers = list('Content-Type' = 'text/plain'),
              body = paste("Error:", e$message)
            )
          }
        )
      }
    )
  )

  url <- sprintf("http://%s:%d", host, port)
  cli::cli_inform(
    c("v" = "Inspect Viewer running at: {.url {url}}"),
    class = "vitals_viewer_start"
  )

  if (interactive() && !is_testing()) {
    utils::browseURL(url)
  }

  invisible(server)
}

parse_query_string <- function(query_string) {
  if (is.null(query_string) || query_string == "") {
    return(list())
  }

  # Remove leading ? if present
  query_string <- sub("^\\?", "", query_string)

  # Replace any ?file= with &file= (except at the start)
  query_string <- gsub("\\?file=", "&file=", query_string)

  parts <- strsplit(query_string, "&")[[1]]
  params <- lapply(parts, function(p) {
    kv <- strsplit(p, "=")[[1]]
    if (length(kv) == 2) {
      val <- utils::URLdecode(kv[2])
      res <- list(val)
      names(res) <- kv[1]
      return(res)
    }
    NULL
  })
  do.call(c, params)
}

path_to_file_url <- function(path) {
  normalized <- normalizePath(path, winslash = "/", mustWork = FALSE)
  if (.Platform$OS.type == "windows") {
    paste0("file:///", normalized)
  } else {
    paste0("file://", normalized)
  }
}

file_url_to_path <- function(url) {
  if (!startsWith(url, "file://")) {
    return(url)
  }
  path <- substr(url, 8, nchar(url))
  if (.Platform$OS.type == "windows" && startsWith(path, "/")) {
    path <- substr(path, 2, nchar(path))
  }
  path
}

get_log_files_metadata <- function(dir) {
  files <- list.files(dir, pattern = "\\.json$", recursive = TRUE)
  files <- files[basename(files) != "listing.json"]
  files <- sort(files, decreasing = TRUE)

  lapply(files, function(f) {
    file_path <- file.path(dir, f)
    info <- file.info(file_path)

    task_info <- tryCatch(
      {
        headers <- eval_log_read_headers(file_path)
        list(
          task = headers$eval$task,
          task_id = headers$eval$task_id
        )
      },
      error = function(e) {
        list(task = NULL, task_id = NULL)
      }
    )

    list(
      name = path_to_file_url(file_path),
      size = info$size,
      mtime = as.numeric(info$mtime) * 1000,
      task = task_info$task,
      task_id = task_info$task_id
    )
  })
}

json_response <- function(data, status = 200) {
  list(
    status = status,
    headers = list(
      'Content-Type' = 'application/json',
      'Cache-Control' = 'no-cache'
    ),
    body = jsonlite::toJSON(
      data,
      auto_unbox = TRUE,
      null = "null"
    )
  )
}

get_api_logs <- function(dir) {
  resp <- list(
    log_dir = path_to_file_url(dir),
    files = get_log_files_metadata(dir)
  )
  json_response(resp)
}

get_api_log_dir <- function(dir) {
  resp <- list(
    log_dir = path_to_file_url(dir)
  )
  json_response(resp)
}

get_api_log_files <- function(dir) {
  resp <- list(
    response_type = "full",
    files = get_log_files_metadata(dir)
  )
  json_response(resp)
}

get_api_log_headers <- function(dir, query) {
  if (!is.null(query$file)) {
    files <- as.list(query)

    headers <- lapply(files, function(f) {
      if (startsWith(f, "file://")) {
        file_path <- file_url_to_path(f)
      } else if (startsWith(f, "/") || grepl("^[A-Za-z]:", f)) {
        file_path <- f
      } else {
        file_path <- file.path(dir, f)
      }

      if (file.exists(file_path)) {
        content <- eval_log_read_headers(file_path)
      } else {
        NULL
      }
    })

    names(headers) <- NULL
    return(json_response(headers))
  }

  json_response(list())
}

get_api_events <- function() {
  json_response(list())
}

get_api_eval_set <- function() {
  json_response(NULL)
}

get_api_flow <- function() {
  json_response(NULL)
}

get_api_log_file <- function(dir, file, query) {
  if (startsWith(file, "file://")) {
    file_path <- file_url_to_path(file)
  } else if (startsWith(file, "/") || grepl("^[A-Za-z]:", file)) {
    file_path <- file
  } else {
    file_path <- file.path(dir, file)
  }

  if (!file.exists(file_path)) {
    return(list(status = 404, body = "File not found"))
  }

  content <- jsonlite::fromJSON(file_path)

  header_only <- query$`header-only`
  if (!is.null(header_only)) {
    header_only <- as.numeric(header_only)
    file_size_mb <- file.info(file_path)$size / (1024 * 1024)

    if (!is.na(header_only) && file_size_mb > header_only) {
      content$samples <- NULL
    }
  }

  json_response(content)
}

test_that("vitals_view() works with .json from Python Inspect", {
  log_dir <- system.file("test/inspect/logs", package = "vitals")
  if (identical(log_dir, "")) {
    skip("Test log files not available")
  }
  expect_condition(vitals_view(log_dir), class = "vitals_viewer_start")
})

test_that("vitals_view() restarts with existing server", {
  log_dir <- system.file("test/inspect/logs", package = "vitals")
  if (identical(log_dir, "")) {
    skip("Test log files not available")
  }
  suppressMessages(vitals_view(log_dir))
  expect_condition(vitals_view(log_dir), class = "vitals_viewer_start")
})

test_that("vitals_view() errors informatively on failure to restart", {
  log_dir <- system.file("test/inspect/logs", package = "vitals")
  if (identical(log_dir, "")) {
    skip("Test log files not available")
  }
  suppressMessages(vitals_view(log_dir))

  testthat::local_mocked_bindings(
    stopServer = function(...) {
      rlang::abort("no way!")
    },
    .package = "httpuv"
  )
  expect_snapshot(vitals_view(log_dir), error = TRUE)
})

test_that("vitals_view() handles nonexistent log directory", {
  log_dir <- "/path/that/does/not/exist"
  expect_snapshot(vitals_view(log_dir), error = TRUE)
})

test_that("parse_query_string() works", {
  expect_equal(parse_query_string(NULL), list())
  expect_equal(parse_query_string(""), list())

  expect_equal(parse_query_string("?key=value"), list(key = "value"))
  expect_equal(parse_query_string("key=value"), list(key = "value"))
  expect_equal(
    parse_query_string("key1=value1&key2=value2"),
    list(key1 = "value1", key2 = "value2")
  )

  # Replaces ?file= with &file= except at start
  result <- parse_query_string("file=first?file=second")
  expect_equal(names(result), c("file", "file"))
  expect_equal(result[[1]], "first")
  expect_equal(result[[2]], "second")

  # URL decodes values
  expect_equal(
    parse_query_string("key=hello%20world"),
    list(key = "hello world")
  )

  # handles parameters without values
  expect_equal(
    parse_query_string("key1&key2=value2"),
    list(key2 = "value2")
  )

  # handles multiple values for same parameter
  result <- parse_query_string("file=first&file=second&file=third")
  expect_equal(names(result), c("file", "file", "file"))
  expect_equal(result[[1]], "first")
  expect_equal(result[[2]], "second")
  expect_equal(result[[3]], "third")
})

# nocov start
.onLoad <- function(libname, pkgname) {
  # run_on_load()
  S7::methods_register()
}

# enable usage of <S7_object>@name in package code
#' @rawNamespace if (getRversion() < "4.3.0") importFrom("S7", "@")
NULL
# nocov end

# Log a function's evaluation
#
# @description
# Inspect has a logging mechanism s.t. the names of functions called as well
# as the arguments supplied are stored in the eval log.
#
# `logged()` takes a function and maps it to a function with this sort of
# "logged" return value. The function is applied to both internal functions
# and user-supplied solvers/scorers/metrics.
#
# @returns
# A modified version of `fn()` with signature `...` and a named list return
# value. That list's return value is:
#
# * `name`: A string representing the name of the function called.
# * `value`: The original return value of `fn()`.
# * `arguments`: Expressions representing arguments passed to `fn()`,
# captured with [rlang::enexprs()].
#
# @examples
# logged(mean)(x = 1:3)
# @keywords internal
logged <- function(fn, fn_name = deparse(substitute(fn))) {
  # TODO: `deparse_substitute()` may not be the intended value? may want a
  # call name instead. we'll see.
  function(...) {
    arg_exprs <- enexprs(...)
    arg_exprs <- map(arg_exprs, expr_deparse)

    res <- list(
      name = paste0(fn_name, collapse = ""),
      value = fn(...),
      arguments = arg_exprs
    )

    res
  }
}

#' Creating and evaluating tasks
#'
#' @description
#' Evaluation `Task`s provide a flexible data structure for evaluating LLM-based
#' tools.
#'
#' 1) **Datasets** contain a set of labelled samples. Datasets are just a
#' tibble with columns `input` and `target`, where `input` is a prompt
#' and `target` is either literal value(s) or grading guidance.
#' 2) **Solvers** evaluate the `input` in the dataset and produce a final result.
#' 3) **Scorers** evaluate the final output of solvers. They may use text
#' comparisons (like [detect_match()]), model grading (like
#' [model_graded_qa()]), or other custom schemes.
#'
#' **The usual flow of LLM evaluation with Tasks calls `$new()` and then `$eval()`.**
#' `$eval()` just calls `$solve()`, `$score()`, `$measure()`, `$log()`,
#' and `$view()` in order. The remaining methods are generally only
#' recommended for expert use.
#'
#' @param solver A function that takes a vector of inputs from the
#' dataset's `input` column as its first argument and determines values
#' approximating `dataset$target`. Its return value must be a list with
#' the following elements:
#'
#' * `result` - A character vector of the final responses, with the same length
#'   as `dataset$input`.
#' * `solver_chat` - A list of ellmer Chat objects that were used to solve
#'   each input, also with the same length as `dataset$input`.
#'
#' Additional output elements can be included in a slot `solver_metadata` that
#' has the same length as `dataset$input`, which will be logged in
#' `solver_metadata`.
#'
#' Additional arguments can be passed to the solver via `$solve(...)`
#' or `$eval(...)`. See the definition of [generate()] for a function that
#' outputs a valid solver that just passes inputs to ellmer Chat objects'
#' `$chat()` method in parallel.
#'
#' @param scorer A function that evaluates how well the solver's return value
#' approximates the corresponding elements of `dataset$target`. The function
#' should take in the `$get_samples()` slot of a Task object and return a list with
#' the following elements:
#'
#' * `score` - A vector of scores with length equal to `nrow(samples)`.
#'   Built-in scorers return ordered factors with
#'   levels `I` < `P` (optionally) < `C` (standing for "Incorrect", "Partially
#'   Correct", and "Correct"). If your scorer returns this output type, the
#'   package will automatically calculate metrics.
#'
#' Optionally:
#' * `scorer_chat` - If your scorer makes use of ellmer, also include a list of
#'   ellmer Chat objects that were used to score each result, also with
#'   length `nrow(samples)`.
#' * `scorer_metadata` - Any intermediate results or other values that you'd
#'   like to be stored in the persistent log. This should also have length
#'   equal to `nrow(samples)`.
#'
#' Scorers will probably make use of `samples$input`, `samples$target`, and
#' `samples$result` specifically. See [model-based scoring][scorer_model]
#' for examples.
#'
#' @param metrics A named list of functions that take in a vector of scores
#' (as in `task$get_samples()$score`) and output a single numeric value.
#'
#' @param epochs The number of times to repeat each sample. Evaluate each sample
#' multiple times to better quantify variation. Optional, defaults to `1L`.
#' The value of `epochs` supplied to `$eval()` or `$score()` will take
#' precedence over the value in `$new()`.
#'
#' @seealso [generate()] for the simplest possible solver, and
#' [scorer_model] and [scorer_detect] for two built-in approaches to
#' scoring.
#' @examples
#' if (!identical(Sys.getenv("ANTHROPIC_API_KEY"), "")) {
#'   # set the log directory to a temporary directory
#'   withr::local_envvar(VITALS_LOG_DIR = withr::local_tempdir())
#'
#'   library(ellmer)
#'   library(tibble)
#'
#'   simple_addition <- tibble(
#'     input = c("What's 2+2?", "What's 2+3?"),
#'     target = c("4", "5")
#'   )
#'
#'   # create a new Task
#'   tsk <- Task$new(
#'     dataset = simple_addition,
#'     solver = generate(chat_claude(model = "claude-sonnet-4-5-20250929")),
#'     scorer = model_graded_qa()
#'   )
#'
#'   # evaluate the task (runs solver and scorer) and opens
#'   # the results in the Inspect log viewer (if interactive)
#'   tsk$eval()
#'
#'   # $eval() is shorthand for:
#'   tsk$solve()
#'   tsk$score()
#'   tsk$measure()
#'   tsk$log()
#'   tsk$view()
#'
#'   # get the evaluation results as a data frame
#'   tsk$get_samples()
#'
#'   # view the task directory with $view() or vitals_view()
#'   vitals_view()
#' }
#'
#' # The `input` column can be a list of 1-row tibbles for per-sample metadata.
#' # Custom solvers can then extract columns from each input:
#' shapes_data <- tibble::tibble(
#'   input = list(
#'     tibble::tibble(shapes = "square, circle, rhombus", pick = "square"),
#'     tibble::tibble(shapes = "square, circle, rhombus", pick = "circle")
#'   ),
#'   target = c("square", "circle")
#' )
#'
#' my_solver <- function(solver_chat = NULL) {
#'   chat <- solver_chat
#'   function(inputs, ..., solver_chat = chat) {
#'     ch <- if (is.function(solver_chat)) solver_chat() else solver_chat$clone()
#'     prompts <- lapply(inputs, function(inp) {
#'       paste0("Always pick ", inp$pick, ". Return only that shape.\n\n", inp$shapes)
#'     })
#'     res <- ellmer::parallel_chat(ch, prompts, ...)
#'     list(result = purrr::map_chr(res, \(c) c$last_turn()@text), solver_chat = res)
#'   }
#' }
#'
#' @export
Task <- R6::R6Class(
  "Task",
  lock_objects = FALSE,
  public = list(
    #' @field dir The directory where evaluation logs will be written to. Defaults
    #' to `vitals_log_dir()`.
    dir = vitals_log_dir(),

    #' @field metrics A named vector of metric values resulting from `$measure()`
    #' (called inside of `$eval()`). Will be `NULL` if metrics have yet to
    #' be applied.
    metrics = NULL,

    #' @description
    #' The typical flow of LLM evaluation with vitals tends to involve first
    #' calling this method and then `$eval()` on the resulting object.

    #' @param dataset A tibble with, minimally, columns `input` and `target`.
    #' The `input` column can be either a character vector or a list-column
    #' of 1-row tibbles. Using 1-row tibbles allows per-sample customization
    #' by including additional metadata that custom solvers can access.
    #' @param name A name for the evaluation task. Defaults to
    #' `deparse(substitute(dataset))`.
    #' @param dir Directory where logs should be stored.
    #'
    #' @return A new Task object.
    initialize = function(
      dataset,
      solver,
      scorer,
      metrics = NULL,
      epochs = NULL,
      name = deparse(substitute(dataset)),
      dir = vitals_log_dir()
    ) {
      force(name)

      solver_name <- deparse(substitute(solver))
      scorer_name <- deparse(substitute(scorer))
      metrics_name <- deparse(substitute(metrics))

      check_dataset(dataset)
      check_log_dir(dir)
      check_function(solver)
      # TODO: for non-built in scorers, what to do?
      check_metrics(metrics)
      check_number_whole(epochs, min = 1, allow_null = TRUE)

      # dataset names can contain dashes or alphanumerics--transition
      # underscores and spaces to dashes (#92)
      private$dataset_name <- gsub(
        "[^[:alnum:]\\-]",
        "",
        gsub("_| ", "-", name)
      )
      self$dir <- dir
      private$solver <- logged(solver, fn_name = solver_name)
      private$scorer <- logged(scorer, fn_name = scorer_name)

      if (!is.null(metrics)) {
        private$metric_fns <- metrics
      }

      private$task_id <- NULL
      private$epochs <- epochs

      private$samples <- set_id_column(dataset)
    },

    #' @description
    #' Evaluates the task by running the solver, scorer, logging results, and
    #' viewing (if interactive). This method works by calling `$solve()`,
    #' `$score()`, `$log()`, and `$view()` in sequence.
    #'
    #' The typical flow of LLM evaluation with vitals tends to involve first
    #' calling `$new()` and then this method on the resulting object.
    #'
    #' @param ... Additional arguments passed to the solver and scorer functions.
    #' All arguments must be named. Arguments are routed based on function
    #' signatures: if an argument name matches a parameter in the solver, it goes
    #' to the solver; if it matches a parameter in the scorer, it goes to the
    #' scorer. Arguments matching both go to both. Unmatched arguments are passed
    #' to any function with `...` in its signature. An error is raised if an
    #' argument matches neither function and neither accepts `...`.
    #' @param view Automatically open the viewer after evaluation (defaults to
    #' TRUE if interactive, FALSE otherwise).
    #'
    #' @return The Task object (invisibly)
    eval = function(..., epochs = NULL, view = interactive()) {
      check_number_whole(epochs, min = 1, allow_null = TRUE)

      if (private$solved || private$scored) {
        private$reset_for_new_eval()
      }

      dots <- list(...)
      if (
        length(dots) > 0 && (is.null(names(dots)) || any(names(dots) == ""))
      ) {
        cli::cli_abort(
          "All arguments in {.code ...} must be named.",
          call = call2("$eval()")
        )
      }

      split_args <- private$split_dots(dots)

      cli::cli_progress_step("Solving")
      do.call(self$solve, c(split_args$solver, list(epochs = epochs)))

      cli::cli_progress_step("Scoring")
      do.call(self$score, split_args$scorer)
      self$measure()

      self$log(self$dir)
      private$stash_last_task()

      cli::cli_progress_done()
      if (view) {
        self$view()
      }

      invisible(self)
    },

    #' @description
    #' The task's samples represent the evaluation in a data frame format.
    #'
    #' [vitals_bind()] row-binds the output of this
    #' function called across several tasks.
    #'
    #' @return
    #' A tibble representing the evaluation. Based on the `dataset`,
    #' `epochs` may duplicate rows, and the solver and scorer will append
    #' columns to this data.
    get_samples = function() {
      private$samples
    },

    #' @description
    #' Solve the task by running the solver
    #'
    #' @param ... Additional arguments passed to the solver function.
    #'
    #' @return The Task object (invisibly)
    solve = function(..., epochs = NULL) {
      withr::local_envvar(IN_VITALS_EVAL = "true")
      check_number_whole(epochs, min = 1, allow_null = TRUE)

      if (private$solved) {
        private$reset_for_new_eval()
      }

      private$timestamps$solve$started_at <- Sys.time()

      private$samples <- join_epochs(
        self$get_samples(),
        epochs %||% private$epochs
      )

      private$run_id <- generate_id()

      private$samples$result <- NA
      private$samples$solver_chat <- NA

      private$track_token_usage("solver_token_usage")
      private$solutions <- private$solver(self$get_samples()$input, ...)

      # TODO: it might be nice to just run one of the inputs async and check for
      # this earlier on so that a full eval's worth of results isn't thrown
      # away if the output format isn't quite right.
      private$check_solver_outputs()
      private$cbind_solutions()

      private$solved <- TRUE
      private$timestamps$solve$completed_at <- Sys.time()
      invisible(self)
    },

    #' @description
    #' Score the task by running the scorer and then applying metrics to
    #' its results.
    #'
    #' @param ... Additional arguments passed to the scorer function.
    #'
    #' @return The Task object (invisibly)
    score = function(...) {
      withr::local_envvar(IN_VITALS_EVAL = "true")
      if (!private$solved) {
        cli::cli_alert_warning(
          "Task has not been solved yet. Run task$solve() first."
        )
        return(invisible(self))
      }

      private$timestamps$score$started_at <- Sys.time()
      private$samples$score <- NA

      private$track_token_usage("scorer_token_usage")
      private$scores <- private$scorer(self$get_samples(), ...)
      private$check_scorer_outputs()
      private$cbind_scores()

      private$scored <- TRUE
      private$timestamps$score$completed_at <- Sys.time()
      invisible(self)
    },

    #' @description
    #' Applies metrics to a scored Task.
    #'
    #' @return The Task object (invisibly)
    measure = function() {
      if (!private$scored) {
        cli::cli_abort(
          "Task has not been scored yet. Run task$score() first."
        )
      }

      if (!is.null(private$metric_fns)) {
        private$apply_metrics()
      } else {
        private$apply_naive_accuracy()
      }

      private$check_metric_output()
      self$metrics <- purrr::map_dbl(
        private$metric_results,
        purrr::pluck,
        "value"
      )

      invisible(self)
    },

    #' @description
    #' Log the task to a directory.
    #'
    #' Note that, if an `VITALS_LOG_DIR` envvar is set, this will happen
    #' automatically in `$eval()`.
    #'
    #' @param dir The directory to write the log to.
    #'
    #' @return The path to the logged file, invisibly.
    log = function(dir = self$dir) {
      if (!private$scored) {
        cli::cli_abort(
          "Task has not been scored yet. Run task$score() first."
        )
      }

      samples <- self$get_samples()
      samples <- private$add_working_times(samples)
      samples <- private$add_working_starts(samples)

      scorer_name <- private$scores$name
      if ("scorer_chat" %in% names(samples)) {
        scorer_name <- paste0(
          c(
            scorer_name,
            " (",
            samples$scorer_chat[[1]]$get_model(),
            ")"
          ),
          collapse = ""
        )
      }

      private$task_id <- private$compute_task_id()

      eval_log <- eval_log(
        eval = translate_to_eval(
          run_id = private$run_id,
          task = private$task_id,
          task_id = private$task_id,
          dataset = list(
            samples = length(unique(samples$id)),
            sample_ids = as.list(seq_len(length(unique(samples$id)))),
            shuffled = FALSE
          ),
          model = private$solver_description(),
          scorers = translate_to_eval_scorers(
            name = samples$scorer[[1]]
          )
        ),
        plan = translate_to_plan(
          steps = translate_to_plan_steps(
            name = private$solver_description(),
            arguments = private$solutions$arguments,
            system_prompt = private$samples$solver_chat[[1]]$get_system_prompt()
          )
        ),
        results = translate_to_results(
          total_samples = nrow(samples),
          completed_samples = nrow(samples),
          scores = results_scores(
            name = scorer_name,
            metrics = map(private$metric_results, rename_metric_fields)
          )
        ),
        stats = translate_to_stats(
          started_at = eval_log_timestamp(private$timestamps$solve$started_at),
          completed_at = eval_log_timestamp(
            private$timestamps$score$completed_at
          ),
          model_usage = sum_model_usage(samples$solver_chat)
        ),
        samples = translate_to_samples(
          samples,
          scores = private$scores,
          timestamps = private$timestamps
        )
      )

      if (is.na(dir)) {
        dir <- tempdir()
      }

      log_path <- eval_log_write(eval_log, dir = dir)

      invisible(log_path)
    },

    #' @description
    #' View the task results in the Inspect log viewer
    #'
    #' @return The Task object (invisibly)
    view = function() {
      if (!private$solved) {
        cli::cli_alert_warning(
          "Task has not been evaluated yet. Run task$eval() first."
        )
        return(invisible(self))
      }

      vitals_view(self$dir)
      invisible(self)
    },

    #' @description
    #' Set the solver function
    #'
    #' @return The Task object (invisibly)
    set_solver = function(solver) {
      solver_name <- deparse(substitute(solver))
      private$solver <- logged(solver, fn_name = solver_name)

      private$reset_solutions()

      invisible(self)
    },

    #' @description
    #' Set the scorer function
    #'
    #' @return The Task object (invisibly)
    set_scorer = function(scorer) {
      scorer_name <- deparse(substitute(scorer))
      private$scorer <- logged(scorer, fn_name = scorer_name)
      private$reset_scores()

      invisible(self)
    },

    #' @description
    #' Set the metrics that will be applied in `$measure()` (and thus `$eval()`).
    #'
    #' @return The Task (invisibly)
    set_metrics = function(metrics) {
      metrics_name <- deparse(substitute(metrics))

      if (is.null(metrics)) {
        private$metric_fns <- NULL
      }

      check_metrics(metrics)
      private$metric_fns <- metrics
      private$reset_metrics()

      invisible(self)
    },

    #' @description The cost of this eval
    #' This is a wrapper around ellmer's `$token_usage()` function.
    #' That function is called at the beginning and end of each call to
    #' `$solve()` and `$score()`; this function returns the cost inferred
    #' by taking the differences in values of `$token_usage()` over time.
    #'
    #' @return A tibble displaying the cost of solving and scoring the
    #' evaluation by model, separately for the solver and scorer.
    get_cost = function() {
      bind_token_usage(
        private$solver_token_usage,
        private$scorer_token_usage
      )
    }
  ),

  private = list(
    dataset_name = NULL,

    solver = NULL,
    solver_token_usage = NULL,
    solved = FALSE,

    solver_description = function() {
      sub_name <- private$solutions$name
      model_name <- private$samples$solver_chat[[1]]$get_model()

      sub_name <- gsub("model = ", "", sub_name, fixed = TRUE)
      sub_name <- gsub(model_name, "", sub_name, fixed = TRUE)
      sub_name <- gsub("\"\"", "", sub_name)
      paste0(
        c(
          sub_name,
          " (",
          model_name,
          ")"
        ),
        collapse = ""
      )
    },

    compute_task_id = function() {
      model_name <- private$samples$solver_chat[[1]]$get_model()

      hash_input <- list(
        solver_name = private$solutions$name,
        solver_args = private$solutions$arguments,
        scorer_name = private$scores$name,
        scorer_args = private$scores$arguments
      )

      args_hash <- substr(hash(hash_input), 1, 22)

      paste(private$dataset_name, model_name, args_hash, sep = "/")
    },

    scorer = NULL,
    scorer_token_usage = NULL,
    scored = FALSE,

    samples = NULL,

    # log when solving/scoring starts
    # eventually, this will be derived from the chats themselves (#112)
    timestamps = list(),

    stash_last_task = function() {
      if (!"pkg:vitals" %in% search()) {
        do.call(
          "attach",
          list(new.env(), pos = length(search()), name = "pkg:vitals")
        )
      }
      env <- as.environment("pkg:vitals")
      env$.last_task <- self
      invisible(NULL)
    },

    reset_solutions = function() {
      private$samples$result <- NA
      private$samples$solver_chat <- NULL
      private$samples$solver_metadata <- NULL
      private$timestamps$solve <- NULL
      private$solved <- FALSE

      if ("epoch" %in% names(self$get_samples())) {
        private$samples <- private$samples[
          private$samples$epoch == 1,
        ]
        private$samples$epoch <- NULL
      }

      if (private$scored) {
        private$reset_scores()
      }

      invisible(NULL)
    },

    reset_scores = function() {
      private$samples$score <- NA
      private$samples$scorer_chat <- NULL
      private$samples$scorer_metadata <- NULL
      private$timestamps$score <- NULL
      private$scored <- FALSE

      if (!is.null(self$metrics)) {
        private$reset_metrics()
      }

      invisible(NULL)
    },

    reset_metrics = function() {
      self$metrics <- NULL
    },

    reset_for_new_eval = function() {
      private$reset_solutions()
      private$reset_scores()
      invisible(NULL)
    },

    check_solver_outputs = function() {
      if (
        !all(c("result", "solver_chat") %in% names(private$solutions$value))
      ) {
        cli::cli_abort(
          "{.arg solver} must return slots {.field result} and 
           {.field solver_chat}.",
          call = call2("$solve()")
        )
      }

      first_solver_chat <- private$solutions$value$solver_chat[[1]]
      if (!inherits(first_solver_chat, "Chat")) {
        cli::cli_abort(
          "Elements in the {.field solver_chat} output from {.arg solver} must be
           ellmer Chat objects, not {.obj_type_friendly {first_solver_chat}}.",
          call = call2("$solve()")
        )
      }
    },

    check_scorer_outputs = function() {
      if (!"score" %in% names(private$scores$value)) {
        cli::cli_abort(
          "{.arg scorer} must return a list with (at least) the slot {.field score}.",
          call = call2("$score()")
        )
      }

      if (!"scorer_chat" %in% names(private$scores$value)) {
        return(invisible())
      }

      first_scorer_chat <- private$scores$value$scorer_chat[[1]]
      if (!inherits(first_scorer_chat, "Chat")) {
        cli::cli_abort(
          "Elements in the {.field scorer_chat} output from {.arg scorer} must be
           ellmer Chat objects, not {.obj_type_friendly {first_scorer_chat}}.",
          call = call2("$score()")
        )
      }
    },

    check_metric_output = function() {
      for (metric_result in private$metric_results) {
        if (!is.numeric(metric_result$value)) {
          cli::cli_abort(
            c(
              "Each metric function must return a single numeric value",
              "{.fun {metric_result$name}} returned 
               {.obj_type_friendly {metric_result$value}}"
            ),
            call = call2("measure")
          )
        }
      }
    },

    cbind_solutions = function() {
      private$samples$result <- private$solutions$value$result
      private$samples$solver_chat <- private$solutions$value$solver_chat

      if ("solver_metadata" %in% names(private$solutions$value)) {
        private$samples$solver_metadata <- private$solutions$value$solver_metadata
      }

      invisible()
    },

    cbind_scores = function() {
      scorer_res <- private$scores$value
      private$samples$score <- scorer_res$score
      if ("scorer_chat" %in% names(scorer_res)) {
        private$samples$scorer_chat <- scorer_res$scorer_chat
      }
      if ("scorer_metadata" %in% names(scorer_res)) {
        private$samples$scorer_metadata <- scorer_res$scorer_metadata
      }
      if ("explanation" %in% names(scorer_res)) {
        private$samples$scorer_explanation <- scorer_res$explanation
      }

      private$samples$scorer <- private$scores$name
    },

    # For a named list of metric functions, apply each as `logged(fn_i)(scores)`
    apply_metrics = function() {
      private$metric_results <-
        purrr::map2(
          private$metric_fns,
          names(private$metric_fns),
          function(metric, metric_name) {
            # TODO: handle errors here? or should `logged()` do that?
            logged(metric, metric_name)(self$get_samples()$score)
          }
        )
    },

    # A default metric, mirroring the default accuracy in Inspect
    apply_naive_accuracy = function() {
      if (
        is.factor(self$get_samples()$score) &&
          (any(c("C", "I") %in% levels(self$get_samples()$score)))
      ) {
        # map factor to numeric for a simple accuracy (#51, #53)
        numeric_scores <- as.numeric(self$get_samples()$score) - 1
        if (any(is.na(numeric_scores))) {
          return()
        }
        max_score <- nlevels(self$get_samples()$score) - 1
        if (max_score > 0) {
          numeric_scores <- numeric_scores / max_score
        }
        private$metric_results <-
          list2(
            accuracy = logged(accuracy)(numeric_scores)
          )
      } else if (is.numeric(self$get_samples()$score)) {
        if (any(is.na(self$get_samples()$score))) {
          return()
        }
        private$metric_results <-
          list2(
            accuracy = logged(accuracy)(self$get_samples()$score)
          )
      }
    },

    # `slot` is one of "solver_token_usage" or "scorer_token_usage"
    # `env` is a function's execution environment
    track_token_usage = function(slot, env = caller_env()) {
      suppressMessages({
        initial_usage <- ellmer::token_usage()
      })

      withr::defer(
        {
          suppressMessages({
            final_usage <- ellmer::token_usage()
          })
          private[[slot]] <- diff_token_usage(initial_usage, final_usage)
        },
        envir = env
      )
    },

    # log working_time values by extracting them from Turn @duration slots (#115)
    add_working_times = function(samples) {
      samples$solver_chat <- purrr::map(
        samples$solver_chat,
        add_working_times_to_turns,
        which = "solve",
        timestamps = private$timestamps,
        n = length(samples$solver_chat)
      )

      if ("scorer_chat" %in% names(samples)) {
        samples$scorer_chat <- purrr::map(
          samples$scorer_chat,
          add_working_times_to_turns,
          which = "score",
          timestamps = private$timestamps,
          n = length(samples$scorer_chat)
        )
      }

      samples
    },

    # log working_start values by accumulating Turn @duration values (#115)
    add_working_starts = function(samples) {
      samples$solver_chat <- add_working_start_to_turns(
        samples$solver_chat,
        which = "solve",
        timestamps = private$timestamps
      )

      if ("scorer_chat" %in% names(samples)) {
        samples$scorer_chat <- add_working_start_to_turns(
          samples$scorer_chat,
          which = "score",
          timestamps = private$timestamps
        )
      }

      samples
    },

    split_dots = function(dots) {
      solver_fn <- environment(private$solver)$fn
      scorer_fn <- environment(private$scorer)$fn

      solver_formals <- names(formals(solver_fn))
      scorer_formals <- names(formals(scorer_fn))

      solver_has_dots <- "..." %in% solver_formals
      scorer_has_dots <- "..." %in% scorer_formals

      solver_args <- list()
      scorer_args <- list()

      for (arg_name in names(dots)) {
        matches_solver <- arg_name %in% solver_formals
        matches_scorer <- arg_name %in% scorer_formals

        if (matches_solver) {
          solver_args[[arg_name]] <- dots[[arg_name]]
        }

        if (matches_scorer) {
          scorer_args[[arg_name]] <- dots[[arg_name]]
        }

        if (!matches_solver && !matches_scorer) {
          if (solver_has_dots) {
            solver_args[[arg_name]] <- dots[[arg_name]]
          }
          if (scorer_has_dots) {
            scorer_args[[arg_name]] <- dots[[arg_name]]
          }
          if (!solver_has_dots && !scorer_has_dots) {
            cli::cli_abort(
              "Argument {.arg {arg_name}} does not match any parameter in the solver or scorer functions.",
              call = call2("$eval()")
            )
          }
        }
      }

      list(solver = solver_args, scorer = scorer_args)
    },

    # The output of `logged(solver)(...)`
    solutions = NULL,

    # The output of `logged(scorer)(...)`.
    scores = NULL,

    # Argument supplied to `metrics` in `Task$new()` or `$set_metrics`.
    # Notably, not `logged()` directly--that function is applied in `apply_metrics()`
    metric_fns = NULL,

    # The output of `metrics(...)` (from directly above).
    metric_results = NULL,

    task_id = NULL,
    run_id = NULL,

    epochs = NULL
  )
)

#' @export
#' @importFrom cli cat_line format_inline col_grey
print.Task <- function(x, ...) {
  dataset_name <- x$.__enclos_env__$private$dataset_name

  cli::cat_line(cli::format_inline(
    "An evaluation {cli::col_blue('task')} {.field {dataset_name}}."
  ))

  if ("score" %in% names(x$get_samples())) {
    cli::cat_line(cli::format_inline(
      "Explore interactively with {.run .last_task$view()}."
    ))
  }

  invisible(x)
}


check_dataset <- function(dataset, call = caller_env()) {
  check_data_frame(dataset)

  required_cols <- c("input", "target")
  missing_cols <- required_cols[!required_cols %in% names(dataset)]

  if (length(missing_cols) > 0) {
    cli::cli_abort(
      "{.arg dataset} is missing required column{?s} {.field {missing_cols}}.",
      call = call
    )
  }

  invisible(dataset)
}

# Must be a named list of functions.
check_metrics <- function(metrics, call = caller_env()) {
  if (is.null(metrics)) {
    return()
  }

  if (
    !is.list(metrics) ||
      (is.list(metrics) && !is_named(metrics))
  ) {
    cli::cli_abort(
      "{.arg metrics} must be a named list of functions or NULL, 
       not {.obj_type_friendly {metrics}}",
      call = call
    )
  }

  for (metric in metrics) {
    check_function(metric, call = call)
  }

  invisible(metrics)
}

set_id_column <- function(dataset, call = caller_env()) {
  if ("id" %in% names(dataset)) {
    if (any(duplicated(dataset$id))) {
      cli::cli_abort(
        "Duplicated values found in the {.field id} column. Each ID must be unique.",
        call = call
      )
    }
  } else {
    dataset$id <- seq_len(nrow(dataset))
  }

  invisible(dataset)
}

join_epochs <- function(samples, epochs) {
  if (is.null(epochs)) {
    epochs <- 1L
  }
  if (abs(epochs - 1) < .1) {
    return(samples)
  }

  dplyr::inner_join(
    samples,
    data.frame(
      id = rep(samples$id, each = epochs),
      epoch = rep(seq_len(epochs), times = nrow(samples))
    ),
    by = "id"
  )
}

has_last_task <- function() {
  if (!"pkg:vitals" %in% search()) {
    return(FALSE)
  }

  exists(".last_task", as.environment("pkg:vitals"))
}

diff_token_usage <- function(before, after) {
  if (is.null(before)) {
    return(after)
  }

  res <- dplyr::left_join(
    after,
    before,
    by = c("provider", "model"),
    suffix = c("", "_before")
  )
  res <- dplyr::mutate(
    res,
    input = input - ifelse(is.na(input_before), 0, input_before)
  )
  res <- dplyr::mutate(
    res,
    output = output - ifelse(is.na(output_before), 0, output_before)
  )
  # if `price_before` is present, so will be `price` (#103)
  if ("price_before" %in% colnames(res)) {
    res <- dplyr::mutate(
      res,
      price = dplyr::if_else(
        is.na(price),
        NA,
        numeric_price(price) - numeric_price(price_before)
      )
    )
  }
  if ("price" %in% colnames(res)) {
    res <- dplyr::mutate(
      res,
      price = dplyr::if_else(is.na(price), NA, sprintf("$%.2f", price))
    )
  }
  res <- dplyr::select(res, provider, model, input, output, any_of("price"))

  dplyr::filter(res, input != 0 | output != 0)
}

bind_token_usage <- function(solver, scorer) {
  solver_token_usage <-
    dplyr::mutate(
      solver,
      source = "solver",
      .before = everything()
    )

  if (is.null(scorer)) {
    return(solver_token_usage)
  }

  dplyr::bind_rows(
    solver_token_usage,
    dplyr::mutate(
      scorer,
      source = "scorer",
      .before = everything()
    )
  )
}

numeric_price <- function(price) {
  result <- as.numeric(gsub("\\$", "", price))
  result[is.na(result)] <- 0
  result
}

test_that("vitals_bind works", {
  skip_on_cran()
  task_1 <- example_task()
  task_2 <- example_task()

  res <- vitals_bind(task_1, task_2)

  expect_s3_class(res, "tbl_df")
  expect_named(res, c("task", "id", "score", "metadata"))
  expect_equal(nrow(res), nrow(task_1$get_samples()) * 2)
  expect_equal(sort(unique(res$task)), c("task_1", "task_2"))

  res_named <- vitals_bind(boop = task_1, bop = task_2)
  expect_equal(sort(unique(res_named$task)), c("boop", "bop"))
})

test_that("translate_turns works", {
  skip_on_cran()
  example_sample <- example_task()$get_samples()[1, , drop = FALSE]
  timestamps <- list(started_at = Sys.time(), completed_at = Sys.time())
  chat_translated <- translate_to_events(
    example_sample,
    timestamps = list(solve = timestamps, score = timestamps)
  )

  inspect_log <- example_inspect_log()
  inspect_log_first_events <- inspect_log$samples[[1]]$events

  expect_contains(
    names(inspect_log_first_events[[1]]),
    names(chat_translated[[1]])
  )

  expect_contains(
    names(inspect_log_first_events[[2]]),
    names(chat_translated[[2]])
  )
})

test_that("vitals writes valid eval logs (scorer tool calls, claude)", {
  vcr::local_cassette("translate-anthropic-scorer-tool-calls")
  key_get("ANTHROPIC_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)
  library(ellmer)

  sum_problems <- tibble::tibble(
    input = c(
      "What is 125 + 267?",
      "What is 543 + 789?",
      "What is 91 + 38?"
    ),
    target = c("392", "1332", "129")
  )

  take_sum <- tool(
    function(a, b) {
      list(sum = a + b)
    },
    name = "sum_tool",
    description = "Add two numbers together to verify arithmetic. Always use this tool.",
    arguments = list(
      a = type_number("First number"),
      b = type_number("Second number")
    )
  )

  scorer_chat <- chat_claude(
    model = "claude-sonnet-4-5-20250929",
    system_prompt = paste(
      "You are a scorer that checks arithmetic answers.",
      "ALWAYS use the sum_tool to verify the answer before grading.",
      "Call sum_tool with the numbers from the question, then compare",
      "the result to the submitted answer."
    )
  )
  scorer_chat$register_tool(take_sum)

  tsk <- Task$new(
    dataset = sum_problems,
    solver = generate(chat_claude(model = "claude-sonnet-4-5-20250929")),
    scorer = model_graded_qa(scorer_chat = scorer_chat)
  )
  tsk$eval()
  expect_valid_log(tsk$log())
})

test_that("model_graded_qa works", {
  vcr::local_cassette("scorer-model-graded-qa")
  key_get("OPENAI_API_KEY")
  skip_on_cran()
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)

  library(ellmer)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(chat_openai(model = "gpt-4.1-nano")),
    scorer = model_graded_qa()
  )

  tsk$eval()
  expect_valid_log(tsk$log())

  # returns scores and a complete scorer chat
  expect_s3_class(tsk$get_samples()$score, "factor")
  expect_true(all(levels(tsk$get_samples()$score) %in% c("I", "C")))
  expect_true(is.ordered(tsk$get_samples()$score))
  expect_s3_class(tsk$get_samples()$solver_chat[[1]], "Chat")
  expect_length(tsk$get_samples()$solver_chat[[1]]$get_turns(), 2)
  expect_s3_class(tsk$get_samples()$scorer_chat[[1]], "Chat")
  expect_length(tsk$get_samples()$scorer_chat[[1]]$get_turns(), 2)

  # by default, scorer detects last model used to solve
  expect_equal(
    tsk$get_samples()$solver_chat[[1]]$get_model(),
    tsk$get_samples()$scorer_chat[[1]]$get_model()
  )
})

test_that("model_graded_fact works", {
  vcr::local_cassette("scorer-model-graded-fact")
  key_get("OPENAI_API_KEY")
  skip_on_cran()
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)

  library(ellmer)

  r_history <- tibble::tibble(
    input = c(
      "Who created R?",
      "In what year was version 1.0.0 of R released?"
    ),
    target = c("Ross Ihaka and Robert Gentleman.", "2000.")
  )

  tsk <- Task$new(
    dataset = r_history,
    solver = generate(chat_openai(model = "gpt-4.1-nano")),
    scorer = model_graded_fact()
  )

  tsk$eval()
  expect_valid_log(tsk$log())

  # returns scores and a complete scorer chat
  expect_s3_class(tsk$get_samples()$score, "factor")
  expect_true(all(levels(tsk$get_samples()$score) %in% c("I", "C")))
  expect_true(is.ordered(tsk$get_samples()$score))
  expect_s3_class(tsk$get_samples()$solver_chat[[1]], "Chat")
  expect_length(tsk$get_samples()$solver_chat[[1]]$get_turns(), 2)
  expect_s3_class(tsk$get_samples()$scorer_chat[[1]], "Chat")
  expect_length(tsk$get_samples()$scorer_chat[[1]]$get_turns(), 2)

  # by default, scorer detects last model used to solve
  expect_equal(
    tsk$get_samples()$solver_chat[[1]]$get_model(),
    tsk$get_samples()$scorer_chat[[1]]$get_model()
  )
})

test_that("model_graded_qa respects non-default grading schemes", {
  vcr::local_cassette("scorer-model-graded-qa-custom-ab")
  key_get("OPENAI_API_KEY")
  skip_on_cran()
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)

  library(ellmer)

  classification_data <- tibble::tibble(
    input = c("The dog ran quickly", "The cat meowed loudly"),
    target = c("A", "B")
  )

  custom_instructions <- "Classify the sentence. Reply with 'GRADE: $LETTER' where LETTER is either A for sentences about dogs or B for sentences about cats."
  custom_pattern <- "(?i)GRADE\\s*:\\s*([AB])(.*)$"

  tsk <- Task$new(
    dataset = classification_data,
    solver = generate(chat_openai(model = "gpt-4.1-nano")),
    scorer = model_graded_qa(
      instructions = custom_instructions,
      grade_pattern = custom_pattern
    )
  )

  tsk$eval()
  expect_valid_log(tsk$log())

  scores <- tsk$get_samples()$score
  expect_type(scores, "character")
  expect_true(all(scores %in% c("A", "B")))
  expect_false(is.factor(scores))
})

test_that("generate works", {
  vcr::local_cassette("solver-generate")
  key_get("OPENAI_API_KEY")
  library(ellmer)

  res <- generate(chat_openai(model = "gpt-4.1-nano"))
  expect_contains(class(res), "function")

  chat_res <- res(list("hey", "hi", "hello"))

  expect_length(chat_res, 2)
  expect_length(chat_res[["result"]], 3)
  expect_length(chat_res[["solver_chat"]], 3)
  expect_type(chat_res[["result"]][[1]], "character")
  expect_s3_class(chat_res[["solver_chat"]][[1]], "Chat")
})

test_that("generate() allows NULL default model", {
  res <- generate()
  expect_contains(class(res), "function")
  expect_snapshot(res(), error = TRUE)
})

test_that("generate_structured works", {
  vcr::local_cassette("solver-generate-structured")
  key_get("ANTHROPIC_API_KEY")
  library(ellmer)

  type_answer <- type_object(
    answer = type_string("The answer to the math question")
  )

  res <- generate_structured(
    solver_chat = chat_anthropic(model = "claude-sonnet-4-20250514"),
    type = type_answer
  )

  expect_contains(class(res), "function")

  chat_res <- res(list("What is 2+2?", "What is 3+3?"))

  expect_length(chat_res, 3)
  expect_length(chat_res[["result"]], 2)
  expect_type(chat_res[["result"]][[1]], "character")
  expect_length(chat_res[["solver_chat"]], 2)
  expect_s3_class(chat_res[["solver_chat"]][[1]], "Chat")
  expect_length(chat_res[["solver_metadata"]], 2)
  expect_true("answer" %in% names(chat_res[["solver_metadata"]][[1]]))
})

test_that("generate_structured() allows NULL default model", {
  res <- generate_structured()
  expect_contains(class(res), "function")
  expect_snapshot(res(), error = TRUE)
})

test_that("generate_structured produces valid logs", {
  vcr::local_cassette("solver-generate-structured-task")
  key_get("ANTHROPIC_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)
  library(ellmer)

  type_answer <- type_object(
    answer = type_string("The numeric answer to the math question")
  )

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate_structured(
      solver_chat = chat_anthropic(model = "claude-sonnet-4-20250514"),
      type = type_answer
    ),
    scorer = model_graded_qa()
  )

  tsk$eval()
  expect_valid_log(tsk$log())

  samples <- tsk$get_samples()
  expect_type(samples$result, "character")
  expect_length(samples$result, 2)
  expect_length(samples$solver_chat, 2)
  expect_s3_class(samples$solver_chat[[1]], "Chat")
  expect_length(samples$solver_metadata, 2)
  expect_true("answer" %in% names(samples$solver_metadata[[1]]))
})

test_that("expect_valid_log fails when log file is nonsense", {
  # use a file name that "looks like" it could be a real Inspect log so that
  # Inspect tries to read it
  tmp_dir <- withr::local_tempdir()
  tmp_file <- file.path(
    tmp_dir,
    "2025-04-02T16-49-36-05-00_simpleaddition_e1e56aeef83a77f6392787.json"
  )
  file.create(tmp_file)
  writeLines(c("beep", "bop", "boop"), tmp_file)

  expect_error(
    suppressWarnings(expect_valid_log(tmp_file)),
    regexp = "Expecting value: line 1 column 1"
  )
})

test_that("vitals writes valid eval logs (basic, openai)", {
  vcr::local_cassette("translate-openai-basic")
  key_get("OPENAI_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(ellmer::chat_openai(model = "gpt-4.1-nano")),
    scorer = model_graded_qa()
  )
  tsk$eval()
  expect_valid_log(tsk$log())
})

test_that("vitals writes valid eval logs (basic, claude)", {
  vcr::local_cassette("translate-anthropic-basic")
  key_get("ANTHROPIC_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(ellmer::chat_claude(
      model = "claude-sonnet-4-5-20250929"
    )),
    scorer = model_graded_qa()
  )
  tsk$eval()
  expect_valid_log(tsk$log())
})

test_that("vitals writes valid eval logs (basic, gemini)", {
  vcr::local_cassette("translate-google-basic")
  key_get("GOOGLE_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(ellmer::chat_google_gemini(model = "gemini-2.0-flash")),
    scorer = model_graded_qa()
  )
  tsk$eval()
  expect_valid_log(tsk$log())
})


test_that("vitals writes valid eval logs (solver tool calls, claude)", {
  vcr::local_cassette("translate-anthropic-tool-calls")
  key_get("ANTHROPIC_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)
  library(ellmer)

  current_date <- tibble::tibble(
    input = c("What's the current date?"),
    target = c("2024-01-01")
  )

  ch <- chat_claude(model = "claude-sonnet-4-5-20250929")
  ch$register_tool(tool(
    function() "2024-01-01",
    name = "get_current_date",
    description = "Return the current date"
  ))

  tsk <- Task$new(
    dataset = current_date,
    solver = generate(ch),
    scorer = function(samples) {
      list(score = factor("C", levels = c("I", "C"), ordered = TRUE))
    }
  )
  tsk$eval()
  expect_valid_log(tsk$log())
})

test_that("vitals writes valid eval logs (solver errors on tool call, claude)", {
  key_get("ANTHROPIC_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)
  library(ellmer)

  current_date <- tibble::tibble(
    input = c("What's the current date?"),
    target = c("2024-01-01")
  )

  ch <- chat_claude(model = "claude-sonnet-4-5-20250929")
  ch$register_tool(
    tool(
      function() stop("Couldn't find the date"),
      name = "get_current_date",
      description = "Return the current date"
    )
  )

  tsk <- Task$new(
    dataset = current_date,
    solver = generate(ch),
    scorer = function(samples) {
      list(score = factor("C", levels = c("I", "C"), ordered = TRUE))
    }
  )

  # a tool call will fail here and raise a warning; this is intentional.
  # since raised from ellmer, we don't expect anything specific about it.
  suppressWarnings(tsk$eval())

  log_file <- list.files(tmp_dir, full.names = TRUE)
  expect_gte(length(log_file), 1)

  expect_valid_log(log_file[1])
})

test_that("vitals writes valid logs with numeric solver results (#145)", {
  vcr::local_cassette("translate-numeric-results")
  key_get("ANTHROPIC_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)

  simple_dataset <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  chat <- ellmer::chat_claude(model = "claude-sonnet-4-5-20250929")
  chat$chat("Hey!", echo = FALSE)

  simple_solver <- function(inputs) {
    list(
      result = rep(1.5, length(inputs)),
      solver_chat = lapply(inputs, function(x) {
        chat
      })
    )
  }

  tsk <- Task$new(
    dataset = simple_dataset,
    solver = simple_solver,
    scorer = model_graded_qa()
  )

  tsk$eval()
  log_path <- tsk$log()
  expect_valid_log(log_path)
})

test_that("as_metadata can make lists into a named vector (#69)", {
  res <- as_metadata(mtcars[1:2, 1:2])
  expect_type(res, "list")
  expect_named(res, c("mpg", "cyl"))
  expect_snapshot(res)

  res <- as_metadata(tibble(x = list(list(a = 1, b = 2)), y = 3))
  expect_type(res, "list")
  expect_named(res, c("x", "y"))
  expect_snapshot(res)

  res <- as_metadata(list(1:3, b = 2))
  expect_type(res, "list")
  expect_named(res, c("1", "b"))
  expect_snapshot(res)

  res <- as_metadata(list(1:3, b = 2))
  expect_type(res, "list")
  expect_named(res, c("1", "b"))
  expect_snapshot(res)

  vec <- c(a = 1, b = 2)
  expect_equal(as_metadata(vec), as.list(vec))
})

test_that("Task R6 class works", {
  vcr::local_cassette("task-basic")
  key_get("OPENAI_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)
  library(ellmer)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(chat_openai(model = "gpt-4.1-nano")),
    scorer = model_graded_qa()
  )

  expect_true(R6::is.R6(tsk))
  expect_true(inherits(tsk, "Task"))
  expect_snapshot(tsk)

  expect_equal(nrow(tsk$get_samples()), nrow(simple_addition))
  expect_named(tsk$get_samples(), c("input", "target", "id"))

  tsk$eval()
  expect_valid_log(tsk$log())
  expect_snapshot(tsk)

  expect_named(
    tsk$get_samples(),
    c(
      "input",
      "target",
      "id",
      "result",
      "solver_chat",
      "score",
      "scorer",
      "scorer_chat",
      "scorer_metadata"
    ),
    ignore.order = TRUE
  )

  expect_equal(tsk, .last_task)
})

test_that("Task with epochs works", {
  vcr::local_cassette("task-epochs")
  key_get("OPENAI_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)

  library(ellmer)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(chat_openai(model = "gpt-4.1-nano")),
    scorer = model_graded_qa()
  )

  tsk$eval(epochs = 2)
  expect_valid_log(tsk$log())

  expect_equal(nrow(tsk$get_samples()), nrow(simple_addition) * 2)
  expect_named(
    tsk$get_samples(),
    c(
      "input",
      "target",
      "id",
      "epoch",
      "result",
      "solver_chat",
      "score",
      "scorer",
      "scorer_chat",
      "scorer_metadata"
    ),
    ignore.order = TRUE
  )
})

test_that("Task respects `$new(epochs)`", {
  vcr::local_cassette("task-new-epochs")
  key_get("OPENAI_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)

  library(ellmer)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(chat_openai(model = "gpt-4.1-nano")),
    scorer = model_graded_qa(),
    epochs = 2
  )

  tsk$eval()
  expect_valid_log(tsk$log())

  expect_equal(nrow(tsk$get_samples()), nrow(simple_addition) * 2)
})

test_that("`$eval(epochs)` takes precedence over `$new(epochs)`", {
  vcr::local_cassette("task-eval-epochs-precedence")
  key_get("OPENAI_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)

  library(ellmer)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(chat_openai(model = "gpt-4.1-nano")),
    scorer = model_graded_qa(),
    epochs = 2
  )

  tsk$eval(epochs = 1)
  expect_valid_log(tsk$log())

  expect_equal(nrow(tsk$get_samples()), nrow(simple_addition))
})

test_that("check_dataset works", {
  expect_snapshot(
    Task$new(
      dataset = data.frame(input = 1),
      solver = function() {},
      scorer = function() {}
    ),
    error = TRUE
  )
  expect_snapshot(
    Task$new(
      dataset = data.frame(target = 1),
      solver = function() {},
      scorer = function() {}
    ),
    error = TRUE
  )
  expect_snapshot(
    Task$new(
      dataset = data.frame(x = 1),
      solver = function() {},
      scorer = function() {}
    ),
    error = TRUE
  )

  d <- data.frame(input = "hey", target = "there")
  expect_equal(d, check_dataset(d))
})

test_that("join_epochs() works", {
  task_data <- data.frame(something = "here", id = 1:3)
  expect_equal(join_epochs(task_data, 1), task_data)

  joined <- join_epochs(task_data, 2)
  expect_equal(nrow(joined), nrow(task_data) * 2)
  expect_equal(joined$epoch, rep(1:2, 3))
  expect_equal(joined$id, rep(1:3, each = 2))
})

test_that("set_id_column works", {
  # no existing `id`
  df <- tibble::tibble(input = c("a", "b"), target = c("c", "d"))
  result <- set_id_column(df)

  expect_equal(nrow(result), 2)
  expect_true("id" %in% names(result))
  expect_equal(result$id, 1:2)

  # existing `id``
  df <- tibble::tibble(input = c("a", "b"), target = c("c", "d"), id = c(5, 10))
  result <- set_id_column(df)

  expect_equal(nrow(result), 2)
  expect_true("id" %in% names(result))
  expect_equal(result$id, c(5, 10))
})

test_that("Task preserves existing id column", {
  withr::local_envvar(list(VITALS_LOG_DIR = withr::local_tempdir()))
  local_mocked_bindings(interactive = function(...) FALSE)

  d <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5"),
    id = c(10, 20)
  )

  tsk <- Task$new(
    dataset = d,
    solver = function() {},
    scorer = function() {}
  )

  expect_equal(tsk$get_samples()$id, c(10, 20))
})

test_that("Task errors informatively with duplicate ids", {
  withr::local_envvar(list(VITALS_LOG_DIR = withr::local_tempdir()))

  d <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5"),
    id = c(10, 10)
  )

  expect_snapshot(
    Task$new(
      dataset = d,
      solver = function() {},
      scorer = function() {}
    ),
    error = TRUE
  )
})

# solver ------------------------------------------------------------------
test_that("set_solver works", {
  key_get("OPENAI_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    result = c("4", "5"),
    target = c("4", "5")
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = function() {},
    scorer = function() {}
  )

  new_solver <- function(inputs) {
    list(
      result = c("4", "5"),
      solver_chat = list(
        ellmer::chat_openai(model = "gpt-4.1-nano"),
        ellmer::chat_openai(model = "gpt-4.1-nano")
      )
    )
  }
  tsk$set_solver(new_solver)
  tsk$solve()

  expect_equal(tsk$get_samples()$result, c("4", "5"))
  expect_false("solver_metadata" %in% names(tsk$get_samples()))

  # set a new solver that includes metadata
  new_solver <- function(inputs) {
    list(
      result = c("4", "5"),
      solver_chat = list(
        ellmer::chat_openai(model = "gpt-4.1-nano"),
        ellmer::chat_openai(model = "gpt-4.1-nano")
      ),
      solver_metadata = c("boop!", "bop!")
    )
  }
  tsk$set_solver(new_solver)
  expect_false(
    any(c("solver_chat", "solver_metadata") %in% names(tsk$get_samples()))
  )
  tsk$solve()

  expect_true("solver_metadata" %in% names(tsk$get_samples()))
})

test_that("set_solver works", {
  key_get("OPENAI_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    result = c("4", "5"),
    target = c("4", "5")
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = function() {},
    scorer = function() {}
  )

  new_solver <- function(inputs) {
    list(
      result = c("4", "5"),
      solver_chat = list(
        ellmer::chat_openai(model = "gpt-4.1-nano"),
        ellmer::chat_openai(model = "gpt-4.1-nano")
      )
    )
  }
  tsk$set_solver(new_solver)
  tsk$solve()

  expect_equal(tsk$get_samples()$result, c("4", "5"))
  expect_false("solver_metadata" %in% names(tsk$get_samples()))

  # set a new solver that includes metadata
  new_solver <- function(inputs) {
    list(
      result = c("4", "5"),
      solver_chat = list(
        ellmer::chat_openai(model = "gpt-4.1-nano"),
        ellmer::chat_openai(model = "gpt-4.1-nano")
      ),
      solver_metadata = c("boop!", "bop!")
    )
  }
  tsk$set_solver(new_solver)
  expect_false(
    any(c("solver_chat", "solver_metadata") %in% names(tsk$get_samples()))
  )
  tsk$solve()

  expect_true("solver_metadata" %in% names(tsk$get_samples()))
})

# scorer ------------------------------------------------------------------
test_that("set_scorer works", {
  key_get("OPENAI_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    result = c("4", "5"),
    target = c("4", "5")
  )

  solver <- function(inputs) {
    list(
      result = c("4", "5"),
      solver_chat = list(
        ellmer::chat_openai(model = "gpt-4.1-nano"),
        ellmer::chat_openai(model = "gpt-4.1-nano")
      )
    )
  }

  tsk <- Task$new(
    dataset = simple_addition,
    solver = solver,
    scorer = function() {}
  )

  tsk$solve()

  # first, return only the score
  scorer_minimal <- function(samples) {
    list(score = c(1, 1))
  }
  tsk$set_scorer(scorer_minimal)
  tsk$score()

  expect_equal(tsk$get_samples()$score, c(1, 1))
  expect_false(any(
    c("scorer_chat", "scorer_metadata") %in% names(tsk$get_samples())
  ))

  # return scorer chats
  scorer_chat <- function(samples) {
    list(
      score = c(1, 1),
      scorer_chat = list(
        ellmer::chat_openai(model = "gpt-4.1-nano"),
        ellmer::chat_openai(model = "gpt-4.1-nano")
      )
    )
  }
  tsk$set_scorer(scorer_chat)
  expect_true(all(is.na(tsk$get_samples()$score)))
  tsk$score()
  expect_true("scorer_chat" %in% names(tsk$get_samples()))

  # return metadata, too
  scorer_metadata <- function(samples) {
    list(
      score = c(1, 1),
      scorer_chat = list(
        ellmer::chat_openai(model = "gpt-4.1-nano"),
        ellmer::chat_openai(model = "gpt-4.1-nano")
      ),
      scorer_metadata = c("beep", "bop")
    )
  }
  tsk$set_scorer(scorer_metadata)
  expect_true(all(is.na(tsk$get_samples()$score)))
  expect_false(any(
    c("scorer_chat", "scorer_metadata") %in% names(tsk$get_samples())
  ))
  tsk$score()
  expect_true(all(
    c("scorer_chat", "scorer_metadata") %in% names(tsk$get_samples())
  ))
})

# metrics ------------------------------------------------------------------
test_that("default metrics are applied effectively", {
  vcr::local_cassette("task-default-metrics")
  key_get("OPENAI_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)
  library(ellmer)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(ellmer::chat_openai(model = "gpt-4.1-nano")),
    scorer = function(...) {
      list(
        score = factor(c("C", "C"), levels = c("I", "P", "C"))
      )
    }
  )

  tsk$eval()

  expect_equal(tsk$metrics, c("accuracy" = 100))
  expect_valid_log(tsk$log())
})

test_that("naive accuracy normalizes by number of factor levels, not observed max", {
  vcr::local_cassette("task-default-metrics")
  key_get("OPENAI_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)
  library(ellmer)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(ellmer::chat_openai(model = "gpt-4.1-nano")),
    scorer = function(...) {
      list(
        score = factor(c("P", "I"), levels = c("I", "P", "C"))
      )
    }
  )

  tsk$eval()

  expect_equal(tsk$metrics, c("accuracy" = 25))
})

test_that("task applies non-default metrics", {
  vcr::local_cassette("task-non-default-metrics")
  key_get("OPENAI_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)
  library(ellmer)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  # via Task$new()...
  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(ellmer::chat_openai(model = "gpt-4.1-nano")),
    scorer = function(...) {
      list(
        score = factor(c("C", "C"), levels = c("I", "P", "C"))
      )
    },
    metrics = list(pct_correct = function(scores) {
      mean(scores == "C") * 100
    })
  )

  tsk$eval()

  expect_equal(tsk$metrics, c("pct_correct" = 100))
  expect_valid_log(tsk$log())

  # via set_metrics...
  tsk$set_metrics(list(prop_correct = function(scores) {
    mean(scores == "C")
  }))
  expect_null(tsk$metrics)
  tsk$measure()
  expect_equal(tsk$metrics, c("prop_correct" = 1))
})

test_that("task errors informatively with bad metrics", {
  vcr::local_cassette("task-bad-metrics")
  key_get("OPENAI_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)
  library(ellmer)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  # wrong type supplied to `$new()`
  expect_snapshot(
    tsk <- Task$new(
      dataset = simple_addition,
      solver = generate(ellmer::chat_openai(model = "gpt-4.1-nano")),
      scorer = function(...) {
        list(
          score = factor(c("C", "C"), levels = c("I", "P", "C"))
        )
      },
      metrics = function(scores) {
        mean(scores == "C") * 100
      }
    ),
    error = TRUE
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(ellmer::chat_openai(model = "gpt-4.1-nano")),
    scorer = function(...) {
      list(
        score = factor(c("C", "C"), levels = c("I", "P", "C"))
      )
    },
    metrics = list(pct_correct = function(scores) {
      mean(scores == "C") * 100
    })
  )

  # wrong type supplied to `$set_metrics()`
  expect_snapshot(
    tsk$set_metrics(function(...) "boop bop"),
    error = TRUE
  )

  # valid type but bad return type
  expect_snapshot(
    {
      tsk$set_metrics(list(
        bad_metric = function(scores) "this is not a numeric"
      ))
      tsk$eval()
    },
    error = TRUE
  )
})

# misc ------------------------------------------------------------------
test_that("task ids are NULL before logging, computed after", {
  key_get("OPENAI_API_KEY")

  tsk <-
    Task$new(
      dataset = are,
      solver = generate(),
      scorer = model_graded_qa()
    )

  expect_null(tsk$.__enclos_env__$private$task_id)
})

test_that("task ids have correct format after logging", {
  vcr::local_cassette("task-id-format")
  key_get("OPENAI_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?"),
    target = c("4")
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(chat_openai(model = "gpt-4.1-nano")),
    scorer = detect_exact()
  )

  tsk$eval()

  task_id <- tsk$.__enclos_env__$private$task_id
  expect_true(grepl("/", task_id, fixed = TRUE))
  expect_true(grepl("gpt-4.1-nano", task_id, fixed = TRUE))
})

test_that("Task completeness is tracked and preserved", {
  vcr::local_cassette("task-completeness")
  key_get("OPENAI_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)
  library(ellmer)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  mock_scorer <- function(samples) {
    list(
      score = c(1),
      metadata = list(NULL)
    )
  }

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(chat_openai(model = "gpt-4.1-nano")),
    scorer = mock_scorer
  )

  expect_false(tsk$.__enclos_env__$private$solved)
  expect_false(tsk$.__enclos_env__$private$scored)

  tsk$solve()
  expect_true(tsk$.__enclos_env__$private$solved)

  tsk$score()
  expect_true(tsk$.__enclos_env__$private$scored)

  tsk$set_solver(generate(chat_openai(model = "gpt-4.1-nano")))
  expect_false(tsk$.__enclos_env__$private$solved)

  tsk$solve()
  expect_true(tsk$.__enclos_env__$private$solved)

  tsk$set_scorer(mock_scorer)
  expect_false(tsk$.__enclos_env__$private$scored)

  tsk$solve()
  tsk$score()

  tsk_clone <- tsk$clone()
  original_results <- tsk$get_samples()$result
  original_scores <- tsk$get_samples()$score

  tsk_clone$eval()
  # TODO: expect_valid_log(tsk$log())
  expect_equal(nrow(tsk_clone$get_samples()), nrow(simple_addition))

  expect_equal(tsk$get_samples()$result, original_results)
  expect_equal(tsk$get_samples()$score, original_scores)

  # test re-evaluation with epochs
  tsk_epochs <- Task$new(
    dataset = simple_addition,
    solver = generate(chat_openai(model = "gpt-4.1-nano")),
    scorer = mock_scorer
  )

  tsk_epochs$eval(epochs = 2)
  # TODO: expect_valid_log(tsk$log())
  expect_equal(nrow(tsk_epochs$get_samples()), nrow(simple_addition) * 2)
  expect_true("epoch" %in% names(tsk_epochs$get_samples()))

  tsk_epochs$eval(epochs = 3)
  # TODO: expect_valid_log(tsk$log())
  expect_equal(nrow(tsk_epochs$get_samples()), nrow(simple_addition) * 3)
  expect_true("epoch" %in% names(tsk_epochs$get_samples()))
})

test_that("Task errors informatively with bad solver output", {
  withr::local_envvar(list(VITALS_LOG_DIR = withr::local_tempdir()))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  bad_solver_missing_fields <- function(inputs) {
    list(
      wrong_name = c("4", "5")
      # missing solver_chat
    )
  }

  tsk <- Task$new(
    dataset = simple_addition,
    solver = bad_solver_missing_fields,
    scorer = function() {}
  )

  expect_snapshot(tsk$solve(), error = TRUE)
})

test_that("Task detects non-Chat objects in solver_chat", {
  withr::local_envvar(list(VITALS_LOG_DIR = withr::local_tempdir()))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  bad_solver_wrong_type <- function(inputs) {
    list(
      result = c("4", "5"),
      solver_chat = list("not a Chat object", "also not a Chat object")
    )
  }

  tsk <- Task$new(
    dataset = simple_addition,
    solver = bad_solver_wrong_type,
    scorer = function() {}
  )

  expect_snapshot(tsk$solve(), error = TRUE)
})

test_that("Task errors informatively with bad scorer output", {
  vcr::local_cassette("task-bad-scorer-output")
  key_get("OPENAI_API_KEY")
  withr::local_envvar(list(VITALS_LOG_DIR = withr::local_tempdir()))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)
  library(ellmer)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(chat_openai(model = "gpt-4.1-nano")),
    scorer = function(samples) {
      list(wrong_name = c("4", "5"))
    }
  )

  expect_snapshot(tsk$eval(), error = TRUE)
})

test_that("Task detects non-Chat objects in scorer_chat", {
  vcr::local_cassette("task-non-chat-scorer")
  key_get("OPENAI_API_KEY")
  withr::local_envvar(list(VITALS_LOG_DIR = withr::local_tempdir()))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)
  library(ellmer)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(chat_openai(model = "gpt-4.1-nano")),
    scorer = function(samples) {
      list(
        score = c("4", "5"),
        scorer_chat = list("not a Chat object", "also not a Chat object")
      )
    }
  )

  expect_snapshot(tsk$eval(), error = TRUE)
})

test_that("token usage is logged correctly", {
  vcr::local_cassette("task-token-usage")
  key_get("OPENAI_API_KEY")
  withr::local_envvar(list(VITALS_LOG_DIR = withr::local_tempdir()))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)
  library(ellmer)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  # use a couple tokens to ensure non-NULL
  chat_openai(model = "gpt-4.1-nano")$chat("hey!", echo = "none")
  usage_before <- ellmer::token_usage()
  usage_before <- dplyr::filter(usage_before, model == "gpt-4.1-nano")

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(chat_openai(model = "gpt-4.1-nano")),
    scorer = model_graded_qa()
  )

  tsk$solve()
  usage_after_solve <- ellmer::token_usage()
  usage_after_solve <- dplyr::filter(usage_after_solve, model == "gpt-4.1-nano")
  cost_after_solve <- tsk$get_cost()
  expect_equal(
    cost_after_solve$input,
    usage_after_solve$input - usage_before$input
  )
  expect_equal(
    cost_after_solve$output,
    usage_after_solve$output - usage_before$output
  )

  tsk$score()
  usage_after_score <- ellmer::token_usage()
  usage_after_score <- dplyr::filter(usage_after_score, model == "gpt-4.1-nano")
  cost_after_score <- tsk$get_cost()
  expect_equal(
    cost_after_score$input,
    c(
      usage_after_solve$input - usage_before$input,
      usage_after_score$input - usage_after_solve$input
    )
  )
  expect_equal(
    cost_after_score$output,
    c(
      usage_after_solve$output - usage_before$output,
      usage_after_score$output - usage_after_solve$output
    )
  )
})


test_that("token usage is logged correctly (with unrelated token usage)", {
  vcr::local_cassette("task-token-usage-unrelated")
  key_get("OPENAI_API_KEY")
  key_get("ANTHROPIC_API_KEY")
  withr::local_envvar(list(VITALS_LOG_DIR = withr::local_tempdir()))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)
  library(ellmer)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  # use a couple tokens to ensure non-NULL
  chat_openai(model = "gpt-4.1-nano")$chat("hey!", echo = "none")
  chat_claude(model = "claude-sonnet-4-5-20250929")$chat(
    "hey!",
    echo = "none"
  )
  usage_before <- ellmer::token_usage()
  usage_before <- dplyr::filter(usage_before, model == "gpt-4.1-nano")

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(chat_openai(model = "gpt-4.1-nano")),
    scorer = model_graded_qa()
  )

  tsk$solve()
  usage_after_solve <- ellmer::token_usage()
  usage_after_solve <- dplyr::filter(usage_after_solve, model == "gpt-4.1-nano")
  cost_after_solve <- tsk$get_cost()
  expect_equal(nrow(cost_after_solve), 1)
  expect_equal(
    cost_after_solve$input,
    usage_after_solve$input - usage_before$input
  )
  expect_equal(
    cost_after_solve$output,
    usage_after_solve$output - usage_before$output
  )
})

# argument routing --------------------------------------------------------
test_that("eval routes arguments correctly to solver and scorer", {
  vcr::local_cassette("task-eval-routes-args")
  key_get("OPENAI_API_KEY")
  withr::local_envvar(list(VITALS_LOG_DIR = withr::local_tempdir()))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)
  library(ellmer)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  solver_arg_tracker <- NULL
  scorer_arg_tracker <- NULL

  ch <- chat_openai(model = "gpt-4.1-nano")
  ch$chat("Hey!", echo = FALSE)

  custom_solver <- function(inputs, solver_param = "default_solver") {
    solver_arg_tracker <<- solver_param
    list(
      result = c("4", "5"),
      solver_chat = list(ch, ch)
    )
  }

  custom_scorer <- function(samples, scorer_param = "default_scorer") {
    scorer_arg_tracker <<- scorer_param
    list(score = c(1, 1))
  }

  tsk <- Task$new(
    dataset = simple_addition,
    solver = custom_solver,
    scorer = custom_scorer
  )

  tsk$eval(solver_param = "passed_to_solver", scorer_param = "passed_to_scorer")

  expect_equal(solver_arg_tracker, "passed_to_solver")
  expect_equal(scorer_arg_tracker, "passed_to_scorer")
})

test_that("eval routes arguments to functions with ellipses", {
  vcr::local_cassette("task-eval-routes-ellipses")
  key_get("OPENAI_API_KEY")
  withr::local_envvar(list(VITALS_LOG_DIR = withr::local_tempdir()))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)
  library(ellmer)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  solver_dots_tracker <- NULL
  scorer_dots_tracker <- NULL

  ch <- chat_openai(model = "gpt-4.1-nano")
  ch$chat("Hey!", echo = FALSE)

  solver_with_dots <- function(inputs, ...) {
    dots <- list(...)
    solver_dots_tracker <<- dots
    list(
      result = c("4", "5"),
      solver_chat = list(ch, ch)
    )
  }

  scorer_with_dots <- function(samples, ...) {
    dots <- list(...)
    scorer_dots_tracker <<- dots
    list(score = c(1, 1))
  }

  tsk <- Task$new(
    dataset = simple_addition,
    solver = solver_with_dots,
    scorer = scorer_with_dots
  )

  tsk$eval(unmatched_param = "goes_to_both")

  expect_equal(solver_dots_tracker$unmatched_param, "goes_to_both")
  expect_equal(scorer_dots_tracker$unmatched_param, "goes_to_both")
})

test_that("eval routes unmatched arguments to solver with ellipses only", {
  vcr::local_cassette("task-eval-routes-unmatched")
  key_get("OPENAI_API_KEY")
  withr::local_envvar(list(VITALS_LOG_DIR = withr::local_tempdir()))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)
  library(ellmer)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  solver_dots_tracker <- NULL

  ch <- chat_openai(model = "gpt-4.1-nano")
  ch$chat("Hey!", echo = FALSE)

  solver_with_dots <- function(inputs, ...) {
    dots <- list(...)
    solver_dots_tracker <<- dots
    list(
      result = c("4", "5"),
      solver_chat = list(ch, ch)
    )
  }

  scorer_no_dots <- function(samples) {
    list(score = c(1, 1))
  }

  tsk <- Task$new(
    dataset = simple_addition,
    solver = solver_with_dots,
    scorer = scorer_no_dots
  )

  tsk$eval(unmatched_param = "goes_to_solver")

  expect_equal(solver_dots_tracker$unmatched_param, "goes_to_solver")
})

test_that("eval errors with unnamed arguments", {
  withr::local_envvar(list(VITALS_LOG_DIR = withr::local_tempdir()))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = function(inputs) {
      list(result = c("4", "5"), solver_chat = list(NULL, NULL))
    },
    scorer = function(samples) {
      list(score = c(1, 1))
    }
  )

  expect_snapshot(tsk$eval("unnamed_arg"), error = TRUE)
})

test_that("eval errors when argument matches neither function and neither has ellipses", {
  vcr::local_cassette("task-eval-errors-bad-arg")
  key_get("OPENAI_API_KEY")
  withr::local_envvar(list(VITALS_LOG_DIR = withr::local_tempdir()))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)
  library(ellmer)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  ch <- chat_openai(model = "gpt-4.1-nano")
  ch$chat("Hey!", echo = FALSE)

  solver_no_dots <- function(inputs, solver_param = "default") {
    list(
      result = c("4", "5"),
      solver_chat = list(ch, ch)
    )
  }

  scorer_no_dots <- function(samples, scorer_param = "default") {
    list(score = c(1, 1))
  }

  tsk <- Task$new(
    dataset = simple_addition,
    solver = solver_no_dots,
    scorer = scorer_no_dots
  )

  expect_snapshot(tsk$eval(unmatched_param = "error"), error = TRUE)
})

test_that("$log() respects dir argument", {
  vcr::local_cassette("task-log-dir-argument")
  key_get("OPENAI_API_KEY")
  tmp_dir_env <- withr::local_tempdir()
  tmp_dir_arg <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir_env))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)
  library(ellmer)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(chat_openai(model = "gpt-4.1-nano")),
    scorer = model_graded_qa()
  )

  tsk$solve()
  tsk$score()
  tsk$measure()

  log_path <- tsk$log(dir = tmp_dir_arg)

  expect_true(file.exists(log_path))
  expect_true(grepl(tmp_dir_arg, log_path, fixed = TRUE))

  files_in_env_dir <- list.files(tmp_dir_env, pattern = "\\.json$")
  files_in_arg_dir <- list.files(tmp_dir_arg, pattern = "\\.json$")

  expect_equal(length(files_in_arg_dir), 1)
  expect_equal(length(files_in_env_dir), 0)
})

test_that("detect_includes works", {
  skip_on_cran()
  tsk <- example_task(scored = FALSE)
  tsk$set_scorer(detect_includes())
  tsk$score()

  expect_s3_class(tsk$get_samples()$score, "factor")
  expect_true(is.ordered(tsk$get_samples()$score))
  expect_equal(as.character(tsk$get_samples()$score), c("C", "C"))
  # TODO: include and test for `metadata` slots (here and in model-based)

  simple_df <- tibble::tibble(
    input = c("Question 1", "Question 2"),
    result = c("The answer is C", "The answer is D"),
    target = c("c", "d")
  )

  tsk_insensitive <- Task$new(
    dataset = simple_df,
    solver = function() {},
    scorer = function() {}
  )
  tsk_insensitive$.__enclos_env__$private$solved <- TRUE
  tsk_insensitive$set_scorer(detect_includes(case_sensitive = FALSE))
  tsk_insensitive$score()

  expect_s3_class(tsk_insensitive$get_samples()$score, "factor")
  expect_true(is.ordered(tsk_insensitive$get_samples()$score))
  expect_equal(as.character(tsk_insensitive$get_samples()$score), c("C", "C"))

  tsk_sensitive <- Task$new(
    dataset = simple_df,
    solver = function() {},
    scorer = function() {}
  )
  tsk_sensitive$.__enclos_env__$private$solved <- TRUE
  tsk_sensitive$set_scorer(detect_includes(case_sensitive = TRUE))
  tsk_sensitive$score()
  expect_s3_class(tsk_sensitive$get_samples()$score, "factor")
  expect_true(is.ordered(tsk_sensitive$get_samples()$score))
  expect_equal(as.character(tsk_sensitive$get_samples()$score), c("I", "I"))
})

test_that("detect_match works", {
  skip_on_cran()
  tsk <- example_task(scored = FALSE)
  tsk$set_scorer(detect_match())
  tsk$score()

  expect_s3_class(tsk$get_samples()$score, "factor")
  expect_true(is.ordered(tsk$get_samples()$score))
  expect_equal(as.character(tsk$get_samples()$score), c("C", "C"))

  simple_df <- tibble::tibble(
    input = c("Question 1", "Question 2"),
    result = c("The answer is C", "The answer is D"),
    target = c("c", "d")
  )

  tsk_insensitive <- Task$new(
    dataset = simple_df,
    solver = function() {},
    scorer = function() {}
  )
  tsk_insensitive$.__enclos_env__$private$solved <- TRUE
  tsk_insensitive$set_scorer(detect_match(case_sensitive = FALSE))
  tsk_insensitive$score()
  expect_s3_class(tsk_insensitive$get_samples()$score, "factor")
  expect_true(is.ordered(tsk_insensitive$get_samples()$score))
  expect_equal(as.character(tsk_insensitive$get_samples()$score), c("C", "C"))

  tsk_sensitive <- Task$new(
    dataset = simple_df,
    solver = function() {},
    scorer = function() {}
  )
  tsk_sensitive$.__enclos_env__$private$solved <- TRUE
  tsk_sensitive$set_scorer(detect_match(case_sensitive = TRUE))
  tsk_sensitive$score()
  expect_s3_class(tsk_sensitive$get_samples()$score, "factor")
  expect_true(is.ordered(tsk_sensitive$get_samples()$score))
  expect_equal(as.character(tsk_sensitive$get_samples()$score), c("I", "I"))
})

test_that("detect_pattern works", {
  skip_on_cran()
  skip_if(getRversion() > "4.4.3")
  tsk <- example_task(scored = FALSE)
  tsk$set_scorer(detect_pattern("(\\d+)\\s*=\\s*(\\d+)"))
  tsk$score()

  expect_s3_class(tsk$get_samples()$score, "factor")
  expect_true(is.ordered(tsk$get_samples()$score))
  expect_equal(as.character(tsk$get_samples()$score), c("C", "C"))

  case_df <- tibble::tibble(
    input = c("Question 1", "Question 2"),
    result = c("The answer contains C", "The answer contains D"),
    target = c("c", "d")
  )

  tsk_insensitive <- Task$new(
    dataset = case_df,
    solver = function() {},
    scorer = function() {}
  )
  tsk_insensitive$.__enclos_env__$private$solved <- TRUE
  tsk_insensitive$set_scorer(detect_pattern(
    "contains\\s+([A-Za-z])",
    case_sensitive = FALSE
  ))
  tsk_insensitive$score()
  expect_s3_class(tsk_insensitive$get_samples()$score, "factor")
  expect_true(is.ordered(tsk_insensitive$get_samples()$score))
  expect_equal(as.character(tsk_insensitive$get_samples()$score), c("C", "C"))

  tsk_sensitive <- Task$new(
    dataset = case_df,
    solver = function() {},
    scorer = function() {}
  )
  tsk_sensitive$.__enclos_env__$private$solved <- TRUE
  tsk_sensitive$set_scorer(detect_pattern(
    "contains\\s+([A-Za-z])",
    case_sensitive = TRUE
  ))
  tsk_sensitive$score()
  expect_s3_class(tsk_sensitive$get_samples()$score, "factor")
  expect_true(is.ordered(tsk_sensitive$get_samples()$score))
  expect_equal(as.character(tsk_sensitive$get_samples()$score), c("I", "I"))

  all_df <- tibble::tibble(
    input = c("Question 1", "Question 2"),
    result = c(
      "Found colors red and blue",
      "Found colors green and yellow"
    ),
    target = c("red", "green")
  )

  tsk_all_false <- Task$new(
    dataset = all_df,
    solver = function() {},
    scorer = function() {}
  )
  tsk_all_false$.__enclos_env__$private$solved <- TRUE
  tsk_all_false$set_scorer(detect_pattern(
    "colors\\s+(\\w+)\\s+and\\s+(\\w+)",
    all = FALSE
  ))
  tsk_all_false$score()
  expect_s3_class(tsk_all_false$get_samples()$score, "factor")
  expect_true(is.ordered(tsk_all_false$get_samples()$score))
  expect_equal(as.character(tsk_all_false$get_samples()$score), c("C", "C"))

  tsk_all_true <- Task$new(
    dataset = all_df,
    solver = function() {},
    scorer = function() {}
  )
  tsk_all_true$.__enclos_env__$private$solved <- TRUE
  tsk_all_true$set_scorer(detect_pattern(
    "colors\\s+(\\w+)\\s+and\\s+(\\w+)",
    all = TRUE
  ))
  tsk_all_true$score()
  expect_s3_class(tsk_all_true$get_samples()$score, "factor")
  expect_true(is.ordered(tsk_all_true$get_samples()$score))
  expect_equal(as.character(tsk_all_true$get_samples()$score), c("I", "I"))
})

test_that("detect_exact works", {
  skip_on_cran()
  ex_task <- example_task(scored = FALSE)
  exact_df <- tibble::tibble(
    input = ex_task$get_samples()$input,
    result = c(ex_task$get_samples()$target[1], "non-matching output"),
    target = ex_task$get_samples()$target
  )

  tsk <- Task$new(
    dataset = exact_df,
    solver = function() {},
    scorer = function() {}
  )
  tsk$.__enclos_env__$private$solved <- TRUE
  tsk$set_scorer(detect_exact())
  tsk$score()

  expect_s3_class(tsk$get_samples()$score, "factor")
  expect_true(is.ordered(tsk$get_samples()$score))
  expect_equal(as.character(tsk$get_samples()$score), c("C", "I"))

  case_df <- tibble::tibble(
    input = c("Question 1", "Question 2"),
    result = c("ANSWER: C", "ANSWER: d"),
    target = c("ANSWER: c", "ANSWER: d")
  )

  tsk_insensitive <- Task$new(
    dataset = case_df,
    solver = function() {},
    scorer = function() {}
  )
  tsk_insensitive$.__enclos_env__$private$solved <- TRUE
  tsk_insensitive$set_scorer(detect_exact(case_sensitive = FALSE))
  tsk_insensitive$score()
  expect_s3_class(tsk_insensitive$get_samples()$score, "factor")
  expect_true(is.ordered(tsk_insensitive$get_samples()$score))
  expect_equal(as.character(tsk_insensitive$get_samples()$score), c("C", "C"))

  tsk_sensitive <- Task$new(
    dataset = case_df,
    solver = function() {},
    scorer = function() {}
  )
  tsk_sensitive$.__enclos_env__$private$solved <- TRUE
  tsk_sensitive$set_scorer(detect_exact(case_sensitive = TRUE))
  tsk_sensitive$score()
  expect_s3_class(tsk_sensitive$get_samples()$score, "factor")
  expect_true(is.ordered(tsk_sensitive$get_samples()$score))
  expect_equal(as.character(tsk_sensitive$get_samples()$score), c("I", "C"))
})

test_that("detect_answer works", {
  skip_on_cran()
  ex_task <- example_task(scored = FALSE)
  answer_df <- tibble::tibble(
    input = ex_task$get_samples()$input,
    result = paste("ANSWER:", ex_task$get_samples()$target),
    target = ex_task$get_samples()$target
  )

  tsk <- Task$new(
    dataset = answer_df,
    solver = function() {},
    scorer = function() {}
  )
  tsk$.__enclos_env__$private$solved <- TRUE
  tsk$set_scorer(detect_answer())
  tsk$score()

  expect_s3_class(tsk$get_samples()$score, "factor")
  expect_true(is.ordered(tsk$get_samples()$score))
  expect_equal(as.character(tsk$get_samples()$score), c("C", "C"))

  whitespace_df <- tibble::tibble(
    input = ex_task$get_samples()$input,
    result = paste("ANSWER: ", ex_task$get_samples()$target),
    target = ex_task$get_samples()$target
  )

  tsk_whitespace <- Task$new(
    dataset = whitespace_df,
    solver = function() {},
    scorer = function() {}
  )
  tsk_whitespace$.__enclos_env__$private$solved <- TRUE
  tsk_whitespace$set_scorer(detect_answer())
  tsk_whitespace$score()
  expect_s3_class(tsk_whitespace$get_samples()$score, "factor")
  expect_true(is.ordered(tsk_whitespace$get_samples()$score))
  expect_equal(as.character(tsk_whitespace$get_samples()$score), c("C", "C"))

  format_df <- tibble::tibble(
    input = c("Question 1", "Question 2"),
    result = c(
      "The solution is:\nANSWER: The Industrial Revolution",
      "ANSWER: C\nExplanation follows..."
    ),
    target = c(
      "The Industrial Revolution",
      "C"
    )
  )

  tsk_line <- Task$new(
    dataset = format_df,
    solver = function() {},
    scorer = function() {}
  )
  tsk_line$.__enclos_env__$private$solved <- TRUE
  tsk_line$set_scorer(detect_answer(format = "line"))
  tsk_line$score()
  expect_s3_class(tsk_line$get_samples()$score, "factor")
  expect_true(is.ordered(tsk_line$get_samples()$score))
  expect_equal(as.character(tsk_line$get_samples()$score), c("C", "I"))

  tsk_word <- Task$new(
    dataset = format_df,
    solver = function() {},
    scorer = function() {}
  )
  tsk_word$.__enclos_env__$private$solved <- TRUE
  tsk_word$set_scorer(detect_answer(format = "word"))
  tsk_word$score()
  expect_s3_class(tsk_word$get_samples()$score, "factor")
  expect_true(is.ordered(tsk_word$get_samples()$score))
  expect_equal(as.character(tsk_word$get_samples()$score), c("I", "C"))

  tsk_letter <- Task$new(
    dataset = format_df,
    solver = function() {},
    scorer = function() {}
  )
  tsk_letter$.__enclos_env__$private$solved <- TRUE
  tsk_letter$set_scorer(detect_answer(format = "letter"))
  tsk_letter$score()
  expect_s3_class(tsk_letter$get_samples()$score, "factor")
  expect_true(is.ordered(tsk_letter$get_samples()$score))
  expect_equal(as.character(tsk_letter$get_samples()$score), c("I", "C"))
})

test_that("detect scorers generate explanations", {
  skip_on_cran()

  simple_df <- tibble::tibble(
    input = c("Question 1", "Question 2"),
    result = c("4", "wrong answer"),
    target = c("4", "5")
  )

  tsk_includes <- Task$new(
    dataset = simple_df,
    solver = function() {},
    scorer = function() {}
  )
  tsk_includes$.__enclos_env__$private$solved <- TRUE
  tsk_includes$set_scorer(detect_includes())
  tsk_includes$score()

  expect_true("scorer_explanation" %in% names(tsk_includes$get_samples()))
  expect_equal(length(tsk_includes$get_samples()$scorer_explanation), 2)
  expect_match(
    tsk_includes$get_samples()$scorer_explanation[[1]],
    "Target '4' was found"
  )
  expect_match(
    tsk_includes$get_samples()$scorer_explanation[[2]],
    "was not found"
  )
})

# In general, we test these scorers completely offline and thus don't `log()`
# as that would require solver chats. Do a "live" test once to ensure that we
# don't assume scorer chats are available while logging. (#77)
test_that("vitals writes valid eval logs (basic, claude)", {
  vcr::local_cassette("scorer-detect-valid-log")
  key_get("OPENAI_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(ellmer::chat_openai(model = "gpt-4.1-nano")),
    scorer = detect_includes()
  )
  tsk$eval()

  expect_valid_log(tsk$log())
})

translate_to_messages <- function(chat) {
  turns <- chat$get_turns(include_system_prompt = TRUE)
  model <- chat$get_model()
  purrr::map(turns, translate_to_message, model = model)
}

translate_to_message <- function(turn, model) {
  role <- turn@role
  source <- if (role == "user") "input" else "generate"

  message <- list(id = generate_id())

  if (role == "system") {
    message$content <- turn@text
    message$role <- role
    return(message)
  } else if (role == "user") {
    if (
      length(turn@contents) == 1 &&
        inherits(turn@contents[[1]], "ellmer::ContentToolResult")
    ) {
      tool_result <- turn@contents[[1]]
      message$content <- collapse_tool_result(tool_result)
      message$role <- "tool"
      message$tool_call_id <- tool_result@request@id
      message$`function` <- tool_result@request@name
      return(message)
    } else {
      message$content <- message_content_from_turn(turn)
      message$source <- source
    }
  } else {
    message$content <- list(list(type = "text", text = turn@text))
    message$source <- source

    tool_requests <- purrr::keep(turn@contents, function(content) {
      inherits(content, "ellmer::ContentToolRequest")
    })

    if (length(tool_requests) > 0) {
      tool_calls <- lapply(tool_requests, function(req) {
        list(
          id = req@id,
          `function` = req@name,
          arguments = req@arguments
        )
      })

      message$tool_calls <- tool_calls
    }

    message$model <- model
  }

  message$role <- role

  message
}

collapse_tool_result <- function(tool_result) {
  if (!is.null(tool_result@error)) {
    return(as.character(tool_result@error))
  }

  value <- tool_result@value
  if (!is.list(value)) {
    value <- list(value)
  }

  content_list <- purrr::map(value, tool_result_value_to_content)

  all_text <- all(purrr::map_lgl(content_list, function(x) x$type == "text"))

  if (all_text) {
    return(paste0(
      purrr::map_chr(content_list, function(x) x$text),
      collapse = "\n"
    ))
  }

  content_list
}

tool_result_value_to_content <- function(x) {
  if (is.atomic(x)) {
    return(list(type = "text", text = paste0(x, collapse = "\n")))
  }

  if (inherits(x, "ellmer::ContentImageInline")) {
    media_type <- x@type %||% "image/png"
    return(list(
      type = "image",
      image = paste0("data:", media_type, ";base64,", x@data)
    ))
  }

  if (inherits(x, "ellmer::ContentText")) {
    return(list(type = "text", text = x@text))
  }

  list(type = "text", text = input_string(tibble::as_tibble(x)))
}

extract_format_from_media_type <- function(media_type, default) {
  parts <- strsplit(media_type, "/")[[1]]
  if (length(parts) == 2) {
    return(parts[2])
  }
  default
}

# translates a sample to Inspect's "event" data structure. The high-level
# structure is something like:
# - Initialization (via the user turn)
# - Tool registration (if applicable)
# - Solver
# - Scorer
#
# TODO: how is the tool registered if it's in the scorer?
translate_to_events <- function(sample, timestamps) {
  events <- translate_events_initialize(sample, timestamps = timestamps)
  events <- translate_events_tool_use(events, sample, timestamps = timestamps)
  events <- translate_events_solver(events, sample, timestamps = timestamps)
  events <- translate_events_scorer(events, sample, timestamps = timestamps)
  events
}

# higher-level helpers ------------------------------------------------------
translate_events_initialize <- function(sample, timestamps) {
  solver_chat <- sample$solver_chat[[1]]
  solver_turns <- solver_chat$get_turns()

  time_user <- timestamps$solve$started_at
  last_working_start <- attr(
    solver_turns[[length(solver_turns)]],
    "working_start"
  )

  events <- list()
  events <- c(events, create_init_begin_event(time_user))
  events <- c(
    events,
    create_sample_init_event(solver_turns[[1]], sample, time_user)
  )
  events <- c(
    events,
    create_init_end_event(time_user, working_start = last_working_start)
  )

  events
}

translate_events_tool_use <- function(events, sample, timestamps) {
  solver_chat <- sample$solver_chat[[1]]
  solver_turns <- solver_chat$get_turns()

  time_user <- timestamps$solve$started_at

  if (has_tool_calls(solver_turns)) {
    events <- c(
      events,
      create_use_tools_begin_event(
        time_user,
        working_start = attr(solver_turns[[1]], "working_start")
      )
    )
    events <- c(events, create_tool_state_event(time_user, solver_chat))
    events <- c(
      events,
      create_use_tools_end_event(
        time_user,
        working_start = attr(
          solver_turns[[length(solver_turns)]],
          "working_start"
        )
      )
    )
  }

  events
}

translate_events_solver <- function(events, sample, timestamps) {
  solver_chat <- sample$solver_chat[[1]]
  solver_turns <- solver_chat$get_turns()
  solver_turn <- solver_chat$last_turn()

  time_user <- timestamps$solve$started_at
  time_solver <- timestamps$solve$started_at

  # From here, the solver logging goes turn-by-turn. For each turn, log
  # the content from that turn as well as the "state" (e.g. previous response
  # history) at that time. Tool calls are logged with a single event, where the
  # "model" event preceding it functions doubly as a user event calling the tool.
  events <- c(events, create_solver_begin_event(time_user))

  for (i in seq_along(solver_turns)) {
    if (i == 1) {
      # First turn is the user query, skip it
      next
    }

    turn <- solver_turns[[i]]

    # For a tool response turn
    if (
      length(turn@contents) == 1 &&
        inherits(turn@contents[[1]], "ellmer::ContentToolResult")
    ) {
      tool_result <- turn@contents[[1]]
      events <- c(
        events,
        create_tool_event(
          turn,
          tool_result,
          timestamp = timestamps$solve$started_at
        )
      )
      next
    }

    if (turn@role == "assistant") {
      events <- c(
        events,
        create_model_event(turn, sample, timestamp = timestamps$solve$started_at)
      )
    }
  }

  events <- c(
    events,
    create_solver_end_event(time_solver, attr(solver_turn, "working_start"))
  )

  events
}

translate_events_scorer <- function(events, sample, timestamps = timestamps) {
  if ("scorer_chat" %in% names(sample)) {
    scorer_chat <- sample$scorer_chat[[1]]
    scorer_turns <- scorer_chat$get_turns()
    scorer_turn <- scorer_chat$last_turn()
    time_scorer <- timestamps$score$started_at

    if (has_tool_calls(scorer_turns)) {
      events <- c(
        events,
        create_use_tools_begin_event(
          time_scorer,
          attr(scorer_turns[[1]], "working_start") %||% 0,
          type = "scorer"
        )
      )
      events <- c(events, create_tool_state_event(time_scorer, scorer_chat))

      tool_result_turns <- purrr::keep(scorer_turns, function(turn) {
        length(turn@contents) == 1 &&
          inherits(turn@contents[[1]], "ellmer::ContentToolResult")
      })

      for (turn in tool_result_turns) {
        tool_result <- turn@contents[[1]]
        events <- c(
          events,
          create_tool_event(
            turn,
            tool_result,
            timestamp = timestamps$score$started_at
          )
        )
      }

      events <- c(
        events,
        create_use_tools_end_event(
          time_scorer,
          attr(scorer_turns[[length(scorer_turns)]], "working_start") %||% 0,
          type = "scorer"
        )
      )
    }

    events <- c(
      events,
      create_scorer_begin_event(
        time_scorer,
        attr(scorer_turn, "working_start")
      )
    )

    assistant_turns <- purrr::keep(
      scorer_turns,
      function(turn) turn@role == "assistant"
    )

    for (turn in assistant_turns) {
      has_tool_request <- any(vapply(
        turn@contents,
        inherits,
        logical(1),
        "ellmer::ContentToolRequest"
      ))

      if (identical(turn, scorer_turn) || has_tool_request) {
        events <- c(
          events,
          create_scoring_model_event(
            turn,
            sample,
            time_scorer
          )
        )
      }
    }

    events <- c(events, create_score_event(scorer_turn, sample, time_scorer))
    events <- c(
      events,
      create_scorer_end_event(
        time_scorer,
        attr(scorer_turn, "working_start")
      )
    )
  }

  events
}

# event-specific helpers ------------------------------------------------------
create_init_begin_event <- function(timestamp) {
  list(list(
    timestamp = events_timestamp(timestamp),
    working_start = 0,
    event = "step",
    action = "begin",
    name = "init"
  ))
}

create_sample_init_event <- function(turn, sample, timestamp) {
  user_message_id <- generate_id()
  message_content <- message_content_from_turn(turn)
  sample_input <- if (is_content_list(message_content)) {
    list(list(
      content = message_content,
      source = "input",
      role = "user"
    ))
  } else {
    message_content
  }

  list(list(
    timestamp = events_timestamp(timestamp),
    working_start = attr(turn, "working_start"),
    event = "sample_init",
    sample = list(
      input = sample_input,
      target = sample$target,
      id = sample$id
    ),
    state = list(
      messages = list(
        list(
          id = user_message_id,
          content = message_content,
          source = "input",
          role = "user"
        )
      ),
      tools = list(),
      tool_choice = NULL,
      store = c(),
      output = list(
        model = sample$solver_chat[[1]]$get_model(),
        choices = list()
      ),
      completed = FALSE,
      metadata = c()
    )
  ))
}

create_init_end_event <- function(timestamp, working_start) {
  list(list(
    timestamp = events_timestamp(timestamp),
    working_start = working_start,
    event = "step",
    action = "end",
    name = "init"
  ))
}

create_use_tools_begin_event <- function(
  timestamp,
  working_start,
  type = "solver"
) {
  list(list(
    timestamp = events_timestamp(timestamp),
    working_start = working_start,
    event = "step",
    action = "begin",
    type = type,
    name = "use_tools"
  ))
}

create_tool_state_event <- function(timestamp, chat) {
  tools_list <- list()

  if (length(chat$get_tools()) > 0) {
    tool_defs <- chat$get_tools()

    for (i in seq_along(tool_defs)) {
      tool_def <- tool_defs[[i]]
      tool_name <- names(tool_defs)[i]

      tool_info <- list(
        op = "add",
        path = paste0("/tools/", i - 1),
        value = list(
          name = tool_name,
          description = tool_def@description,
          parameters = list(
            type = "object",
            properties = c(),
            required = list(),
            additionalProperties = FALSE
          )
        )
      )

      tools_list <- append(tools_list, list(tool_info))
    }
  }

  tools_list <- append(
    tools_list,
    list(list(
      op = "replace",
      path = "/tool_choice",
      value = "auto"
    ))
  )

  list(list(
    timestamp = events_timestamp(timestamp),
    working_start = attr(chat$get_turns()[[1]], "working_start"),
    event = "state",
    changes = tools_list
  ))
}

create_use_tools_end_event <- function(
  timestamp,
  working_start,
  type = "solver"
) {
  list(list(
    timestamp = events_timestamp(timestamp),
    working_start = working_start,
    event = "step",
    action = "end",
    type = type,
    name = "use_tools"
  ))
}

create_tool_event <- function(turn, tool_result, timestamp) {
  list(list(
    timestamp = events_timestamp(timestamp),
    working_start = attr(turn, "working_start"),
    event = "tool",
    type = "function",
    id = tool_result@request@id,
    `function` = tool_result@request@name,
    arguments = tool_result@request@arguments,
    result = if (!is.null(tool_result@error)) {
      as.character(tool_result@error)
    } else {
      collapse_tool_result(tool_result)
    },
    events = list(),
    completed = events_timestamp(timestamp),
    working_time = attr(turn, "working_time")
  ))
}

create_solver_begin_event <- function(timestamp) {
  list(list(
    timestamp = events_timestamp(timestamp),
    working_start = 0,
    event = "step",
    action = "begin",
    type = "solver",
    name = "generate"
  ))
}

create_model_event <- function(turn, sample, timestamp) {
  user_message_id <- generate_id()
  solver_chat <- sample$solver_chat[[1]]

  turns <- solver_chat$get_turns()
  previous_turns <- list()

  for (j in seq_along(turns)) {
    if (identical(turns[[j]], turn)) {
      break
    }
    previous_turns[[length(previous_turns) + 1]] <- turns[[j]]
  }

  input_messages <- lapply(previous_turns, function(prev_turn) {
    if (prev_turn@role == "user") {
      if (
        length(prev_turn@contents) == 1 &&
          inherits(prev_turn@contents[[1]], "ellmer::ContentToolResult")
      ) {
        tool_result <- prev_turn@contents[[1]]
        return(list(
          id = generate_id(),
          content = if (!is.null(tool_result@error)) {
            as.character(tool_result@error)
          } else {
            collapse_tool_result(tool_result)
          },
          role = "tool",
          tool_call_id = tool_result@request@id,
          `function` = tool_result@request@name
        ))
      } else {
        return(list(
          id = generate_id(),
          content = message_content_from_turn(prev_turn),
          source = "input",
          role = "user"
        ))
      }
    } else {
      message <- list(
        id = generate_id(),
        content = list(list(type = "text", text = prev_turn@text)),
        source = "generate",
        role = "assistant"
      )

      tool_requests <- purrr::keep(prev_turn@contents, function(content) {
        inherits(content, "ellmer::ContentToolRequest")
      })

      if (length(tool_requests) > 0) {
        tool_calls <- lapply(tool_requests, function(req) {
          list(
            id = req@id,
            `function` = req@name,
            arguments = req@arguments
          )
        })

        message$tool_calls <- tool_calls
      }

      return(message)
    }
  })

  has_tool_calls_in_turn <- any(sapply(turn@contents, function(content) {
    inherits(content, "ellmer::ContentToolRequest")
  }))

  tool_calls_list <- list()
  if (has_tool_calls_in_turn) {
    tool_requests <- purrr::keep(turn@contents, function(content) {
      inherits(content, "ellmer::ContentToolRequest")
    })

    tool_calls_list <- lapply(tool_requests, function(req) {
      list(
        id = req@id,
        `function` = req@name,
        arguments = req@arguments
      )
    })
  }

  stop_reason <- ifelse(has_tool_calls_in_turn, "tool_calls", "stop")

  tools_list <- chat_tools_list(solver_chat)

  output_message <- list(
    id = generate_id(),
    content = list(list(type = "text", text = turn@text)),
    source = "generate",
    role = "assistant"
  )

  if (has_tool_calls_in_turn) {
    output_message$tool_calls <- tool_calls_list
  }

  output_message$model <- solver_chat$get_model()

  request_messages <- lapply(input_messages, function(msg) {
    if (msg$role == "tool") {
      return(list(
        role = "user",
        content = list(list(
          tool_use_id = msg$tool_call_id,
          type = "tool_result",
          content = if (is.character(msg$content)) {
            list(list(type = "text", text = msg$content))
          } else {
            # Handle tool results that may contain image objects
            lapply(msg$content, function(item) {
              if (
                is.list(item) &&
                  identical(item$type, "image") &&
                  !is.null(item$source)
              ) {
                # Convert image object to ContentImage format
                list(
                  type = "image",
                  image = paste0(
                    "data:",
                    item$source$media_type,
                    ";base64,",
                    item$source$data
                  )
                )
              } else {
                item
              }
            })
          },
          # This depends specifically on previous helpers using
          # `as_character()` on conditions to extract error messages
          is_error = if (is.character(msg$content)) {
            grepl("Error in", msg$content)
          } else {
            FALSE
          }
        ))
      ))
    } else if (msg$role == "user") {
      return(list(
        role = "user",
        content = ensure_content_list(msg$content)
      ))
    } else if (msg$role == "assistant") {
      if ("tool_calls" %in% names(msg)) {
        tool_use_elements <- lapply(msg$tool_calls, function(tc) {
          list(
            type = "tool_use",
            id = tc$id,
            name = tc$`function`,
            input = tc$arguments
          )
        })

        combined_content <- c(
          list(list(type = "text", text = msg$content[[1]]$text)),
          tool_use_elements
        )

        return(list(
          role = "assistant",
          content = combined_content
        ))
      } else {
        # Handle content that may contain image objects
        processed_content <- if (is.list(msg$content)) {
          lapply(msg$content, function(item) {
            if (
              is.list(item) &&
                identical(item$type, "image") &&
                !is.null(item$source)
            ) {
              # Convert image object to ContentImage format
              list(
                type = "image",
                image = paste0(
                  "data:",
                  item$source$media_type,
                  ";base64,",
                  item$source$data
                )
              )
            } else {
              item
            }
          })
        } else {
          msg$content
        }

        return(list(
          role = "assistant",
          content = processed_content
        ))
      }
    }
  })

  list(list(
    timestamp = events_timestamp(timestamp),
    working_start = attr(turn, "working_start"),
    event = "model",
    model = solver_chat$get_model(),
    input = input_messages,
    tools = tools_list,
    tool_choice = if (length(tools_list) > 0) "auto" else "none",
    config = drop_nulls(list(
      max_tokens = solver_chat$get_provider()@params$max_tokens
    )),
    output = list(
      model = solver_chat$get_model(),
      choices = list(
        list(
          message = output_message,
          stop_reason = stop_reason
        )
      ),
      usage = turn_tokens(turn),
      time = attr(turn, "working_time")
    ),
    call = list(
      request = drop_nulls(list(
        messages = request_messages,
        tools = tools_list,
        tool_choice = if (length(tools_list) > 0) {
          list(type = "auto")
        } else {
          "none"
        },
        model = solver_chat$get_model(),
        max_tokens = solver_chat$get_provider()@params$max_tokens,
        extra_headers = list(
          `x-irid` = generate_id()
        )
      )),
      response = list(
        id = paste0("msg_", generate_id()),
        content = if (has_tool_calls_in_turn) {
          c(
            list(list(
              citations = NULL,
              text = turn@text,
              type = "text"
            )),
            lapply(tool_calls_list, function(tc) {
              list(
                id = tc$id,
                input = tc$arguments,
                name = tc$`function`,
                type = "tool_use"
              )
            })
          )
        } else {
          list(list(
            citations = NULL,
            text = turn@text,
            type = "text"
          ))
        },
        model = solver_chat$get_model(),
        role = "assistant",
        stop_reason = if (has_tool_calls_in_turn) "tool_use" else "end_turn",
        stop_sequence = NULL,
        type = "message",
        usage = turn_tokens(turn)
      ),
      time = attr(turn, "working_time")
    ),
    completed = events_timestamp(timestamp),
    working_time = attr(turn, "working_time")
  ))
}

chat_tools_list <- function(chat) {
  if (length(chat$get_tools()) == 0) {
    return(list())
  }

  tools <- chat$get_tools()
  lapply(seq_along(tools), function(i) {
    tool <- tools[[i]]
    tool_name <- names(tools)[i]

    list(
      name = tool_name,
      description = tool@description,
      parameters = list(
        type = "object",
        properties = c(),
        required = list(),
        additionalProperties = FALSE
      )
    )
  })
}

create_solver_end_event <- function(timestamp, working_start) {
  list(list(
    timestamp = events_timestamp(timestamp),
    working_start = working_start,
    event = "step",
    action = "end",
    type = "solver",
    name = "generate"
  ))
}

create_scorer_begin_event <- function(timestamp, working_start) {
  list(list(
    timestamp = events_timestamp(timestamp),
    working_start = working_start,
    event = "step",
    action = "begin",
    type = "scorer",
    name = "model_graded_qa"
  ))
}

create_scoring_model_event <- function(turn, sample, timestamp) {
  scorer_chat <- sample$scorer_chat[[1]]
  turns <- scorer_chat$get_turns()
  previous_turns <- list()

  for (j in seq_along(turns)) {
    if (identical(turns[[j]], turn)) {
      break
    }
    previous_turns[[length(previous_turns) + 1]] <- turns[[j]]
  }

  input_messages <- lapply(previous_turns, function(prev_turn) {
    if (prev_turn@role == "user") {
      if (
        length(prev_turn@contents) == 1 &&
          inherits(prev_turn@contents[[1]], "ellmer::ContentToolResult")
      ) {
        tool_result <- prev_turn@contents[[1]]
        return(list(
          id = generate_id(),
          content = if (!is.null(tool_result@error)) {
            as.character(tool_result@error)
          } else {
            collapse_tool_result(tool_result)
          },
          role = "tool",
          tool_call_id = tool_result@request@id,
          `function` = tool_result@request@name
        ))
      } else {
        return(list(
          id = generate_id(),
          content = message_content_from_turn(prev_turn),
          source = "input",
          role = "user"
        ))
      }
    } else {
      message <- list(
        id = generate_id(),
        content = list(list(type = "text", text = prev_turn@text)),
        source = "generate",
        role = "assistant"
      )

      tool_requests <- purrr::keep(prev_turn@contents, function(content) {
        inherits(content, "ellmer::ContentToolRequest")
      })

      if (length(tool_requests) > 0) {
        tool_calls <- lapply(tool_requests, function(req) {
          list(
            id = req@id,
            `function` = req@name,
            arguments = req@arguments
          )
        })

        message$tool_calls <- tool_calls
      }

      return(message)
    }
  })

  has_tool_calls_in_turn <- any(sapply(turn@contents, function(content) {
    inherits(content, "ellmer::ContentToolRequest")
  }))

  tool_calls_list <- list()
  if (has_tool_calls_in_turn) {
    tool_requests <- purrr::keep(turn@contents, function(content) {
      inherits(content, "ellmer::ContentToolRequest")
    })

    tool_calls_list <- lapply(tool_requests, function(req) {
      list(
        id = req@id,
        `function` = req@name,
        arguments = req@arguments
      )
    })
  }

  stop_reason <- ifelse(has_tool_calls_in_turn, "tool_calls", "stop")
  tools_list <- chat_tools_list(scorer_chat)

  output_message <- list(
    id = generate_id(),
    content = list(list(type = "text", text = turn@text)),
    source = "generate",
    role = "assistant"
  )

  if (has_tool_calls_in_turn) {
    output_message$tool_calls <- tool_calls_list
  }

  output_message$model <- scorer_chat$get_model()

  request_messages <- lapply(input_messages, function(msg) {
    if (msg$role == "tool") {
      return(list(
        role = "user",
        content = list(list(
          tool_use_id = msg$tool_call_id,
          type = "tool_result",
          content = if (is.character(msg$content)) {
            list(list(type = "text", text = msg$content))
          } else {
            lapply(msg$content, function(item) {
              if (
                is.list(item) &&
                  identical(item$type, "image") &&
                  !is.null(item$source)
              ) {
                list(
                  type = "image",
                  image = paste0(
                    "data:",
                    item$source$media_type,
                    ";base64,",
                    item$source$data
                  )
                )
              } else {
                item
              }
            })
          },
          is_error = if (is.character(msg$content)) {
            grepl("Error in", msg$content)
          } else {
            FALSE
          }
        ))
      ))
    } else if (msg$role == "user") {
      return(list(
        role = "user",
        content = ensure_content_list(msg$content)
      ))
    } else if (msg$role == "assistant") {
      if ("tool_calls" %in% names(msg)) {
        tool_use_elements <- lapply(msg$tool_calls, function(tc) {
          list(
            type = "tool_use",
            id = tc$id,
            name = tc$`function`,
            input = tc$arguments
          )
        })

        combined_content <- c(
          list(list(type = "text", text = msg$content[[1]]$text)),
          tool_use_elements
        )

        return(list(
          role = "assistant",
          content = combined_content
        ))
      } else {
        processed_content <- if (is.list(msg$content)) {
          lapply(msg$content, function(item) {
            if (
              is.list(item) &&
                identical(item$type, "image") &&
                !is.null(item$source)
            ) {
              list(
                type = "image",
                image = paste0(
                  "data:",
                  item$source$media_type,
                  ";base64,",
                  item$source$data
                )
              )
            } else {
              item
            }
          })
        } else {
          msg$content
        }

        return(list(
          role = "assistant",
          content = processed_content
        ))
      }
    }
  })

  list(list(
    timestamp = events_timestamp(timestamp),
    working_start = attr(turn, "working_start"),
    event = "model",
    model = scorer_chat$get_model(),
    input = input_messages,
    tools = tools_list,
    tool_choice = if (length(tools_list) > 0) "auto" else "none",
    config = drop_nulls(list(
      max_tokens = scorer_chat$get_provider()@params$max_tokens
    )),
    output = list(
      model = scorer_chat$get_model(),
      choices = list(
        list(
          message = output_message,
          stop_reason = stop_reason
        )
      ),
      usage = turn_tokens(turn),
      time = attr(turn, "working_time")
    ),
    call = list(
      request = drop_nulls(list(
        messages = request_messages,
        tools = tools_list,
        tool_choice = if (length(tools_list) > 0) {
          list(type = "auto")
        } else {
          "none"
        },
        model = scorer_chat$get_model(),
        max_tokens = scorer_chat$get_provider()@params$max_tokens,
        extra_headers = list(
          `x-irid` = generate_id()
        )
      )),
      response = list(
        id = paste0("msg_", generate_id()),
        content = if (has_tool_calls_in_turn) {
          c(
            list(list(
              citations = NULL,
              text = turn@text,
              type = "text"
            )),
            lapply(tool_calls_list, function(tc) {
              list(
                id = tc$id,
                input = tc$arguments,
                name = tc$`function`,
                type = "tool_use"
              )
            })
          )
        } else {
          list(list(
            citations = NULL,
            text = turn@text,
            type = "text"
          ))
        },
        model = scorer_chat$get_model(),
        role = "assistant",
        stop_reason = if (has_tool_calls_in_turn) "tool_use" else "end_turn",
        stop_sequence = NULL,
        type = "message",
        usage = turn_tokens(turn),
        time = attr(turn, "working_time")
      )
    ),
    completed = events_timestamp(timestamp),
    working_time = attr(turn, "working_time")
  ))
}

create_score_event <- function(turn, sample, timestamp) {
  scorer_user_turn <- sample$scorer_chat[[1]]$get_turns()[[1]]

  list(list(
    timestamp = events_timestamp(timestamp),
    working_start = attr(turn, "working_start"),
    event = "score",
    score = list(
      value = sample$score,
      answer = as.character(sample$result),
      explanation = turn@text,
      metadata = list(
        grading = list(
          list(
            id = generate_id(),
            content = scorer_user_turn@text,
            role = "user"
          ),
          list(
            id = generate_id(),
            content = list(
              list(
                type = "text",
                text = turn@text
              )
            ),
            source = "generate",
            role = "assistant"
          )
        )
      )
    ),
    target = sample$target,
    intermediate = FALSE
  ))
}

create_scorer_end_event <- function(timestamp, working_start) {
  list(list(
    timestamp = events_timestamp(timestamp),
    working_start = working_start,
    event = "step",
    action = "end",
    type = "scorer",
    name = "model_graded_qa"
  ))
}

drop_nulls <- function(x) {
  x[!vapply(x, is.null, logical(1))]
}

# misc helpers -------------------------------------------------------------
# the events log the timestamp a bit differently than everywhere
# else in the log
events_timestamp <- function(time) {
  sub(
    pattern = "(\\d{2})(\\d{2})$",
    replacement = "\\1:\\2",
    x = format(time, "%Y-%m-%dT%H:%M:%OS6%z")
  )
}

turn_tokens <- function(turn) {
  tokens_io <- turn@tokens

  input_tokens <- if (is.na(tokens_io[1])) 0L else as.integer(tokens_io[1])
  output_tokens <- if (is.na(tokens_io[2])) 0L else as.integer(tokens_io[2])

  list(
    input_tokens = input_tokens,
    input_tokens_cache_write = 0L,
    input_tokens_cache_read = 0L,
    output_tokens = output_tokens
  )
}

turn_duration <- function(turn) {
  duration <- 0
  if (inherits(turn, "ellmer::AssistantTurn")) {
    if (!is.na(turn@duration)) {
      duration <- turn@duration
    }
  }
  duration
}

# log working_time values by extracting them from the Turn @duration slots (#115).
# `working_time` is the duration of the turn in seconds. User turns have NA duration,
# assistant turns have the actual request duration as measured by httr2.
add_working_times_to_turns <- function(chat, which, timestamps, n) {
  turns <- chat$get_turns()

  if (length(turns) < 2) {
    return(chat)
  }

  for (i in seq_along(turns)) {
    attr(turns[[i]], "working_time") <- turn_duration(turns[[i]])
  }

  chat$set_turns(turns)

  chat
}

# log working_start values by accumulating turn durations (#115).
# `working_start` is the clock time when a turn started minus the clock time
# when the first chat started, in seconds. This is computed by accumulating
# the @duration values from previous turns.
add_working_start_to_turns <- function(chats, which, timestamps) {
  current_working_start <- 0

  for (i in seq_along(chats)) {
    chat <- chats[[i]]
    chat_turns <- chat$get_turns()

    for (j in seq_along(chat_turns)) {
      attr(chat_turns[[j]], "working_start") <- current_working_start
      current_working_start <- current_working_start + turn_duration(chat_turns[[j]])
    }

    chat$set_turns(chat_turns)
    chats[[i]] <- chat
  }

  chats
}

#' Model-based scoring
#'
#' @description
#' Model-based scoring makes use of a model to score output from a solver.
#'
#' * `model_graded_qa()` scores how well a solver answers a question/answer task.
#' * `model_graded_fact()` determines whether a solver includes a given fact
#' in its response.
#'
#' The two scorers are quite similar in their implementation, but use a different
#' default `template` to evaluate correctness.
#'
#' @param template Grading template to use--a `glue()` string which will take
#' substitutions `input`, `answer`, `criterion`, `instructions`.
#' @param instructions Grading instructions. If provided, this completely
#' replaces the default instructions, which specify e.g. how the grader should
#' format its output (e.g. `"GRADE: C"`).
#' @param grade_pattern A regex pattern to extract the final grade from the
#' judge model's response.
#' @param partial_credit Whether to allow partial credit.
#' @param scorer_chat An ellmer chat used to grade the model output, e.g.
#' [ellmer::chat_claude()].
#'
#' @returns
#' A function that will grade model responses according to the given instructions.
#' See [Task]'s `scorer` argument for a description of the returned function.
#' The functions that `model_graded_qa()` and `model_graded_fact()` output
#' can be passed directly to `$eval()`.
#'
#' See the documentation for the `scorer` argument in [Task] for more
#' information on the return type.
#'
#' @seealso [scorer_detect] for string detection-based scoring.
#'
#' @examples
#' # Quality assurance -----------------------------
#' if (!identical(Sys.getenv("ANTHROPIC_API_KEY"), "")) {
#'   # set the log directory to a temporary directory
#'   withr::local_envvar(VITALS_LOG_DIR = withr::local_tempdir())
#'
#'   library(ellmer)
#'   library(tibble)
#'
#'   simple_addition <- tibble(
#'     input = c("What's 2+2?", "What's 2+3?"),
#'     target = c("4", "5")
#'   )
#'
#'   tsk <- Task$new(
#'     dataset = simple_addition,
#'     solver = generate(solver_chat = chat_claude(model = "claude-sonnet-4-5-20250929")),
#'     scorer = model_graded_qa()
#'   )
#'
#'   tsk$eval()
#' }
#'
#' # Factual response -------------------------------
#' if (!identical(Sys.getenv("ANTHROPIC_API_KEY"), "")) {
#'   # set the log directory to a temporary directory
#'   withr::local_envvar(VITALS_LOG_DIR = withr::local_tempdir())
#'
#'   library(ellmer)
#'   library(tibble)
#'
#'   r_history <- tibble(
#'     input = c(
#'       "Who created the R programming language?",
#'       "In what year was version 1.0 of R released?"
#'     ),
#'     target = c("Ross Ihaka and Robert Gentleman.", "2000.")
#'   )
#'
#'   tsk <- Task$new(
#'     dataset = r_history,
#'     solver = generate(solver_chat = chat_claude(model = "claude-sonnet-4-5-20250929")),
#'     scorer = model_graded_fact()
#'   )
#'
#'   tsk$eval()
#' }
#'
#' @name scorer_model
#' @export
model_graded_qa <- function(
  template = NULL,
  instructions = NULL,
  grade_pattern = "(?i)GRADE\\s*:\\s*([CPI])(.*)$",
  partial_credit = FALSE,
  scorer_chat = NULL
) {
  ch <- scorer_chat

  function(samples, ..., scorer_chat = ch) {
    model_graded_qa_impl(
      samples = samples,
      template = template,
      instructions = instructions,
      grade_pattern = grade_pattern,
      partial_credit = partial_credit,
      scorer_chat = scorer_chat
    )
  }
}

model_graded_qa_impl <- function(
  samples,
  template = NULL,
  instructions = NULL,
  grade_pattern = "(?i)GRADE\\s*:\\s*([CPI])(.*)$",
  partial_credit = FALSE,
  scorer_chat = NULL
) {
  template <- template %||% qa_default_template()
  instructions <- instructions %||% qa_default_instructions(partial_credit)

  prompts <- purrr::map_chr(seq_len(nrow(samples)), function(i) {
    qa_format_prompt(
      template,
      samples$input[i],
      samples$result[i],
      samples$target[i],
      instructions
    )
  })

  if (is.null(scorer_chat)) {
    scorer_chat <- solver_chat(samples[1, ])
  }

  scorer_chat <- scorer_chat$clone()
  responses <- ellmer::parallel_chat(scorer_chat, as.list(prompts))

  grades <- purrr::map_chr(responses, function(response_chat) {
    response_text <- response_chat$last_turn()@text
    qa_extract_grade(response_text, grade_pattern, partial_credit)
  })

  scores <- process_grades(grades, partial_credit)

  metadata <- purrr::map(seq_along(prompts), function(i) {
    list(
      prompt = prompts[i],
      response = responses[[i]]$last_turn()@text,
      grade_pattern = grade_pattern
    )
  })

  list(
    score = scores,
    scorer_chat = responses,
    scorer_metadata = metadata
  )
}

qa_format_prompt <- function(template, input, output, target, instructions) {
  glue::glue(
    template,
    input = input,
    answer = output,
    criterion = target,
    instructions = instructions
  )
}

qa_extract_grade <- function(response, pattern, partial_credit = FALSE) {
  grade_letter <- regmatches(
    response,
    regexec(pattern, response, perl = TRUE)
  )[[1]][2]

  if (is.na(grade_letter)) {
    return(NA)
  }

  toupper(grade_letter)
}

process_grades <- function(grades, partial_credit) {
  unique_grades <- unique(grades)
  is_ipc_grading <- (any(c("I", "C") %in% unique_grades) &&
    all(unique_grades %in% c("I", "P", "C")))

  if (!is_ipc_grading) {
    return(grades)
  }

  if (partial_credit) {
    factor(
      grades,
      levels = c("I", "P", "C"),
      ordered = TRUE
    )
  } else {
    factor(
      grades,
      levels = c("I", "C"),
      ordered = TRUE
    )
  }
}

qa_default_instructions <- function(partial_credit = FALSE) {
  partial_letter <- if (partial_credit) ", P, or " else " or "
  partial_prompt <- if (partial_credit) {
    '"P" for partially correct answers,'
  } else {
    ""
  }

  glue::glue(
    "After assessing the submitted answer, reply with 'GRADE: $LETTER' where
    LETTER is one of C{partial_letter}I.
    Please choose ONE option: either 'C' for correct answers, {partial_prompt}
    or 'I' for incorrect answers.
    First **briefly** explain your reasoning, then end with GRADE: $LETTER.
    Do not format the grading string and do not include any punctuation or
    exposition after it."
  )
}

# based on https://github.com/UKGovernmentBEIS/inspect_ai/blob/f60a6ff8502fa43ef1b9a7a905eab2a21608f1c1/src/inspect_ai/scorer/_model.py#L206
qa_default_template <- function() {
  "You are assessing a submitted answer on a given task based on a criterion.
  [BEGIN DATA]
  ***
  [Task]: {input}
  ***
  [Submission]: {answer}
  ***
  [Criterion]: {criterion}
  ***
  [END DATA]
  Does the submission meet the criterion?
  {instructions}"
}

#' @rdname scorer_model
#' @export
model_graded_fact <- function(
  template = NULL,
  instructions = NULL,
  grade_pattern = "(?i)GRADE\\s*:\\s*([CPI])(.*)$",
  partial_credit = FALSE,
  scorer_chat = NULL
) {
  ch <- scorer_chat

  function(samples, scorer_chat = ch) {
    model_graded_qa_impl(
      samples = samples,
      template = template %||% fact_default_template(),
      instructions = instructions,
      grade_pattern = grade_pattern,
      partial_credit = partial_credit,
      scorer_chat = scorer_chat
    )
  }
}

fact_default_template <- function() {
  "You are comparing a submitted answer to an expert answer on a given question.
  [BEGIN DATA]
  ************
  [Question]: {input}
  ************
  [Expert]: {criterion}
  ************
  [Submission]: {answer}
  ************
  [END DATA]
  Compare the factual content of the submitted answer with the expert answer.
  Ignore any differences in style, grammar, or punctuation.
  Does the submission contain the content in the expert answer?
  {instructions}"
}

# validating log files (for dev use only) --------------------------------------
# invokes Inspect's pydantic models on an eval log file so that
# we can ensure we're writing files that are compatible with the
# viewer.
# first-time setup (optional) (requires Python >=3.10):
#   reticulate::install_python(version = '3.14.0')
#   # restart R session
#   reticulate::virtualenv_create(envname = "vitals-venv")
#   reticulate::py_install("inspect_ai", envname = "vitals-venv")
python_cmd <- function() {
  if (!is_installed("reticulate")) {
    return("python3")
  }
  if (asNamespace("reticulate")$py_available()) {
    return(asNamespace("reticulate")$py_config()$python)
  }
  asNamespace("reticulate")$py_require(c("inspect-ai", "pydantic"))
  tryCatch(
    asNamespace("reticulate")$py_config()$python,
    error = function(e) "python3"
  )
}

expect_valid_log <- local({
  .pydantic_skip_status <- if (
    !interactive() && !isTRUE(as.logical(Sys.getenv("NOT_CRAN", "false")))
  ) {
    "On CRAN."
  } else if (
    system2(python_cmd(), "--version", stdout = FALSE, stderr = FALSE) != 0
  ) {
    "Python is not available"
  } else if (
    system2(
      python_cmd(),
      "-c 'import inspect_ai'",
      stdout = FALSE,
      stderr = FALSE
    ) !=
      0
  ) {
    "inspect_ai Python module is not available"
  } else if (
    system2(
      python_cmd(),
      "-c 'import pydantic'",
      stdout = FALSE,
      stderr = FALSE
    ) !=
      0
  ) {
    "pydantic Python module is not available"
  } else if (
    !file.exists(system.file("test/validate_log.py", package = "vitals"))
  ) {
    "Python validation script not found."
  } else {
    NULL
  }

  function(x) {
    if (!is.null(.pydantic_skip_status)) {
      skip(.pydantic_skip_status)
    }

    if (!file.exists(x)) {
      skip(paste0("Log file ", x, " does not exist."))
    }

    result <- system2(
      python_cmd(),
      args = c(system.file("test/validate_log.py", package = "vitals"), x),
      stdout = TRUE,
      stderr = TRUE
    )
    status <- attr(result, "status")

    expect(
      is.null(status) || status == 0,
      paste0(
        c("The generated log did not pass the pydantic model: ", result),
        collapse = "\n"
      )
    )
  }
})

is_positron <- function() {
  Sys.getenv("POSITRON") == "1"
}

is_testing <- function() {
  identical(Sys.getenv("TESTTHAT"), "true")
}

is_snapshot <- function() {
  identical(Sys.getenv("TESTTHAT_IS_SNAPSHOT"), "true")
}

is_replaying <- function() {
  as.logical(Sys.getenv("VCR_IS_REPLAYING", "FALSE"))
}

key_get <- function(name, error_call = caller_env()) {
  val <- Sys.getenv(name)
  if (!identical(val, "")) {
    invisible(val)
  } else {
    if (is_replaying()) {
      ""
    } else if (is_testing()) {
      testthat::skip(sprintf("%s env var is not configured", name))
    } else {
      cli::cli_abort("Can't find env var {.code {name}}.", call = error_call)
    }
  }
}

key_exists <- function(name) {
  !identical(Sys.getenv(name), "")
}

# ad-hoc check functions
check_inherits <- function(x, cls, x_arg = caller_arg(x), call = caller_env()) {
  if (!inherits(x, cls)) {
    cli::cli_abort(
      "{.arg {x_arg}} must be a {.cls {cls}}, not {.obj_type_friendly {x}}",
      call = call
    )
  }

  invisible()
}

check_log_dir <- function(x, call = caller_env()) {
  if (is.na(x)) {
    cli::cli_warn(c(
      "!" = "{.pkg vitals} could not find a log directory; evaluation log
             files will be written to a temporary directory.",
      "i" = 'Set a log directory with e.g.
             {.code vitals::vitals_log_dir_set("./logs")}, 
             perhaps in {.file ~/.Rprofile}, to quiet this warning.'
    ))
  }
  invisible(x)
}

# miscellaneous ---------
solver_chat <- function(sample) {
  solver <- sample$solver_chat[[1]]
  res <- solver$clone()
  res$set_turns(list())
  res$set_system_prompt(NULL)
  res
}

interactive <- NULL

regenerate_example_objects <- function() {
  source("inst/regenerate-example-objects.R")
}

accuracy <- function(...) {
  mean(...) * 100
}

is_old_task_id_format <- function(task_id) {
  !is.null(task_id) && !grepl("/", task_id, fixed = TRUE)
}

warn_old_task_id_format <- function() {
  cli::cli_warn(
    c(
      "Log directory contains logs with old-format task IDs.",
      "i" = "The log viewer may collapse different evals as
             {.emph retried} logs.",
      "i" = "Click {.emph Show retried logs} in the viewer to see all logs.",
      "i" = "To silence this warning, regenerate logs with the current
             {.pkg vitals} version."
    ),
    .frequency = "once",
    .frequency_id = "vitals_old_task_id_format"
  )
}

test_that("is_positron works", {
  # default when unset
  withr::local_envvar(POSITRON = "")
  expect_false(is_positron())

  # explicit values
  withr::local_envvar(POSITRON = "1")
  expect_true(is_positron())

  withr::local_envvar(POSITRON = "0")
  expect_false(is_positron())
})

test_that("is_testing works", {
  # reflects current testing status
  expect_equal(is_testing(), identical(Sys.getenv("TESTTHAT"), "true"))

  # can be overridden
  withr::local_envvar(TESTTHAT = "true")
  expect_true(is_testing())

  withr::local_envvar(TESTTHAT = "false")
  expect_false(is_testing())
})

test_that("check_inherits works", {
  # errors informatively with wrong class
  expect_snapshot(check_inherits(1, "list"), error = TRUE)
  expect_snapshot(check_inherits(1, "list", x_arg = "my_arg"), error = TRUE)

  # no condition otherwise
  expect_null(check_inherits(list(), "list"))
  expect_null(check_inherits(mtcars, c("data.frame", "list")))
})

test_that("solver_chat works", {
  skip_on_cran()
  example_sample <- example_task()$get_samples()[1, , drop = FALSE]

  res <- solver_chat(example_sample)
  expect_s3_class(res, "Chat")
  expect_equal(res$get_turns(), list())
})

test_that("check_log_dir warns informatively", {
  withr::local_envvar(VITALS_LOG_DIR = NA)
  expect_snapshot(
    res <- Task$new(
      tibble(input = 1, target = 1),
      function() {},
      function() {}
    )
  )
})

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @import rlang
#
# quiet "All declared Imports should be used" NOTE
#' @importFrom R6 R6Class
#' @importFrom ellmer content_image_url
#' @importFrom tibble tibble
#' @importFrom withr defer
## usethis namespace: end
NULL

utils::globalVariables(c(
  "any_of",
  "capture.output",
  "chat_claude",
  "everything",
  "head",
  "input",
  "input_before",
  "model",
  "output",
  "output_before",
  "path",
  "price",
  "price_before",
  "provider",
  "qnorm",
  "res",
  "setNames",
  "str",
  "timestamp",
  "var"
))

#' Concatenate task samples for analysis
#'
#' @description
#' Combine multiple [Task] objects into a single tibble for comparison.
#'
#' This function takes multiple (optionally named) [Task] objects and row-binds
#' their `$get_samples()` together, adding a `task` column to identify the source of each
#' row. The resulting tibble nests additional columns into a `metadata` column
#' and is ready for further analysis.
#'
#' @param ... `Task` objects to combine, optionally named.
#'
#' @returns A tibble with the combined samples from all tasks, with a `task`
#' column indicating the source and a nested `metadata` column containing
#' additional fields.
#'
#' @examples
#' if (!identical(Sys.getenv("ANTHROPIC_API_KEY"), "")) {
#'   # set the log directory to a temporary directory
#'   withr::local_envvar(VITALS_LOG_DIR = withr::local_tempdir())
#'
#'   library(ellmer)
#'   library(tibble)
#'
#'   simple_addition <- tibble(
#'     input = c("What's 2+2?", "What's 2+3?"),
#'     target = c("4", "5")
#'   )
#'
#'   tsk1 <- Task$new(
#'     dataset = simple_addition,
#'     solver = generate(chat_claude(model = "claude-sonnet-4-5-20250929")),
#'     scorer = model_graded_qa()
#'   )
#'   tsk1$eval()
#'
#'   tsk2 <- Task$new(
#'     dataset = simple_addition,
#'     solver = generate(chat_claude(model = "claude-sonnet-4-5-20250929")),
#'     scorer = detect_includes()
#'   )
#'   tsk2$eval()
#'
#'   combined <- vitals_bind(model_graded = tsk1, string_detection = tsk2)
#' }
#'
#' @export
vitals_bind <- function(...) {
  x <- dots_list(..., .named = TRUE)
  lapply(x, check_inherits, cls = "Task", call = caller_env(0))

  x <- purrr::map2(x, names(x), function(task, task_name) {
    dplyr::mutate(
      task$get_samples(),
      task = task_name,
      .before = everything()
    )
  })

  res <- purrr::list_rbind(x)
  res <- tidyr::nest(
    res,
    .by = any_of(c("task", "id", "epoch", "score")),
    .key = "metadata"
  )
  res
}

# model usage ------------------------------------------------------------------
translate_to_model_usage <- function(chat) {
  tokens <- as.data.frame(chat$get_tokens())
  model <- chat$get_model()

  dots_list(
    !!model := list(
      input_tokens = sum(tokens$input, na.rm = TRUE),
      output_tokens = sum(tokens$output, na.rm = TRUE),
      total_tokens = sum(tokens$input, tokens$output, na.rm = TRUE),
      input_tokens_cache_write = 0,
      input_tokens_cache_read = sum(tokens$cached_input, na.rm = TRUE)
    )
  )
}

# given the list of solvers in a dataset, sum across all of their token usage
sum_model_usage <- function(solvers) {
  chat <- solvers[[1]]

  usage_per_solver <- lapply(
    solvers,
    function(chat) {
      translate_to_model_usage(chat)[[1]]
    }
  )
  res <- Reduce(function(x, y) Map(`+`, x, y), usage_per_solver)

  # TODO: ultimately, this needs to be per-model
  dots_list(!!chat$get_model() := res)
}

# output ----------------------------------------------------------------------
translate_to_output <- function(chat) {
  last_assistant_turn <- chat$last_turn()

  list(
    model = chat$get_model(),
    choices = translate_assistant_choices(last_assistant_turn),
    usage = translate_to_model_usage(chat)[[1]],
    time = turn_duration(last_assistant_turn)
  )
}

translate_assistant_choices <- function(turn) {
  text_contents <- first_text_contents(turn)
  text <- if (!is.null(text_contents)) text_contents@text else ""
  list(list(
    message = list(
      id = generate_id(),
      content = list(list(
        type = "text",
        text = text
      )),
      source = "generate",
      role = turn@role
    ),
    # turn@json$stop_reason gives the actual stop reason (often 'end_turn'), but
    # Inspect requires "Input should be 'stop', 'max_tokens', 'model_length',
    # 'tool_calls', 'content_filter' or 'unknown'" . (#7)
    stop_reason = "stop"
  ))
}

# miscellaneous ----------------------------------------------------------------
first_text_contents <- function(turn) {
  turn_contents <- turn@contents
  is_text <- vapply(turn_contents, inherits, logical(1), "ellmer::ContentText")
  turn_contents[[which(is_text)[1]]]
}

message_content_from_turn <- function(turn) {
  contents <- turn@contents
  if (length(contents) == 0) {
    return(turn@text)
  }

  entries <- lapply(contents, translate_ellmer_content)
  if (length(entries) > 0) {
    only_text <- all(vapply(
      entries,
      function(entry) {
        is.list(entry) && identical(entry$type, "text")
      },
      logical(1)
    ))
    if (only_text) {
      return(turn@text)
    }
  }

  entries
}

translate_ellmer_content <- function(content) {
  if (inherits(content, "ellmer::ContentText")) {
    return(list(type = "text", text = content@text))
  }

  if (inherits(content, "ellmer::ContentImageInline")) {
    media_type <- content@type %||% "image/png"
    return(list(
      type = "image",
      image = paste0("data:", media_type, ";base64,", content@data)
    ))
  }

  if (inherits(content, "ellmer::ContentImageRemote")) {
    detail <- content@detail %||% "auto"
    data_uri <- remote_url_to_data_uri(content@url)
    result <- list(type = "image", image = data_uri)
    if (!is.null(detail)) {
      result$detail <- detail
    }
    return(result)
  }

  fallback <- paste0(
    utils::capture.output(content),
    collapse = "\n"
  )
  list(type = "text", text = fallback)
}

remote_url_to_data_uri <- function(url) {
  resp <- httr2::request(url) |>
    httr2::req_perform()

  content_type <- httr2::resp_content_type(resp)
  raw_data <- httr2::resp_body_raw(resp)

  base64_data <- jsonlite::base64_enc(raw_data)

  paste0("data:", content_type, ";base64,", base64_data)
}

is_content_list <- function(x) {
  if (!is.list(x) || length(x) == 0) {
    return(FALSE)
  }

  first <- x[[1]]
  is.list(first) && !is.null(first$type)
}

ensure_content_list <- function(content) {
  if (is_content_list(content)) {
    return(content)
  }

  list(list(type = "text", text = as.character(content)))
}

has_tool_calls <- function(turns) {
  any(sapply(turns, function(turn) {
    any(sapply(turn@contents, function(content) {
      inherits(content, "ellmer::ContentToolRequest")
    }))
  }))
}

eval_log_timestamp <- function(time = Sys.time()) {
  timestamp <- format(time, "%Y-%m-%dT%H:%M:%S%z")
  gsub("([+-][0-9]{2})([0-9]{2})$", "\\1:\\2", timestamp)
}

generate_id <- function(length = 22) {
  chars <- c(letters, LETTERS, 0:9)
  paste0(sample(chars, length, replace = TRUE), collapse = "")
}

eval_log_filename <- function(eval_log) {
  paste0(
    clean_filename_component(eval_log$eval$created),
    "_",
    clean_filename_component(eval_log$eval$task_id),
    ".json"
  )
}

clean_filename_component <- function(x) {
  gsub("[_/:]", "-", x)
}

results_scores <- function(name, metrics) {
  if (length(metrics) == 0) {
    metrics <- c()
  }
  list(list(
    name = name,
    scorer = name,
    params = structure(list(), names = character(0)),
    metrics = metrics
  ))
}

rename_metric_fields <- function(metrics) {
  metrics$params <- metrics$arguments
  metrics$arguments <- NULL
  metrics
}

test_that("eval_log_filename sanitizes colons in model names (#207)", {
  eval_log <- list(eval = list(
    created = "2025-02-08T15:51:00-06:00",
    task_id = "dataset/ministral-3:14b/abc123"
  ))
  filename <- eval_log_filename(eval_log)
  expect_false(grepl(":", filename))
  expect_equal(
    filename,
    "2025-02-08T15-51-00-06-00_dataset-ministral-3-14b-abc123.json"
  )
})

test_that("translate_to_model_usage works with example turns", {
  skip_on_cran()
  ellmer_usage <- translate_to_model_usage(example_ellmer_solver())

  inspect_usage <- example_inspect_log()[["samples"]][[1]][["model_usage"]]

  expect_type(ellmer_usage, "list")
  expect_equal(names(ellmer_usage), example_ellmer_solver()$get_model())
  expect_equal(length(ellmer_usage[[1]]), length(inspect_usage[[1]]))
})

test_that("translate_to_output works with example turns", {
  skip_on_cran()
  ellmer_output <- translate_to_output(example_ellmer_solver())

  inspect_output <- example_inspect_log()[["samples"]][[1]][["output"]]

  expect_equal(names(ellmer_output), names(inspect_output))
  expect_equal(length(ellmer_output$usage), length(ellmer_output$usage))
})

mock_chat_turns <- function(..., system_prompt = NULL, model = "test-model") {
  chat <- ellmer::chat_openai_compatible(
    base_url = "https://example.com",
    model = model,
    system_prompt = system_prompt,
    credentials = function() "fake-key"
  )
  pairs <- list(...)
  turns <- list()
  for (pair in pairs) {
    turns <- c(
      turns,
      list(
        ellmer::UserTurn(contents = list(ellmer::ContentText(pair$user))),
        ellmer::AssistantTurn(
          contents = list(ellmer::ContentText(pair$assistant))
        )
      )
    )
  }
  chat$set_turns(turns)
  chat
}

example_ellmer_solver <- function() {
  mock_chat_turns(list(user = "What's 2+2?", assistant = "2+2=4"))
}

# a log actually written by Python Inspect
example_inspect_log <- function() {
  log_path <- system.file(
    "test/inspect/logs/2025-03-24T10-39-36-05-00_simple-arithmetic_fQ9mYnqZFhtEuUenPpJgKL.json",
    package = "vitals"
  )
  if (identical(log_path, "")) {
    testthat::skip("Test log files not available")
  }
  eval_log_read(log_path)
}

example_task <- function(solved = TRUE, scored = TRUE) {
  # loads a cached `tsk` with example output.
  # regenerate with `inst/regenerate-example-objects.R`
  load(
    system.file(
      "test/example-task.rda",
      package = "vitals"
    )
  )

  simple_addition <- tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  res <- Task$new(
    dataset = simple_addition,
    solver = function(...) {},
    scorer = function(...) {}
  )

  if (!solved) {
    return(res)
  }

  res$.__enclos_env__$private$samples$result <- tsk$get_samples()$result
  res$.__enclos_env__$private$samples$solver_chat <- tsk$get_samples()$solver_chat
  res$.__enclos_env__$private$solved <- TRUE

  if (!scored) {
    return(res)
  }

  res$.__enclos_env__$private$samples$score <- tsk$get_samples()$score
  res$.__enclos_env__$private$samples$scorer_chat <- tsk$get_samples()$scorer_chat
  res$.__enclos_env__$private$scored <- TRUE

  res
}

test_that("translate_to_messages works with example turns", {
  skip_on_cran()
  ellmer_messages <- translate_to_messages(example_ellmer_solver()$set_system_prompt(
    NULL
  ))

  inspect_messages <- example_inspect_log()[["samples"]][[1]][["messages"]]

  zap_volatile <- function(l) {
    l[["id"]] <- NULL
    l[["model"]] <- NULL
    l
  }
  expect_equal(
    purrr::map(ellmer_messages, zap_volatile),
    purrr::map(inspect_messages, zap_volatile)
  )
})

test_that("assistant messages include model field", {
  chat <- mock_chat_turns(list(user = "Hi", assistant = "Hello"))
  messages <- translate_to_messages(chat)
  expect_equal(messages[[2]]$model, "test-model")
})

test_that("translate_to_messages handles image inputs", {
  key_get("ANTHROPIC_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)
  library(ellmer)

  dataset <- tibble::tibble(
    input = "What does this image show?",
    target = "The image shows a bike."
  )

  image_solver <- function(
    inputs,
    solver_chat = chat_claude(model = "claude-sonnet-4-5-20250929")
  ) {
    image_file <- system.file("test/x.png", package = "vitals")
    ch <- solver_chat$clone()
    ch$chat(inputs[1], content_image_file(image_file), echo = FALSE)
    list(result = ch$last_turn()@text, solver_chat = list(ch))
  }

  tsk <- Task$new(
    dataset = dataset,
    solver = image_solver,
    scorer = model_graded_qa()
  )

  tsk$eval()
  expect_valid_log(tsk$log())
  expect_true(any(grepl("data:image", readLines(tsk$log()), fixed = TRUE)))
})

test_that("translate_to_messages handles remote image URLs", {
  skip_on_cran()
  key_get("ANTHROPIC_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)
  library(ellmer)

  dataset <- tibble::tibble(
    input = "What does this image show?",
    target = "The R logo."
  )

  image_url_solver <- function(
    inputs,
    solver_chat = chat_claude(model = "claude-sonnet-4-5-20250929")
  ) {
    ch <- solver_chat$clone()
    ch$chat(
      inputs[1],
      content_image_url("https://www.r-project.org/Rlogo.png"),
      echo = FALSE
    )
    list(result = ch$last_turn()@text, solver_chat = list(ch))
  }

  tsk <- Task$new(
    dataset = dataset,
    solver = image_url_solver,
    scorer = model_graded_qa()
  )

  tsk$eval()
  log_path <- tsk$log()
  expect_valid_log(log_path)

  log_content <- readLines(log_path)
  expect_true(any(grepl("data:image", log_content, fixed = TRUE)))
  expect_false(any(grepl("https://www.r-project.org", log_content, fixed = TRUE)))
})


test_that("logs including system prompts are compatible with inspect", {
  vcr::local_cassette("translate-messages-system-prompts")
  key_get("OPENAI_API_KEY")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(list(VITALS_LOG_DIR = tmp_dir))
  withr::local_options(cli.default_handler = function(...) {})
  local_mocked_bindings(interactive = function(...) FALSE)

  library(ellmer)

  simple_addition <- tibble::tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(chat_openai(
      system_prompt = "Be terse.",
      model = "gpt-4.1-nano"
    )),
    scorer = model_graded_qa()
  )

  tsk$eval()
  expect_valid_log(tsk$log())
  expect_true(any(grepl("Be terse", readLines(tsk$log()))))
})

#' Convert a chat to a solver function
#'
#' @description
#' `generate()` is the simplest possible solver one might use with
#' vitals; it just passes its inputs to the supplied model and returns
#' its raw responses. The inputs are evaluated in parallel,
#' not in the sense of multiple R sessions, but in the sense of multiple,
#' asynchronous HTTP requests using [ellmer::parallel_chat()]. `generate()`'s output
#' can be passed directory to the `solver` argument of [Task]'s `$new()`
#' method.
#'
#' @param solver_chat An ellmer chat object, such as from
#'   [ellmer::chat_claude()], or a zero-argument function that returns one.
#'
#' @returns
#' The output of `generate()` is another function. That function takes in
#' a vector of `input`s, as well as a solver chat by the
#' name of `solver_chat` with the default supplied to `generate()` itself.
#'
#' See the documentation for the `solver` argument in [Task] for more
#' information on the return type.
#'
#' @seealso [generate_structured()] for structured output extraction.
#'
#' @inherit Task examples
#' @export
generate <- function(solver_chat = NULL) {
  chat <- solver_chat

  function(inputs, ..., solver_chat = chat) {
    if (is.function(solver_chat)) {
      ch <- solver_chat()
      check_inherits(ch, "Chat")
    } else {
      check_inherits(solver_chat, "Chat")
      ch <- solver_chat$clone()
    }

    res <- ellmer::parallel_chat(ch, as.list(inputs), ...)

    list(
      result = purrr::map_chr(res, function(c) c$last_turn()@text),
      solver_chat = res
    )
  }
}

#' Convert a chat to a solver function with structured output
#'
#' @description
#' `generate_structured()` is a variant of [generate()] that uses
#' [ellmer::parallel_chat_structured()] to extract structured data from
#' the model's responses. This allows you to define a schema for the
#' expected output using ellmer's `type_*()` functions.
#'
#' Because `parallel_chat_structured()` returns structured data rather than
#' Chat objects, `generate_structured()` creates synthetic Chat objects
#' for logging purposes. These "mock" chats contain the input and
#' JSON-serialized output as turns, but won't include actual token usage
#' or timing metadata from the API.
#'
#' The `result` field contains JSON-serialized strings for compatibility
#' with existing scorers. The raw structured data is available in
#' `$get_samples()$solver_metadata` after calling `$solve()` or `$eval()`.
#'
#' @inheritParams generate
#' @param type A type specification for the extracted data, created with
#'   ellmer's `type_*()` functions (e.g., [ellmer::type_object()],
#'   [ellmer::type_string()]). This defines the schema for the structured
#'   output.
#'
#' @inherit generate return
#'
#' @seealso [generate()] for unstructured output, [ellmer::type_object()] and
#'   related functions for defining type specifications.
#'
#' @examples
#' if (FALSE) {
#'   library(ellmer)
#'
#'   type_answer <- type_object(
#'     answer = type_string(
#'       "The author's first name, with no other formatting."
#'     )
#'   )
#'
#'   names <- tibble::tribble(
#'     ~input,                                  ~target,
#'     "Name's Josiah, how's it going?",        "Josiah",
#'     "I'm Lin, what's your name?",            "Lin",
#'     "My name is Em Fields, how about you?",  "Em"
#'   )
#'
#'   tsk <- Task$new(
#'     dataset = names,
#'     solver = generate_structured(
#'       solver_chat = chat_anthropic(model = "claude-sonnet-4-20250514"),
#'       type = type_answer
#'     ),
#'     scorer = detect_match("any")
#'   )
#'
#'   tsk$eval()
#'
#'   # the result is JSON-serialized for compatibility with scorers
#'   tsk$get_samples()$result
#'
#'   # raw structured data is available in solver_metadata
#'   tsk$get_samples()$solver_metadata
#'
#'   # solver_chat contains synthetic turns for logging
#'   tsk$get_samples()$solver_chat[[1]]
#' }
#' @export
generate_structured <- function(solver_chat = NULL, type = NULL) {
  chat <- solver_chat
  type_spec <- type

  function(inputs, ..., solver_chat = chat, type = type_spec) {
    if (is.function(solver_chat)) {
      ch <- solver_chat()
      check_inherits(ch, "Chat")
    } else {
      check_inherits(solver_chat, "Chat")
      ch <- solver_chat$clone()
    }

    res <- ellmer::parallel_chat_structured(
      ch,
      as.list(inputs),
      type = type,
      ...
    )

    res_per_input <- split_structured_result(res, length(inputs))

    solver_chats <- purrr::map2(
      inputs,
      res_per_input,
      mock_chat,
      solver_chat = ch
    )

    result_strings <- purrr::map_chr(
      res_per_input,
      function(x) as.character(jsonlite::toJSON(x, auto_unbox = TRUE))
    )

    list(
      result = result_strings,
      solver_chat = solver_chats,
      solver_metadata = res_per_input
    )
  }
}

split_structured_result <- function(res, n) {
  if (is.data.frame(res)) {
    lapply(seq_len(nrow(res)), function(i) res[i, , drop = FALSE])
  } else if (is.list(res) && length(res) == n) {
    res
  } else {
    rep(list(res), n)
  }
}

mock_chat <- function(input, result, solver_chat) {
  chat <- solver_chat$clone()

  user_turn <- ellmer::UserTurn(
    contents = list(ellmer::ContentText(as.character(input)))
  )

  result_json <- jsonlite::toJSON(result, auto_unbox = TRUE)
  assistant_turn <- ellmer::AssistantTurn(
    contents = list(ellmer::ContentText(as.character(result_json)))
  )

  chat$set_turns(list(user_turn, assistant_turn))

  chat
}

#' Prepare logs for deployment
#'
#' @description
#' This function creates a standalone bundle of the Inspect viewer with log files
#' that can be deployed statically. It copies the UI viewer files, log files, and
#' generates the necessary configuration files.
#'
#' @param log_dir Path to the directory containing log files. Defaults to
#' `vitals_log_dir()`.
#' @param output_dir Path to the directory where the bundled output will be placed.
#' @param overwrite Whether to overwrite an existing output directory. Defaults to FALSE.
#'
#' @section Deployment:
#'
#' This function generates a directory that's ready for deployment to any
#' static web server such as GitHub Pages, S3 buckets, or Netlify. If you
#' have a connection to Posit Connect configured, you can deploy a directory
#' of log files with the following:
#'
#' ```r
#' tmp_dir <- withr::local_tempdir()
#' vitals_bundle(output_dir = tmp_dir, overwrite = TRUE)
#' rsconnect::deployApp(tmp_dir)
#' ```
#'
#' @return Invisibly returns the output directory path. That directory contains:
#'
#' ```
#' output_dir
#' |-- index.html
#' |-- robots.txt
#' |-- assets
#'     |--  ..
#' |-- logs
#'     |--  ..
#' ```
#'
#' `robots.txt` prevents crawlers from indexing the viewer. That said, many
#' crawlers only read the `robots.txt` at the root directory of a package, so
#' the file will likely be ignored if this folder isn't the root directory of
#' the deployed page. `assets/` is the bundled source for the viewer. `logs/`
#' is the `log_dir` as well as a `listing.json`, which is a manifest file for the
#' directory.
#'
#' @examples
#' if (!identical(Sys.getenv("ANTHROPIC_API_KEY"), "")) {
#'   # set the log directory to a temporary directory
#'   withr::local_envvar(VITALS_LOG_DIR = withr::local_tempdir())
#'
#'   library(ellmer)
#'   library(tibble)
#'
#'   simple_addition <- tibble(
#'     input = c("What's 2+2?", "What's 2+3?"),
#'     target = c("4", "5")
#'   )
#'
#'   tsk <- Task$new(
#'     dataset = simple_addition,
#'     solver = generate(chat_claude(model = "claude-sonnet-4-5-20250929")),
#'     scorer = model_graded_qa()
#'   )
#'
#'   tsk$eval()
#'
#'   output_dir <- tempdir()
#'   vitals_bundle(output_dir = output_dir, overwrite = TRUE)
#' }
#'
#' @export
vitals_bundle <- function(
  log_dir = vitals_log_dir(),
  output_dir = NULL,
  overwrite = FALSE
) {
  if (!dir.exists(log_dir)) {
    cli::cli_abort(
      "{.arg log_dir} {.file {log_dir}} doesn't exist."
    )
  }
  check_string(output_dir)

  if (dir.exists(output_dir) && !overwrite) {
    cli::cli_abort(c(
      "'{.arg output_dir}' already exists.",
      "i" = "Choose another output directory or use overwrite = TRUE."
    ))
  }

  working_dir <- withr::local_tempdir()

  dist_dir <- system.file("dist", package = "vitals")
  copy_dir_contents(dist_dir, working_dir)

  log_dir_name <- "logs"
  view_logs_dir <- file.path(working_dir, log_dir_name)
  dir.create(view_logs_dir, recursive = TRUE, showWarnings = FALSE)

  copy_log_files(log_dir, view_logs_dir)

  write_log_dir_manifest(view_logs_dir)

  inject_configuration(file.path(working_dir, "index.html"), log_dir_name)
  write_robots_txt(working_dir)

  move_output(working_dir, output_dir)

  cli::cli_alert_success("Bundle {.file {output_dir}} created!")
  invisible(output_dir)
}

copy_dir_contents <- function(source_dir, dest_dir) {
  if (!dir.exists(dest_dir)) {
    dir.create(dest_dir, recursive = TRUE, showWarnings = FALSE)
  }

  files <- list.files(
    source_dir,
    recursive = TRUE,
    all.files = TRUE,
    include.dirs = FALSE
  )

  for (file in files) {
    src_file <- file.path(source_dir, file)
    dest_file <- file.path(dest_dir, file)

    parent_dir <- dirname(dest_file)
    if (!dir.exists(parent_dir)) {
      dir.create(parent_dir, recursive = TRUE, showWarnings = FALSE)
    }

    file.copy(src_file, dest_file, overwrite = TRUE)
  }
}

copy_log_files <- function(log_dir, target_dir, call = caller_env()) {
  if (!dir.exists(log_dir)) {
    cli::cli_abort("The log directory {.file {log_dir}} doesn't exist.")
  }

  log_files <- list.files(
    log_dir,
    pattern = "\\.json$",
    full.names = TRUE,
    recursive = TRUE
  )

  if (length(log_files) == 0) {
    cli::cli_abort(
      "The log directory {.file {log_dir}} doesn't contain any JSON log files.",
      call = call
    )
  }

  for (file in log_files) {
    rel_path <- sub(
      paste0("^", gsub("([.\\])", "\\\\\\1", normalizePath(log_dir))),
      "",
      normalizePath(file)
    )
    rel_path <- sub("^[/\\\\]", "", rel_path)

    dest_file <- file.path(target_dir, rel_path)
    dest_dir <- dirname(dest_file)

    if (!dir.exists(dest_dir)) {
      dir.create(dest_dir, recursive = TRUE, showWarnings = FALSE)
    }

    file.copy(file, dest_file, overwrite = TRUE)
  }
}

write_log_dir_manifest <- function(log_dir) {
  manifest_file <- file.path(log_dir, "listing.json")

  log_files <- list.files(
    log_dir,
    pattern = "\\.json$",
    recursive = TRUE,
    full.names = TRUE
  )
  log_files <- setdiff(log_files, file.path(log_dir, "listing.json"))

  headers <- lapply(log_files, eval_log_read_headers)
  overviews <- lapply(headers, log_header_to_overview)
  manifest <- setNames(overviews, basename(log_files))

  task_ids <- vapply(headers, function(h) h$eval$task_id %||% NA_character_, character(1))
  if (any(vapply(task_ids, is_old_task_id_format, logical(1)), na.rm = TRUE)) {
    warn_old_task_id_format()
  }

  jsonlite::write_json(
    manifest,
    manifest_file,
    auto_unbox = TRUE,
    pretty = TRUE
  )
}

log_header_to_overview <- function(header) {
  primary_metric <- NULL
  if (!is.null(header$results) &&
      !is.null(header$results$scores) &&
      nrow(header$results$scores) > 0) {
    metrics_df <- header$results$scores$metrics[[1]]
    if (!is.null(metrics_df) && nrow(metrics_df) > 0) {
      primary_metric <- list(
        name = metrics_df$name[1],
        value = metrics_df$value[1],
        params = as.list(metrics_df$params[1, , drop = FALSE])
      )
    }
  }

  list(
    eval_id = header$eval$eval_id,
    run_id = header$eval$run_id,
    task = header$eval$task,
    task_id = header$eval$task_id,
    task_version = header$eval$task_version,
    version = header$version,
    status = header$status,
    model = header$eval$model,
    started_at = header$eval$created,
    completed_at = if (!is.null(header$stats)) header$stats$completed_at else NULL,
    primary_metric = primary_metric
  )
}

inject_configuration <- function(html_file, log_dir) {
  content <- readLines(html_file, warn = FALSE)

  config_data <- list(
    log_dir = log_dir,
    web_path = ".",
    deployment_type = "static",
    version = "1.0"
  )

  # Convert to JSON string and properly escape for HTML embedding
  config_json <- jsonlite::toJSON(config_data, auto_unbox = TRUE)
  log_dir_script <- sprintf(
    '  <script id="log_dir_context" type="application/json">%s</script>',
    config_json
  )

  head_close_pos <- grep("</head>", content, fixed = TRUE)[1]
  if (!is.na(head_close_pos)) {
    content <- c(
      content[1:(head_close_pos - 1)],
      log_dir_script,
      content[head_close_pos:length(content)]
    )
  }

  writeLines(content, html_file)
}

write_robots_txt <- function(dir) {
  file_path <- file.path(dir, "robots.txt")
  content <- "User-agent: *\nDisallow: /\n"
  writeLines(content, file_path)
}

move_output <- function(from_dir, to_dir) {
  if (dir.exists(to_dir)) {
    unlink(to_dir, recursive = TRUE)
  }

  dir.create(to_dir, recursive = TRUE, showWarnings = FALSE)

  copy_dir_contents(from_dir, to_dir)
}

test_that("vitals_bundle creates a valid bundle", {
  output_dir <- file.path(withr::local_tempdir(), "test-inspect-bundle")
  if (dir.exists(output_dir)) {
    unlink(output_dir, recursive = TRUE)
  }

  log_dir <- system.file("test/inspect/logs", package = "vitals")
  if (identical(log_dir, "")) {
    skip("Test log files not available")
  }

  dir.create(output_dir)
  expect_snapshot(
    error = TRUE,
    vitals_bundle(log_dir = log_dir, output_dir = output_dir, overwrite = FALSE)
  )
  unlink(output_dir, recursive = TRUE)

  expect_snapshot(
    result <- vitals_bundle(
      log_dir = log_dir,
      output_dir = output_dir,
      overwrite = TRUE
    ),
    transform = function(x) gsub("'[^']*'", "'***'", x)
  )

  expect_equal(result, output_dir)

  expect_true(dir.exists(output_dir))
  expect_true(file.exists(file.path(output_dir, "index.html")))
  expect_true(file.exists(file.path(output_dir, "robots.txt")))
  expect_true(dir.exists(file.path(output_dir, "logs")))
  expect_true(
    length(list.files(file.path(output_dir, "logs"), pattern = "\\.json$")) > 0
  )

  unlink(output_dir, recursive = TRUE)
})

test_that("vitals_bundle errors informatively", {
  output_dir <- file.path(withr::local_tempdir(), "test-inspect-bundle-empty")
  if (dir.exists(output_dir)) {
    unlink(output_dir, recursive = TRUE)
  }
  withr::defer(unlink(output_dir, recursive = TRUE))

  empty_log_dir <- file.path(withr::local_tempdir(), "empty-logs")
  if (!dir.exists(empty_log_dir)) {
    dir.create(empty_log_dir)
  }
  withr::defer(unlink(empty_log_dir, recursive = TRUE))

  expect_snapshot(
    error = TRUE,
    vitals_bundle(log_dir = empty_log_dir, output_dir = output_dir),
    transform = function(x) gsub("'[^']*'", "'***'", x)
  )

  non_existent_dir <- file.path(withr::local_tempdir(), "non-existent-dir")
  if (dir.exists(non_existent_dir)) {
    unlink(non_existent_dir, recursive = TRUE)
  }

  expect_snapshot(
    error = TRUE,
    vitals_bundle(log_dir = non_existent_dir, output_dir = output_dir),
    transform = function(x) gsub("'[^']*'", "'***'", x)
  )
})

#' The log directory
#'
#' @description
#' vitals supports the `VITALS_LOG_DIR` environment variable,
#' which sets a default directory to write logs to in [Task]'s `$eval()`
#' and `$log()` methods.
#'
#' @param dir A directory to configure the environment variable
#' `VITALS_LOG_DIR` to.
#'
#' @returns
#' Both `vitals_log_dir()` and `vitals_log_dir_set()` return the current
#' value of the environment variable `VITALS_LOG_DIR`. `vitals_log_dir_set()`
#' additionally sets it to a new value.
#'
#' To set this variable in every new R session, you might consider adding it
#' to your `.Rprofile`, perhaps with `usethis::edit_r_profile()`.
#'
#' @examples
#' vitals_log_dir()
#'
#' dir <- tempdir()
#'
#' vitals_log_dir_set(dir)
#'
#' vitals_log_dir()
#' @export
vitals_log_dir <- function() {
  Sys.getenv("VITALS_LOG_DIR", unset = NA)
}

#' @rdname vitals_log_dir
#' @export
vitals_log_dir_set <- function(dir) {
  old <- Sys.getenv("VITALS_LOG_DIR", unset = NA)
  Sys.setenv(VITALS_LOG_DIR = dir)
  invisible(old)
}

# Read and write evaluation logs
#
# @description
# Utilities for reading and writing evaluation logs.
#
# These utilities support the `VITALS_LOG_DIR` environment variable,
# which sets a default directory to write logs to.
#
# @param x For `eval_log_read()`, a path to the log file to read. For
# For `eval_log_write()`, an evaluation log object to write to disk.
# @param dir Directory where logs should be stored.
#
# @examples
# file <-
#   system.file(
#     "logs/2025-02-08T15-51-00-06-00_simple-arithmetic_o3cKtmsvqQtmXGZhvfDrKB.json",
#      package = "vitals"
#   )
#
# example_eval_log <- eval_log_read(file)
#
# example_eval_log
# @name eval_log
eval_log_read <- function(x) {
  structure(jsonlite::read_json(x), class = c("eval_log", "list"))
}

# @rdname eval_log
eval_log_write <- function(x = eval_log(), dir = vitals_log_dir()) {
  if (!dir.exists(dir)) {
    dir.create(dir, showWarnings = FALSE, recursive = TRUE)
  }

  path <- file.path(dir, eval_log_filename(x))
  jsonlite::write_json(
    x = structure(x, class = "list"),
    path = path,
    auto_unbox = TRUE,
    pretty = TRUE
  )
  path
}

# eval log files are quite relational, where the `samples` and `logging` fields
# take up the most storage by far. reading with `simplifyVector = FALSE` is
# much faster, so read the whole thing, subset out what we need, and
# then write/read the simplified version. (#26)
header_fields <-
  c("version", "status", "eval", "plan", "results", "stats")

eval_log_read_headers <- function(x) {
  # read only the needed fields without simplifying
  res_fields <- jsonlite::fromJSON(x, simplifyVector = FALSE)[header_fields]

  # write/read the small fields, simplifying this time around
  jsonlite::fromJSON(jsonlite::toJSON(res_fields, auto_unbox = TRUE))
}

test_that("vitals_log_dir works", {
  current_dir <- vitals_log_dir()
  withr::defer(vitals_log_dir_set(current_dir))

  # returns NA if not set
  withr::local_envvar(VITALS_LOG_DIR = NULL)
  expect_identical(vitals_log_dir(), NA_character_)

  # returns value if set
  withr::local_envvar(VITALS_LOG_DIR = "path/to/logs")
  expect_equal(vitals_log_dir(), "path/to/logs")
})

test_that("vitals_log_dir_set works", {
  current_dir <- vitals_log_dir()
  withr::defer(vitals_log_dir_set(current_dir))

  # returns previous value invisibly
  withr::local_envvar(VITALS_LOG_DIR = NULL)
  expect_equal(vitals_log_dir_set("abc123"), NA_character_)

  withr::local_envvar(VITALS_LOG_DIR = "abc123")
  expect_equal(vitals_log_dir_set("def456"), "abc123")

  # sets new value for future calls
  withr::local_envvar(VITALS_LOG_DIR = NULL)
  vitals_log_dir_set("abc123")
  expect_equal(vitals_log_dir(), "abc123")

  vitals_log_dir_set("def456")
  expect_equal(vitals_log_dir(), "def456")
})

test_that("vitals_view() works with read + written eval file", {
  log_dir <- tempdir()
  if (!dir.exists(log_dir)) dir.create(log_dir)
  withr::defer(unlink(log_dir, recursive = TRUE))

  # file generated from Python Inspect
  eval_log <- example_inspect_log()
  eval_log_write(x = eval_log, dir = log_dir)

  expect_condition(vitals_view(log_dir), class = "vitals_viewer_start")
})

#' Scoring with string detection
#'
#' @description
#' The following functions use string pattern detection to score model outputs.
#'
#' - `detect_includes()`: Determine whether the `target` from the sample
#' appears anywhere inside the model output. Can be case sensitive or
#' insensitive (defaults to the latter).
#' - `detect_match()`: Determine whether the `target` from the sample appears
#' at the beginning or end of model output (defaults to looking at the end).
#' Has options for ignoring case, white-space, and punctuation
#' (all are ignored by default).
#' - `detect_pattern()`: Extract matches of a pattern from the model response
#' and determine whether those matches also appear in `target`.
#' - `detect_answer()`: Scorer for model output that precedes answers with
#' "ANSWER: ". Can extract letters, words, or the remainder of the line.
#' - `detect_exact()`: Scorer which will normalize the text of the answer and
#' target(s) and perform an exact matching comparison of the text. This
#' scorer will return `CORRECT` when the answer is an exact match to one
#' or more targets.
#'
#' @param case_sensitive Logical, whether comparisons are case sensitive.
#' @param location Where to look for match: one of `"end"`, `"begin"`,
#' `"any"`, or `"exact"`. Defaults to `"end"`.
#' @param pattern Regular expression pattern to extract answer.
#' @param all Logical: for multiple captures, whether all must match.
#' @param format What to extract after `"ANSWER:"`: `"letter"`, `"word"`,
#' or `"line"`. Defaults to `"line"`.
#'
#' @seealso [model_graded_qa()] and [model_graded_fact()] for model-based
#' scoring.
#'
#' @returns
#' A function that scores model output based on string matching. Pass the
#' returned value to `$eval(scorer)`. See the documentation for the `scorer`
#' argument in [Task] for more information on the return type.
#'
#' @examples
#' if (!identical(Sys.getenv("ANTHROPIC_API_KEY"), "")) {
#'   # set the log directory to a temporary directory
#'   withr::local_envvar(VITALS_LOG_DIR = withr::local_tempdir())
#'
#'   library(ellmer)
#'   library(tibble)
#'
#'   simple_addition <- tibble(
#'     input = c("What's 2+2?", "What's 2+3?"),
#'     target = c("4", "5")
#'   )
#'
#'   # create a new Task
#'   tsk <- Task$new(
#'     dataset = simple_addition,
#'     solver = generate(solver_chat = chat_claude(model = "claude-sonnet-4-5-20250929")),
#'     scorer = detect_includes()
#'   )
#'
#'   # evaluate the task (runs solver and scorer)
#'   tsk$eval()
#' }
#'
#' @name scorer_detect
#' @export
detect_includes <- function(case_sensitive = FALSE) {
  check_bool(case_sensitive)

  function(samples) {
    results <- purrr::pmap(
      samples,
      function(...) {
        detect_includes_impl(
          list(...),
          case_sensitive = case_sensitive
        )
      }
    )

    list(
      score = factor(
        ifelse(purrr::map_lgl(results, "result"), "C", "I"),
        levels = c("I", "C"),
        ordered = TRUE
      ),
      explanation = purrr::map_chr(results, "explanation"),
      scorer_metadata = purrr::map(results, "metadata")
    )
  }
}

detect_includes_impl <- function(sample, case_sensitive) {
  answer <- sample$result
  target <- sample$target

  if (!case_sensitive) {
    answer <- tolower(answer)
    target <- tolower(target)
  }

  result <- grepl(target, answer, fixed = TRUE)

  explanation <- if (result) {
    paste0("Target '", sample$target, "' was found in the answer.")
  } else {
    paste0("Target '", sample$target, "' was not found in the answer.")
  }

  list(
    result = result,
    explanation = explanation,
    metadata = list(
      matched = result,
      answer = answer
    )
  )
}

#' @rdname scorer_detect
#' @export
detect_match <- function(
  location = c("end", "begin", "any", "exact"),
  case_sensitive = FALSE
) {
  location <- arg_match(location)
  check_bool(case_sensitive)

  function(samples) {
    results <- purrr::pmap(
      samples,
      function(...) {
        detect_match_impl(
          list(...),
          location = location,
          case_sensitive = case_sensitive
        )
      }
    )

    list(
      score = factor(
        ifelse(purrr::map_lgl(results, "result"), "C", "I"),
        levels = c("I", "C"),
        ordered = TRUE
      ),
      explanation = purrr::map_chr(results, "explanation"),
      scorer_metadata = purrr::map(results, "metadata")
    )
  }
}

detect_match_impl <- function(sample, location, case_sensitive) {
  answer <- trimws(sample$result)
  target <- trimws(sample$target)

  if (!case_sensitive) {
    answer <- tolower(answer)
    target <- tolower(target)
  }

  result <- switch(
    location,
    begin = startsWith(answer, target),
    end = endsWith(answer, target),
    any = grepl(target, answer, fixed = TRUE),
    exact = answer == target,
    FALSE
  )

  location_desc <- switch(
    location,
    begin = "at the beginning of",
    end = "at the end of",
    any = "anywhere in",
    exact = "exactly matching",
    "in"
  )

  explanation <- if (result) {
    paste0("Target '", sample$target, "' was found ", location_desc, " the answer.")
  } else {
    paste0("Target '", sample$target, "' was not found ", location_desc, " the answer.")
  }

  list(
    result = result,
    explanation = explanation,
    metadata = list(
      matched = result,
      answer = answer
    )
  )
}

#' @rdname scorer_detect
#' @export
detect_pattern <- function(pattern, case_sensitive = FALSE, all = FALSE) {
  check_string(pattern)
  check_bool(case_sensitive)
  check_bool(all)

  function(samples) {
    results <- purrr::pmap(
      samples,
      function(...) {
        detect_pattern_impl(
          list(...),
          pattern = pattern,
          case_sensitive = case_sensitive,
          all = all
        )
      }
    )

    list(
      score = factor(
        ifelse(purrr::map_lgl(results, "result"), "C", "I"),
        levels = c("I", "C"),
        ordered = TRUE
      ),
      explanation = purrr::map_chr(results, "explanation"),
      scorer_metadata = purrr::map(results, "metadata")
    )
  }
}

detect_pattern_impl <- function(sample, pattern, case_sensitive, all) {
  flags <- if (!case_sensitive) ignore.case = TRUE else NULL
  matches <- regexec(pattern, sample$result, perl = TRUE, flags)
  if (matches[[1]][1] == -1) {
    return(list(
      result = FALSE,
      explanation = paste0(
        "Pattern '", pattern, "' did not match any text in the answer."
      ),
      metadata = list(
        matched = FALSE,
        answer = NA
      )
    ))
  }

  groups <- regmatches(sample$result, matches)[[1]][-1]
  target <- sample$target

  if (!case_sensitive) {
    groups <- tolower(groups)
    target <- tolower(target)
  }

  matched <- if (all) {
    all(groups %in% target)
  } else {
    any(groups %in% target)
  }

  match_desc <- if (all) "All matches" else "At least one match"
  explanation <- if (matched) {
    paste0(match_desc, " from pattern '", pattern, "' matched the target.")
  } else {
    paste0(match_desc, " from pattern '", pattern, "' did not match the target.")
  }

  list(
    result = matched,
    explanation = explanation,
    metadata = list(
      matched = matched,
      answer = groups[1]
    )
  )
}

#' @rdname scorer_detect
#' @export
detect_exact <- function(case_sensitive = FALSE) {
  check_bool(case_sensitive)

  function(samples) {
    results <- purrr::pmap(
      samples,
      function(...) {
        detect_exact_impl(
          list(...),
          case_sensitive = case_sensitive
        )
      }
    )

    list(
      score = factor(
        ifelse(purrr::map_lgl(results, "result"), "C", "I"),
        levels = c("I", "C"),
        ordered = TRUE
      ),
      explanation = purrr::map_chr(results, "explanation"),
      scorer_metadata = purrr::map(results, "metadata")
    )
  }
}

detect_exact_impl <- function(sample, case_sensitive) {
  answer <- trimws(gsub("[[:punct:]]", "", sample$result))
  target <- trimws(gsub("[[:punct:]]", "", sample$target))

  if (!case_sensitive) {
    answer <- tolower(answer)
    target <- tolower(target)
  }

  matched <- answer == target

  explanation <- if (matched) {
    paste0("Normalized answer exactly matched target '", sample$target, "'.")
  } else {
    paste0("Normalized answer did not exactly match target '", sample$target, "'.")
  }

  list(
    result = matched,
    explanation = explanation,
    scorer = "exact",
    metadata = list(
      matched = matched,
      answer = answer
    )
  )
}

#' @rdname scorer_detect
#' @export
detect_answer <- function(format = c("line", "word", "letter")) {
  format <- arg_match(format)

  function(samples) {
    results <- purrr::pmap(
      samples,
      function(...) {
        detect_answer_impl(
          list(...),
          format = format
        )
      }
    )

    list(
      score = factor(
        ifelse(purrr::map_lgl(results, "result"), "C", "I"),
        levels = c("I", "C"),
        ordered = TRUE
      ),
      explanation = purrr::map_chr(results, "explanation"),
      scorer_metadata = purrr::map(results, "metadata")
    )
  }
}

detect_answer_impl <- function(sample, format) {
  pattern <- switch(
    format,
    letter = "ANSWER:\\s*([A-Za-z])",
    word = "ANSWER:\\s*(\\w+)",
    line = "ANSWER:\\s*(.+)$",
    "ANSWER:\\s*(.+)$"
  )

  matches <- regexec(pattern, sample$result, perl = TRUE)
  if (matches[[1]][1] == -1) {
    return(list(
      result = FALSE,
      explanation = "No 'ANSWER:' prefix found in the response.",
      metadata = list(
        matched = FALSE,
        answer = NA
      )
    ))
  }

  answer <- regmatches(sample$result, matches)[[1]][2]
  matched <- tolower(trimws(answer)) == tolower(trimws(sample$target))

  explanation <- if (matched) {
    paste0("Extracted answer '", answer, "' matched target '", sample$target, "'.")
  } else {
    paste0("Extracted answer '", answer, "' did not match target '", sample$target, "'.")
  }

  list(
    result = matched,
    explanation = explanation,
    metadata = list(
      matched = matched,
      answer = answer
    )
  )
}

test_that("are is an appropriate task dataset", {
  withr::local_envvar(VITALS_LOG_DIR = withr::local_tempdir())
  tsk <- Task$new(
    dataset = are,
    name = "The R Eval",
    solver = function(...) list(...),
    scorer = function(...) list(...)
  )
  expect_true(R6::is.R6(tsk))
})

# generate() allows NULL default model

    Code
      res()
    Condition
      Error in `res()`:
      ! `solver_chat` must be a <Chat>, not NULL

# generate_structured() allows NULL default model

    Code
      res()
    Condition
      Error in `res()`:
      ! `solver_chat` must be a <Chat>, not NULL

# check_inherits works

    Code
      check_inherits(1, "list")
    Condition
      Error:
      ! `1` must be a <list>, not a number

---

    Code
      check_inherits(1, "list", x_arg = "my_arg")
    Condition
      Error:
      ! `my_arg` must be a <list>, not a number

# check_log_dir warns informatively

    Code
      res <- Task$new(tibble(input = 1, target = 1), function() { }, function() { })
    Condition
      Warning:
      ! vitals could not find a log directory; evaluation log files will be written to a temporary directory.
      i Set a log directory with e.g. `vitals::vitals_log_dir_set("./logs")`, perhaps in '~/.Rprofile', to quiet this warning.

# vitals_bundle creates a valid bundle

    Code
      vitals_bundle(log_dir = log_dir, output_dir = output_dir, overwrite = FALSE)
    Condition
      Error in `vitals_bundle()`:
      ! '`output_dir`' already exists.
      i Choose another output directory or use overwrite = TRUE.

---

    Code
      result <- vitals_bundle(log_dir = log_dir, output_dir = output_dir, overwrite = TRUE)
    Condition
      Warning:
      Log directory contains logs with old-format task IDs.
      i The log viewer may collapse different evals as retried logs.
      i Click Show retried logs in the viewer to see all logs.
      i To silence this warning, regenerate logs with the current vitals version.
      This warning is displayed once per session.
    Message
      v Bundle '***' created!

# vitals_bundle errors informatively

    Code
      vitals_bundle(log_dir = empty_log_dir, output_dir = output_dir)
    Condition
      Error in `vitals_bundle()`:
      ! The log directory '***' doesn't contain any JSON log files.

---

    Code
      vitals_bundle(log_dir = non_existent_dir, output_dir = output_dir)
    Condition
      Error in `vitals_bundle()`:
      ! `log_dir` '***' doesn't exist.

# as_metadata can make lists into a named vector (#69)

    Code
      res
    Output
      $mpg
      [1] 21 21
      
      $cyl
      [1] 6 6
      

---

    Code
      res
    Output
      $x
      [1] "List of 1"      " $ :List of 2"  "  ..$ a: num 1" "  ..$ b: num 2"
      
      $y
      [1] 3
      

---

    Code
      res
    Output
      $`1`
      [1] 1 2 3
      
      $b
      [1] 2
      

---

    Code
      res
    Output
      $`1`
      [1] 1 2 3
      
      $b
      [1] 2
      

# Task R6 class works

    Code
      tsk
    Output
      An evaluation task simple-addition.

---

    Code
      tsk
    Output
      An evaluation task simple-addition.
      Explore interactively with `.last_task$view()`.

# check_dataset works

    Code
      Task$new(dataset = data.frame(input = 1), solver = function() { }, scorer = function()
        { })
    Condition
      Error in `initialize()`:
      ! `dataset` is missing required column target.

---

    Code
      Task$new(dataset = data.frame(target = 1), solver = function() { }, scorer = function()
        { })
    Condition
      Error in `initialize()`:
      ! `dataset` is missing required column input.

---

    Code
      Task$new(dataset = data.frame(x = 1), solver = function() { }, scorer = function()
        { })
    Condition
      Error in `initialize()`:
      ! `dataset` is missing required columns input and target.

# Task errors informatively with duplicate ids

    Code
      Task$new(dataset = d, solver = function() { }, scorer = function() { })
    Condition
      Error in `initialize()`:
      ! Duplicated values found in the id column. Each ID must be unique.

# task errors informatively with bad metrics

    Code
      tsk <- Task$new(dataset = simple_addition, solver = generate(ellmer::chat_openai(
        model = "gpt-4.1-nano")), scorer = function(...) {
        list(score = factor(c("C", "C"), levels = c("I", "P", "C")))
      }, metrics = function(scores) {
        mean(scores == "C") * 100
      })
    Condition
      Error in `initialize()`:
      ! `metrics` must be a named list of functions or NULL, not a function

---

    Code
      tsk$set_metrics(function(...) "boop bop")
    Condition
      Error:
      ! `metrics` must be a named list of functions or NULL, not a function

---

    Code
      tsk$set_metrics(list(bad_metric = function(scores) "this is not a numeric"))
      tsk$eval()
    Condition
      Error in `measure()`:
      ! Each metric function must return a single numeric value
      `bad_metric()` returned a string

# Task errors informatively with bad solver output

    Code
      tsk$solve()
    Condition
      Error in `$solve()`:
      ! `solver` must return slots result and solver_chat.

# Task detects non-Chat objects in solver_chat

    Code
      tsk$solve()
    Condition
      Error in `$solve()`:
      ! Elements in the solver_chat output from `solver` must be ellmer Chat objects, not a string.

# Task errors informatively with bad scorer output

    Code
      tsk$eval()
    Condition
      Error in `$score()`:
      ! `scorer` must return a list with (at least) the slot score.

# Task detects non-Chat objects in scorer_chat

    Code
      tsk$eval()
    Condition
      Error in `$score()`:
      ! Elements in the scorer_chat output from `scorer` must be ellmer Chat objects, not a string.

# eval errors with unnamed arguments

    Code
      tsk$eval("unnamed_arg")
    Condition
      Error in `$eval()`:
      ! All arguments in `...` must be named.

# eval errors when argument matches neither function and neither has ellipses

    Code
      tsk$eval(unmatched_param = "error")
    Condition
      Error in `$eval()`:
      ! Argument `unmatched_param` does not match any parameter in the solver or scorer functions.

# vitals_view() errors informatively on failure to restart

    Code
      vitals_view(log_dir)
    Condition
      Error in `vitals_view()`:
      ! Unable to terminate the existing server.
      Caused by error in `httpuv::stopServer()`:
      ! no way!

# vitals_view() handles nonexistent log directory

    Code
      vitals_view(log_dir)
    Condition
      Error in `vitals_view()`:
      ! Log directory '/path/that/does/not/exist' not found.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bind.R
\name{vitals_bind}
\alias{vitals_bind}
\title{Concatenate task samples for analysis}
\usage{
vitals_bind(...)
}
\arguments{
\item{...}{\code{Task} objects to combine, optionally named.}
}
\value{
A tibble with the combined samples from all tasks, with a \code{task}
column indicating the source and a nested \code{metadata} column containing
additional fields.
}
\description{
Combine multiple \link{Task} objects into a single tibble for comparison.

This function takes multiple (optionally named) \link{Task} objects and row-binds
their \verb{$get_samples()} together, adding a \code{task} column to identify the source of each
row. The resulting tibble nests additional columns into a \code{metadata} column
and is ready for further analysis.
}
\examples{
if (!identical(Sys.getenv("ANTHROPIC_API_KEY"), "")) {
  # set the log directory to a temporary directory
  withr::local_envvar(VITALS_LOG_DIR = withr::local_tempdir())

  library(ellmer)
  library(tibble)

  simple_addition <- tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  tsk1 <- Task$new(
    dataset = simple_addition,
    solver = generate(chat_claude(model = "claude-sonnet-4-5-20250929")),
    scorer = model_graded_qa()
  )
  tsk1$eval()

  tsk2 <- Task$new(
    dataset = simple_addition,
    solver = generate(chat_claude(model = "claude-sonnet-4-5-20250929")),
    scorer = detect_includes()
  )
  tsk2$eval()

  combined <- vitals_bind(model_graded = tsk1, string_detection = tsk2)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scorer-detect.R
\name{scorer_detect}
\alias{scorer_detect}
\alias{detect_includes}
\alias{detect_match}
\alias{detect_pattern}
\alias{detect_exact}
\alias{detect_answer}
\title{Scoring with string detection}
\usage{
detect_includes(case_sensitive = FALSE)

detect_match(
  location = c("end", "begin", "any", "exact"),
  case_sensitive = FALSE
)

detect_pattern(pattern, case_sensitive = FALSE, all = FALSE)

detect_exact(case_sensitive = FALSE)

detect_answer(format = c("line", "word", "letter"))
}
\arguments{
\item{case_sensitive}{Logical, whether comparisons are case sensitive.}

\item{location}{Where to look for match: one of \code{"end"}, \code{"begin"},
\code{"any"}, or \code{"exact"}. Defaults to \code{"end"}.}

\item{pattern}{Regular expression pattern to extract answer.}

\item{all}{Logical: for multiple captures, whether all must match.}

\item{format}{What to extract after \code{"ANSWER:"}: \code{"letter"}, \code{"word"},
or \code{"line"}. Defaults to \code{"line"}.}
}
\value{
A function that scores model output based on string matching. Pass the
returned value to \verb{$eval(scorer)}. See the documentation for the \code{scorer}
argument in \link{Task} for more information on the return type.
}
\description{
The following functions use string pattern detection to score model outputs.
\itemize{
\item \code{detect_includes()}: Determine whether the \code{target} from the sample
appears anywhere inside the model output. Can be case sensitive or
insensitive (defaults to the latter).
\item \code{detect_match()}: Determine whether the \code{target} from the sample appears
at the beginning or end of model output (defaults to looking at the end).
Has options for ignoring case, white-space, and punctuation
(all are ignored by default).
\item \code{detect_pattern()}: Extract matches of a pattern from the model response
and determine whether those matches also appear in \code{target}.
\item \code{detect_answer()}: Scorer for model output that precedes answers with
"ANSWER: ". Can extract letters, words, or the remainder of the line.
\item \code{detect_exact()}: Scorer which will normalize the text of the answer and
target(s) and perform an exact matching comparison of the text. This
scorer will return \code{CORRECT} when the answer is an exact match to one
or more targets.
}
}
\examples{
if (!identical(Sys.getenv("ANTHROPIC_API_KEY"), "")) {
  # set the log directory to a temporary directory
  withr::local_envvar(VITALS_LOG_DIR = withr::local_tempdir())

  library(ellmer)
  library(tibble)

  simple_addition <- tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  # create a new Task
  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(solver_chat = chat_claude(model = "claude-sonnet-4-5-20250929")),
    scorer = detect_includes()
  )

  # evaluate the task (runs solver and scorer)
  tsk$eval()
}

}
\seealso{
\code{\link[=model_graded_qa]{model_graded_qa()}} and \code{\link[=model_graded_fact]{model_graded_fact()}} for model-based
scoring.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/are.R
\docType{data}
\name{are}
\alias{are}
\title{An R Eval}
\format{
A tibble with 29 rows and 7 columns:
\describe{
\item{id}{Character. Unique identifier/title for the code problem.}
\item{input}{Character. The question to be answered.}
\item{target}{Character. The solution, often with a description of notable
features of a correct solution.}
\item{domain}{Character. The technical domain
(e.g., Data Analysis, Programming, or Authoring).}
\item{task}{Character. Type of task
(e.g., Debugging, New feature, or Translation.)}
\item{source}{Character. URL or source of the problem. \code{NA}s indicate that
the problem was written originally for this eval.}
\item{knowledge}{List. Required knowledge/concepts for solving the problem.}
}
}
\source{
Posit Community, GitHub issues, R4DS solutions, etc. For row-level
references, see \code{source}.
}
\usage{
are
}
\description{
An R Eval is a dataset of challenging R coding problems. Each \code{input} is a
question about R code which could be solved on first-read only by experts
and, with a chance to read documentation and run some code, by
fluent data scientists. Solutions are in \code{target()} and enable a fluent
data scientist to evaluate whether the solution deserves full, partial, or
no credit.

Pass this dataset to \code{Task$new()} to situate it inside of an evaluation
task.
}
\examples{
are

dplyr::glimpse(are)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bundle.R
\name{vitals_bundle}
\alias{vitals_bundle}
\title{Prepare logs for deployment}
\usage{
vitals_bundle(log_dir = vitals_log_dir(), output_dir = NULL, overwrite = FALSE)
}
\arguments{
\item{log_dir}{Path to the directory containing log files. Defaults to
\code{vitals_log_dir()}.}

\item{output_dir}{Path to the directory where the bundled output will be placed.}

\item{overwrite}{Whether to overwrite an existing output directory. Defaults to FALSE.}
}
\value{
Invisibly returns the output directory path. That directory contains:

\if{html}{\out{<div class="sourceCode">}}\preformatted{output_dir
|-- index.html
|-- robots.txt
|-- assets
    |--  ..
|-- logs
    |--  ..
}\if{html}{\out{</div>}}

\code{robots.txt} prevents crawlers from indexing the viewer. That said, many
crawlers only read the \code{robots.txt} at the root directory of a package, so
the file will likely be ignored if this folder isn't the root directory of
the deployed page. \verb{assets/} is the bundled source for the viewer. \verb{logs/}
is the \code{log_dir} as well as a \code{listing.json}, which is a manifest file for the
directory.
}
\description{
This function creates a standalone bundle of the Inspect viewer with log files
that can be deployed statically. It copies the UI viewer files, log files, and
generates the necessary configuration files.
}
\section{Deployment}{


This function generates a directory that's ready for deployment to any
static web server such as GitHub Pages, S3 buckets, or Netlify. If you
have a connection to Posit Connect configured, you can deploy a directory
of log files with the following:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{tmp_dir <- withr::local_tempdir()
vitals_bundle(output_dir = tmp_dir, overwrite = TRUE)
rsconnect::deployApp(tmp_dir)
}\if{html}{\out{</div>}}
}

\examples{
if (!identical(Sys.getenv("ANTHROPIC_API_KEY"), "")) {
  # set the log directory to a temporary directory
  withr::local_envvar(VITALS_LOG_DIR = withr::local_tempdir())

  library(ellmer)
  library(tibble)

  simple_addition <- tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(chat_claude(model = "claude-sonnet-4-5-20250929")),
    scorer = model_graded_qa()
  )

  tsk$eval()

  output_dir <- tempdir()
  vitals_bundle(output_dir = output_dir, overwrite = TRUE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/log-dir.R
\name{vitals_log_dir}
\alias{vitals_log_dir}
\alias{vitals_log_dir_set}
\title{The log directory}
\usage{
vitals_log_dir()

vitals_log_dir_set(dir)
}
\arguments{
\item{dir}{A directory to configure the environment variable
\code{VITALS_LOG_DIR} to.}
}
\value{
Both \code{vitals_log_dir()} and \code{vitals_log_dir_set()} return the current
value of the environment variable \code{VITALS_LOG_DIR}. \code{vitals_log_dir_set()}
additionally sets it to a new value.

To set this variable in every new R session, you might consider adding it
to your \code{.Rprofile}, perhaps with \code{usethis::edit_r_profile()}.
}
\description{
vitals supports the \code{VITALS_LOG_DIR} environment variable,
which sets a default directory to write logs to in \link{Task}'s \verb{$eval()}
and \verb{$log()} methods.
}
\examples{
vitals_log_dir()

dir <- tempdir()

vitals_log_dir_set(dir)

vitals_log_dir()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/solver.R
\name{generate}
\alias{generate}
\title{Convert a chat to a solver function}
\usage{
generate(solver_chat = NULL)
}
\arguments{
\item{solver_chat}{An ellmer chat object, such as from
\code{\link[ellmer:chat_anthropic]{ellmer::chat_claude()}}, or a zero-argument function that returns one.}
}
\value{
The output of \code{generate()} is another function. That function takes in
a vector of \code{input}s, as well as a solver chat by the
name of \code{solver_chat} with the default supplied to \code{generate()} itself.

See the documentation for the \code{solver} argument in \link{Task} for more
information on the return type.
}
\description{
\code{generate()} is the simplest possible solver one might use with
vitals; it just passes its inputs to the supplied model and returns
its raw responses. The inputs are evaluated in parallel,
not in the sense of multiple R sessions, but in the sense of multiple,
asynchronous HTTP requests using \code{\link[ellmer:parallel_chat]{ellmer::parallel_chat()}}. \code{generate()}'s output
can be passed directory to the \code{solver} argument of \link{Task}'s \verb{$new()}
method.
}
\examples{
if (!identical(Sys.getenv("ANTHROPIC_API_KEY"), "")) {
  # set the log directory to a temporary directory
  withr::local_envvar(VITALS_LOG_DIR = withr::local_tempdir())

  library(ellmer)
  library(tibble)

  simple_addition <- tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  # create a new Task
  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(chat_claude(model = "claude-sonnet-4-5-20250929")),
    scorer = model_graded_qa()
  )

  # evaluate the task (runs solver and scorer) and opens
  # the results in the Inspect log viewer (if interactive)
  tsk$eval()

  # $eval() is shorthand for:
  tsk$solve()
  tsk$score()
  tsk$measure()
  tsk$log()
  tsk$view()

  # get the evaluation results as a data frame
  tsk$get_samples()

  # view the task directory with $view() or vitals_view()
  vitals_view()
}

# The `input` column can be a list of 1-row tibbles for per-sample metadata.
# Custom solvers can then extract columns from each input:
shapes_data <- tibble::tibble(
  input = list(
    tibble::tibble(shapes = "square, circle, rhombus", pick = "square"),
    tibble::tibble(shapes = "square, circle, rhombus", pick = "circle")
  ),
  target = c("square", "circle")
)

my_solver <- function(solver_chat = NULL) {
  chat <- solver_chat
  function(inputs, ..., solver_chat = chat) {
    ch <- if (is.function(solver_chat)) solver_chat() else solver_chat$clone()
    prompts <- lapply(inputs, function(inp) {
      paste0("Always pick ", inp$pick, ". Return only that shape.\n\n", inp$shapes)
    })
    res <- ellmer::parallel_chat(ch, prompts, ...)
    list(result = purrr::map_chr(res, \(c) c$last_turn()@text), solver_chat = res)
  }
}

}
\seealso{
\code{\link[=generate_structured]{generate_structured()}} for structured output extraction.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/solver.R
\name{generate_structured}
\alias{generate_structured}
\title{Convert a chat to a solver function with structured output}
\usage{
generate_structured(solver_chat = NULL, type = NULL)
}
\arguments{
\item{solver_chat}{An ellmer chat object, such as from
\code{\link[ellmer:chat_anthropic]{ellmer::chat_claude()}}, or a zero-argument function that returns one.}

\item{type}{A type specification for the extracted data, created with
ellmer's \verb{type_*()} functions (e.g., \code{\link[ellmer:type_boolean]{ellmer::type_object()}},
\code{\link[ellmer:type_boolean]{ellmer::type_string()}}). This defines the schema for the structured
output.}
}
\value{
The output of \code{generate()} is another function. That function takes in
a vector of \code{input}s, as well as a solver chat by the
name of \code{solver_chat} with the default supplied to \code{generate()} itself.

See the documentation for the \code{solver} argument in \link{Task} for more
information on the return type.
}
\description{
\code{generate_structured()} is a variant of \code{\link[=generate]{generate()}} that uses
\code{\link[ellmer:parallel_chat]{ellmer::parallel_chat_structured()}} to extract structured data from
the model's responses. This allows you to define a schema for the
expected output using ellmer's \verb{type_*()} functions.

Because \code{parallel_chat_structured()} returns structured data rather than
Chat objects, \code{generate_structured()} creates synthetic Chat objects
for logging purposes. These "mock" chats contain the input and
JSON-serialized output as turns, but won't include actual token usage
or timing metadata from the API.

The \code{result} field contains JSON-serialized strings for compatibility
with existing scorers. The raw structured data is available in
\verb{$get_samples()$solver_metadata} after calling \verb{$solve()} or \verb{$eval()}.
}
\examples{
if (FALSE) {
  library(ellmer)

  type_answer <- type_object(
    answer = type_string(
      "The author's first name, with no other formatting."
    )
  )

  names <- tibble::tribble(
    ~input,                                  ~target,
    "Name's Josiah, how's it going?",        "Josiah",
    "I'm Lin, what's your name?",            "Lin",
    "My name is Em Fields, how about you?",  "Em"
  )

  tsk <- Task$new(
    dataset = names,
    solver = generate_structured(
      solver_chat = chat_anthropic(model = "claude-sonnet-4-20250514"),
      type = type_answer
    ),
    scorer = detect_match("any")
  )

  tsk$eval()

  # the result is JSON-serialized for compatibility with scorers
  tsk$get_samples()$result

  # raw structured data is available in solver_metadata
  tsk$get_samples()$solver_metadata

  # solver_chat contains synthetic turns for logging
  tsk$get_samples()$solver_chat[[1]]
}
}
\seealso{
\code{\link[=generate]{generate()}} for unstructured output, \code{\link[ellmer:type_boolean]{ellmer::type_object()}} and
related functions for defining type specifications.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scorer-model.R
\name{scorer_model}
\alias{scorer_model}
\alias{model_graded_qa}
\alias{model_graded_fact}
\title{Model-based scoring}
\usage{
model_graded_qa(
  template = NULL,
  instructions = NULL,
  grade_pattern = "(?i)GRADE\\\\s*:\\\\s*([CPI])(.*)$",
  partial_credit = FALSE,
  scorer_chat = NULL
)

model_graded_fact(
  template = NULL,
  instructions = NULL,
  grade_pattern = "(?i)GRADE\\\\s*:\\\\s*([CPI])(.*)$",
  partial_credit = FALSE,
  scorer_chat = NULL
)
}
\arguments{
\item{template}{Grading template to use--a \code{glue()} string which will take
substitutions \code{input}, \code{answer}, \code{criterion}, \code{instructions}.}

\item{instructions}{Grading instructions. If provided, this completely
replaces the default instructions, which specify e.g. how the grader should
format its output (e.g. \code{"GRADE: C"}).}

\item{grade_pattern}{A regex pattern to extract the final grade from the
judge model's response.}

\item{partial_credit}{Whether to allow partial credit.}

\item{scorer_chat}{An ellmer chat used to grade the model output, e.g.
\code{\link[ellmer:chat_anthropic]{ellmer::chat_claude()}}.}
}
\value{
A function that will grade model responses according to the given instructions.
See \link{Task}'s \code{scorer} argument for a description of the returned function.
The functions that \code{model_graded_qa()} and \code{model_graded_fact()} output
can be passed directly to \verb{$eval()}.

See the documentation for the \code{scorer} argument in \link{Task} for more
information on the return type.
}
\description{
Model-based scoring makes use of a model to score output from a solver.
\itemize{
\item \code{model_graded_qa()} scores how well a solver answers a question/answer task.
\item \code{model_graded_fact()} determines whether a solver includes a given fact
in its response.
}

The two scorers are quite similar in their implementation, but use a different
default \code{template} to evaluate correctness.
}
\examples{
# Quality assurance -----------------------------
if (!identical(Sys.getenv("ANTHROPIC_API_KEY"), "")) {
  # set the log directory to a temporary directory
  withr::local_envvar(VITALS_LOG_DIR = withr::local_tempdir())

  library(ellmer)
  library(tibble)

  simple_addition <- tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(solver_chat = chat_claude(model = "claude-sonnet-4-5-20250929")),
    scorer = model_graded_qa()
  )

  tsk$eval()
}

# Factual response -------------------------------
if (!identical(Sys.getenv("ANTHROPIC_API_KEY"), "")) {
  # set the log directory to a temporary directory
  withr::local_envvar(VITALS_LOG_DIR = withr::local_tempdir())

  library(ellmer)
  library(tibble)

  r_history <- tibble(
    input = c(
      "Who created the R programming language?",
      "In what year was version 1.0 of R released?"
    ),
    target = c("Ross Ihaka and Robert Gentleman.", "2000.")
  )

  tsk <- Task$new(
    dataset = r_history,
    solver = generate(solver_chat = chat_claude(model = "claude-sonnet-4-5-20250929")),
    scorer = model_graded_fact()
  )

  tsk$eval()
}

}
\seealso{
\link{scorer_detect} for string detection-based scoring.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vitals-package.R
\docType{package}
\name{vitals-package}
\alias{vitals}
\alias{vitals-package}
\title{vitals: Large Language Model Evaluation}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

A port of 'Inspect', a widely adopted 'Python' framework for large language model evaluation. Specifically aimed at 'ellmer' users who want to measure the effectiveness of their large language model-based products, the package supports prompt engineering, tool usage, multi-turn dialog, and model graded evaluations.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/tidyverse/vitals}
  \item \url{https://vitals.tidyverse.org}
  \item Report bugs at \url{https://github.com/tidyverse/vitals/issues}
}

}
\author{
\strong{Maintainer}: Simon Couch \email{simon.couch@posit.co} (\href{https://orcid.org/0000-0001-5676-5107}{ORCID})

Other contributors:
\itemize{
  \item Max Kuhn \email{max@posit.co} [contributor]
  \item Hadley Wickham \email{hadley@posit.co} (\href{https://orcid.org/0000-0003-4757-117X}{ORCID}) [contributor]
  \item Mine Cetinkaya-Rundel \email{mine@posit.co} (\href{https://orcid.org/0000-0001-6452-2420}{ORCID}) [contributor]
  \item Posit Software, PBC (\href{https://ror.org/03wc8by49}{ROR}) [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/view.R
\name{vitals_view}
\alias{vitals_view}
\title{Interactively view local evaluation logs}
\usage{
vitals_view(dir = vitals_log_dir(), host = "127.0.0.1", port = NULL)
}
\arguments{
\item{dir}{Path to a directory containing task eval logs.}

\item{host}{Host to serve on. Defaults to "127.0.0.1".}

\item{port}{Port to serve on. If NULL, will find a random available port.}
}
\value{
The server object (invisibly)
}
\description{
vitals bundles the Inspect log viewer, an interactive app for exploring
evaluation logs. Supply a path to a directory of tasks written to json.
For individual \link{Task} objects, use the \verb{$view()} method instead.
}
\examples{
if (!identical(Sys.getenv("ANTHROPIC_API_KEY"), "")) {
  # set the log directory to a temporary directory
  withr::local_envvar(VITALS_LOG_DIR = withr::local_tempdir())

  library(ellmer)
  library(tibble)

  simple_addition <- tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  # create a new Task
  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(chat_claude(model = "claude-sonnet-4-5-20250929")),
    scorer = model_graded_qa()
  )

  # evaluate the task (runs solver and scorer) and opens
  # the results in the Inspect log viewer (if interactive)
  tsk$eval()

  # $eval() is shorthand for:
  tsk$solve()
  tsk$score()
  tsk$measure()
  tsk$log()
  tsk$view()

  # get the evaluation results as a data frame
  tsk$get_samples()

  # view the task directory with $view() or vitals_view()
  vitals_view()
}

# The `input` column can be a list of 1-row tibbles for per-sample metadata.
# Custom solvers can then extract columns from each input:
shapes_data <- tibble::tibble(
  input = list(
    tibble::tibble(shapes = "square, circle, rhombus", pick = "square"),
    tibble::tibble(shapes = "square, circle, rhombus", pick = "circle")
  ),
  target = c("square", "circle")
)

my_solver <- function(solver_chat = NULL) {
  chat <- solver_chat
  function(inputs, ..., solver_chat = chat) {
    ch <- if (is.function(solver_chat)) solver_chat() else solver_chat$clone()
    prompts <- lapply(inputs, function(inp) {
      paste0("Always pick ", inp$pick, ". Return only that shape.\n\n", inp$shapes)
    })
    res <- ellmer::parallel_chat(ch, prompts, ...)
    list(result = purrr::map_chr(res, \(c) c$last_turn()@text), solver_chat = res)
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/task.R
\name{Task}
\alias{Task}
\title{Creating and evaluating tasks}
\description{
Evaluation \code{Task}s provide a flexible data structure for evaluating LLM-based
tools.
\enumerate{
\item \strong{Datasets} contain a set of labelled samples. Datasets are just a
tibble with columns \code{input} and \code{target}, where \code{input} is a prompt
and \code{target} is either literal value(s) or grading guidance.
\item \strong{Solvers} evaluate the \code{input} in the dataset and produce a final result.
\item \strong{Scorers} evaluate the final output of solvers. They may use text
comparisons (like \code{\link[=detect_match]{detect_match()}}), model grading (like
\code{\link[=model_graded_qa]{model_graded_qa()}}), or other custom schemes.
}

\strong{The usual flow of LLM evaluation with Tasks calls \verb{$new()} and then \verb{$eval()}.}
\verb{$eval()} just calls \verb{$solve()}, \verb{$score()}, \verb{$measure()}, \verb{$log()},
and \verb{$view()} in order. The remaining methods are generally only
recommended for expert use.
}
\examples{
if (!identical(Sys.getenv("ANTHROPIC_API_KEY"), "")) {
  # set the log directory to a temporary directory
  withr::local_envvar(VITALS_LOG_DIR = withr::local_tempdir())

  library(ellmer)
  library(tibble)

  simple_addition <- tibble(
    input = c("What's 2+2?", "What's 2+3?"),
    target = c("4", "5")
  )

  # create a new Task
  tsk <- Task$new(
    dataset = simple_addition,
    solver = generate(chat_claude(model = "claude-sonnet-4-5-20250929")),
    scorer = model_graded_qa()
  )

  # evaluate the task (runs solver and scorer) and opens
  # the results in the Inspect log viewer (if interactive)
  tsk$eval()

  # $eval() is shorthand for:
  tsk$solve()
  tsk$score()
  tsk$measure()
  tsk$log()
  tsk$view()

  # get the evaluation results as a data frame
  tsk$get_samples()

  # view the task directory with $view() or vitals_view()
  vitals_view()
}

# The `input` column can be a list of 1-row tibbles for per-sample metadata.
# Custom solvers can then extract columns from each input:
shapes_data <- tibble::tibble(
  input = list(
    tibble::tibble(shapes = "square, circle, rhombus", pick = "square"),
    tibble::tibble(shapes = "square, circle, rhombus", pick = "circle")
  ),
  target = c("square", "circle")
)

my_solver <- function(solver_chat = NULL) {
  chat <- solver_chat
  function(inputs, ..., solver_chat = chat) {
    ch <- if (is.function(solver_chat)) solver_chat() else solver_chat$clone()
    prompts <- lapply(inputs, function(inp) {
      paste0("Always pick ", inp$pick, ". Return only that shape.\n\n", inp$shapes)
    })
    res <- ellmer::parallel_chat(ch, prompts, ...)
    list(result = purrr::map_chr(res, \(c) c$last_turn()@text), solver_chat = res)
  }
}

}
\seealso{
\code{\link[=generate]{generate()}} for the simplest possible solver, and
\link{scorer_model} and \link{scorer_detect} for two built-in approaches to
scoring.
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{dir}}{The directory where evaluation logs will be written to. Defaults
to \code{vitals_log_dir()}.}

\item{\code{metrics}}{A named vector of metric values resulting from \verb{$measure()}
(called inside of \verb{$eval()}). Will be \code{NULL} if metrics have yet to
be applied.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Task-new}{\code{Task$new()}}
\item \href{#method-Task-eval}{\code{Task$eval()}}
\item \href{#method-Task-get_samples}{\code{Task$get_samples()}}
\item \href{#method-Task-solve}{\code{Task$solve()}}
\item \href{#method-Task-score}{\code{Task$score()}}
\item \href{#method-Task-measure}{\code{Task$measure()}}
\item \href{#method-Task-log}{\code{Task$log()}}
\item \href{#method-Task-view}{\code{Task$view()}}
\item \href{#method-Task-set_solver}{\code{Task$set_solver()}}
\item \href{#method-Task-set_scorer}{\code{Task$set_scorer()}}
\item \href{#method-Task-set_metrics}{\code{Task$set_metrics()}}
\item \href{#method-Task-get_cost}{\code{Task$get_cost()}}
\item \href{#method-Task-clone}{\code{Task$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-new"></a>}}
\if{latex}{\out{\hypertarget{method-Task-new}{}}}
\subsection{Method \code{new()}}{
The typical flow of LLM evaluation with vitals tends to involve first
calling this method and then \verb{$eval()} on the resulting object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Task$new(
  dataset,
  solver,
  scorer,
  metrics = NULL,
  epochs = NULL,
  name = deparse(substitute(dataset)),
  dir = vitals_log_dir()
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{dataset}}{A tibble with, minimally, columns \code{input} and \code{target}.
The \code{input} column can be either a character vector or a list-column
of 1-row tibbles. Using 1-row tibbles allows per-sample customization
by including additional metadata that custom solvers can access.}

\item{\code{solver}}{A function that takes a vector of inputs from the
dataset's \code{input} column as its first argument and determines values
approximating \code{dataset$target}. Its return value must be a list with
the following elements:
\itemize{
\item \code{result} - A character vector of the final responses, with the same length
as \code{dataset$input}.
\item \code{solver_chat} - A list of ellmer Chat objects that were used to solve
each input, also with the same length as \code{dataset$input}.
}

Additional output elements can be included in a slot \code{solver_metadata} that
has the same length as \code{dataset$input}, which will be logged in
\code{solver_metadata}.

Additional arguments can be passed to the solver via \verb{$solve(...)}
or \verb{$eval(...)}. See the definition of \code{\link[=generate]{generate()}} for a function that
outputs a valid solver that just passes inputs to ellmer Chat objects'
\verb{$chat()} method in parallel.}

\item{\code{scorer}}{A function that evaluates how well the solver's return value
approximates the corresponding elements of \code{dataset$target}. The function
should take in the \verb{$get_samples()} slot of a Task object and return a list with
the following elements:
\itemize{
\item \code{score} - A vector of scores with length equal to \code{nrow(samples)}.
Built-in scorers return ordered factors with
levels \code{I} < \code{P} (optionally) < \code{C} (standing for "Incorrect", "Partially
Correct", and "Correct"). If your scorer returns this output type, the
package will automatically calculate metrics.
}

Optionally:
\itemize{
\item \code{scorer_chat} - If your scorer makes use of ellmer, also include a list of
ellmer Chat objects that were used to score each result, also with
length \code{nrow(samples)}.
\item \code{scorer_metadata} - Any intermediate results or other values that you'd
like to be stored in the persistent log. This should also have length
equal to \code{nrow(samples)}.
}

Scorers will probably make use of \code{samples$input}, \code{samples$target}, and
\code{samples$result} specifically. See \link[=scorer_model]{model-based scoring}
for examples.}

\item{\code{metrics}}{A named list of functions that take in a vector of scores
(as in \code{task$get_samples()$score}) and output a single numeric value.}

\item{\code{epochs}}{The number of times to repeat each sample. Evaluate each sample
multiple times to better quantify variation. Optional, defaults to \code{1L}.
The value of \code{epochs} supplied to \verb{$eval()} or \verb{$score()} will take
precedence over the value in \verb{$new()}.}

\item{\code{name}}{A name for the evaluation task. Defaults to
\code{deparse(substitute(dataset))}.}

\item{\code{dir}}{Directory where logs should be stored.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new Task object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-eval"></a>}}
\if{latex}{\out{\hypertarget{method-Task-eval}{}}}
\subsection{Method \code{eval()}}{
Evaluates the task by running the solver, scorer, logging results, and
viewing (if interactive). This method works by calling \verb{$solve()},
\verb{$score()}, \verb{$log()}, and \verb{$view()} in sequence.

The typical flow of LLM evaluation with vitals tends to involve first
calling \verb{$new()} and then this method on the resulting object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Task$eval(..., epochs = NULL, view = interactive())}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Additional arguments passed to the solver and scorer functions.
All arguments must be named. Arguments are routed based on function
signatures: if an argument name matches a parameter in the solver, it goes
to the solver; if it matches a parameter in the scorer, it goes to the
scorer. Arguments matching both go to both. Unmatched arguments are passed
to any function with \code{...} in its signature. An error is raised if an
argument matches neither function and neither accepts \code{...}.}

\item{\code{epochs}}{The number of times to repeat each sample. Evaluate each sample
multiple times to better quantify variation. Optional, defaults to \code{1L}.
The value of \code{epochs} supplied to \verb{$eval()} or \verb{$score()} will take
precedence over the value in \verb{$new()}.}

\item{\code{view}}{Automatically open the viewer after evaluation (defaults to
TRUE if interactive, FALSE otherwise).}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The Task object (invisibly)
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-get_samples"></a>}}
\if{latex}{\out{\hypertarget{method-Task-get_samples}{}}}
\subsection{Method \code{get_samples()}}{
The task's samples represent the evaluation in a data frame format.

\code{\link[=vitals_bind]{vitals_bind()}} row-binds the output of this
function called across several tasks.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Task$get_samples()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A tibble representing the evaluation. Based on the \code{dataset},
\code{epochs} may duplicate rows, and the solver and scorer will append
columns to this data.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-solve"></a>}}
\if{latex}{\out{\hypertarget{method-Task-solve}{}}}
\subsection{Method \code{solve()}}{
Solve the task by running the solver
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Task$solve(..., epochs = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Additional arguments passed to the solver function.}

\item{\code{epochs}}{The number of times to repeat each sample. Evaluate each sample
multiple times to better quantify variation. Optional, defaults to \code{1L}.
The value of \code{epochs} supplied to \verb{$eval()} or \verb{$score()} will take
precedence over the value in \verb{$new()}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The Task object (invisibly)
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-score"></a>}}
\if{latex}{\out{\hypertarget{method-Task-score}{}}}
\subsection{Method \code{score()}}{
Score the task by running the scorer and then applying metrics to
its results.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Task$score(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Additional arguments passed to the scorer function.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The Task object (invisibly)
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-measure"></a>}}
\if{latex}{\out{\hypertarget{method-Task-measure}{}}}
\subsection{Method \code{measure()}}{
Applies metrics to a scored Task.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Task$measure()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
The Task object (invisibly)
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-log"></a>}}
\if{latex}{\out{\hypertarget{method-Task-log}{}}}
\subsection{Method \code{log()}}{
Log the task to a directory.

Note that, if an \code{VITALS_LOG_DIR} envvar is set, this will happen
automatically in \verb{$eval()}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Task$log(dir = self$dir)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{dir}}{The directory to write the log to.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The path to the logged file, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-view"></a>}}
\if{latex}{\out{\hypertarget{method-Task-view}{}}}
\subsection{Method \code{view()}}{
View the task results in the Inspect log viewer
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Task$view()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
The Task object (invisibly)
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-set_solver"></a>}}
\if{latex}{\out{\hypertarget{method-Task-set_solver}{}}}
\subsection{Method \code{set_solver()}}{
Set the solver function
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Task$set_solver(solver)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{solver}}{A function that takes a vector of inputs from the
dataset's \code{input} column as its first argument and determines values
approximating \code{dataset$target}. Its return value must be a list with
the following elements:
\itemize{
\item \code{result} - A character vector of the final responses, with the same length
as \code{dataset$input}.
\item \code{solver_chat} - A list of ellmer Chat objects that were used to solve
each input, also with the same length as \code{dataset$input}.
}

Additional output elements can be included in a slot \code{solver_metadata} that
has the same length as \code{dataset$input}, which will be logged in
\code{solver_metadata}.

Additional arguments can be passed to the solver via \verb{$solve(...)}
or \verb{$eval(...)}. See the definition of \code{\link[=generate]{generate()}} for a function that
outputs a valid solver that just passes inputs to ellmer Chat objects'
\verb{$chat()} method in parallel.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The Task object (invisibly)
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-set_scorer"></a>}}
\if{latex}{\out{\hypertarget{method-Task-set_scorer}{}}}
\subsection{Method \code{set_scorer()}}{
Set the scorer function
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Task$set_scorer(scorer)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{scorer}}{A function that evaluates how well the solver's return value
approximates the corresponding elements of \code{dataset$target}. The function
should take in the \verb{$get_samples()} slot of a Task object and return a list with
the following elements:
\itemize{
\item \code{score} - A vector of scores with length equal to \code{nrow(samples)}.
Built-in scorers return ordered factors with
levels \code{I} < \code{P} (optionally) < \code{C} (standing for "Incorrect", "Partially
Correct", and "Correct"). If your scorer returns this output type, the
package will automatically calculate metrics.
}

Optionally:
\itemize{
\item \code{scorer_chat} - If your scorer makes use of ellmer, also include a list of
ellmer Chat objects that were used to score each result, also with
length \code{nrow(samples)}.
\item \code{scorer_metadata} - Any intermediate results or other values that you'd
like to be stored in the persistent log. This should also have length
equal to \code{nrow(samples)}.
}

Scorers will probably make use of \code{samples$input}, \code{samples$target}, and
\code{samples$result} specifically. See \link[=scorer_model]{model-based scoring}
for examples.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The Task object (invisibly)
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-set_metrics"></a>}}
\if{latex}{\out{\hypertarget{method-Task-set_metrics}{}}}
\subsection{Method \code{set_metrics()}}{
Set the metrics that will be applied in \verb{$measure()} (and thus \verb{$eval()}).
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Task$set_metrics(metrics)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{metrics}}{A named list of functions that take in a vector of scores
(as in \code{task$get_samples()$score}) and output a single numeric value.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The Task (invisibly)
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-get_cost"></a>}}
\if{latex}{\out{\hypertarget{method-Task-get_cost}{}}}
\subsection{Method \code{get_cost()}}{
The cost of this eval
This is a wrapper around ellmer's \verb{$token_usage()} function.
That function is called at the beginning and end of each call to
\verb{$solve()} and \verb{$score()}; this function returns the cost inferred
by taking the differences in values of \verb{$token_usage()} over time.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Task$get_cost()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A tibble displaying the cost of solving and scoring the
evaluation by model, separately for the solver and scorer.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Task-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Task$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}
